module hdf5_wrapper

  use hdf5
  implicit none
  save
  public

  ! Parameters that may be altered from within user programs:
  ! VERBOSITY, 0=Quiet, 1=Talkative
  integer, public :: HDF_VERBOSITY = 0

  ! 0=Don't suppress errors, 1=Suppress Errors
  integer, public :: HDF_ERROR_SUPPRESS = 1

  ! maximum number of files which can be open
  integer, private, parameter :: nfilemax = 1000

  ! table of currently open files
  integer, private :: nopen = 0
  integer(hid_t), dimension(nfilemax), private :: file_id
  logical, dimension(nfilemax), private        :: read_only
  character(len=1000), dimension(nfilemax)     :: fname_table

  ! is this the first call?
  logical, private :: needInit = .true.

  ! Data types for reading integers (there's no KIND=... support in HDF5!)
  integer(hid_t), private :: NATIVE_INTEGER4
  integer(hid_t), private :: NATIVE_INTEGER8

  ! Data types for reading reals (default REAL and DOUBLE PRECISION may be the 
  ! same, so we can't use those if the library is to be portable)
  integer(hid_t), private :: NATIVE_REAL4
  integer(hid_t), private :: NATIVE_REAL8


  ! Interface for generic read data routine
  interface hdf5_read_data
     module procedure read_scalar_real
     module procedure read_1d_real_array
     module procedure read_2d_real_array
     module procedure read_3d_real_array
     module procedure read_4d_real_array
     module procedure read_5d_real_array
     module procedure read_6d_real_array
     module procedure read_7d_real_array
     module procedure read_scalar_double
     module procedure read_1d_double_array
     module procedure read_2d_double_array
     module procedure read_3d_double_array
     module procedure read_4d_double_array
     module procedure read_5d_double_array
     module procedure read_6d_double_array
     module procedure read_7d_double_array
     module procedure read_scalar_integer4
     module procedure read_1d_integer4_array
     module procedure read_2d_integer4_array
     module procedure read_3d_integer4_array
     module procedure read_4d_integer4_array
     module procedure read_5d_integer4_array
     module procedure read_6d_integer4_array
     module procedure read_7d_integer4_array
     module procedure read_scalar_integer8
     module procedure read_1d_integer8_array
     module procedure read_2d_integer8_array
     module procedure read_3d_integer8_array
     module procedure read_4d_integer8_array
     module procedure read_5d_integer8_array
     module procedure read_6d_integer8_array
     module procedure read_7d_integer8_array
     module procedure read_scalar_string
     module procedure read_1d_string_array
     module procedure read_2d_string_array
     module procedure read_3d_string_array
     module procedure read_4d_string_array
     module procedure read_5d_string_array
     module procedure read_6d_string_array
     module procedure read_7d_string_array
  end interface

  ! interface for generic write_data routine
  interface hdf5_write_data
     module procedure write_1d_integer4_array
     module procedure write_2d_integer4_array
     module procedure write_3d_integer4_array
     module procedure write_4d_integer4_array
     module procedure write_5d_integer4_array
     module procedure write_6d_integer4_array
     module procedure write_7d_integer4_array
     module procedure write_scalar_integer4
     module procedure write_1d_integer8_array
     module procedure write_2d_integer8_array
     module procedure write_3d_integer8_array
     module procedure write_4d_integer8_array
     module procedure write_5d_integer8_array
     module procedure write_6d_integer8_array
     module procedure write_7d_integer8_array
     module procedure write_scalar_integer8
     module procedure write_1d_real_array
     module procedure write_2d_real_array
     module procedure write_3d_real_array
     module procedure write_4d_real_array
     module procedure write_5d_real_array
     module procedure write_6d_real_array
     module procedure write_7d_real_array
     module procedure write_scalar_real
     module procedure write_1d_double_array
     module procedure write_2d_double_array
     module procedure write_3d_double_array
     module procedure write_4d_double_array
     module procedure write_5d_double_array
     module procedure write_6d_double_array
     module procedure write_7d_double_array
     module procedure write_scalar_double
     module procedure write_1d_string_array
     module procedure write_2d_string_array
     module procedure write_3d_string_array
     module procedure write_4d_string_array
     module procedure write_5d_string_array
     module procedure write_6d_string_array
     module procedure write_7d_string_array
     module procedure write_scalar_string
  end interface

  ! interface for generic read_attribute routine
  interface hdf5_read_attribute
     module procedure read_1d_integer4_array_attribute
     module procedure read_2d_integer4_array_attribute
     module procedure read_3d_integer4_array_attribute
     module procedure read_4d_integer4_array_attribute
     module procedure read_5d_integer4_array_attribute
     module procedure read_6d_integer4_array_attribute
     module procedure read_7d_integer4_array_attribute
     module procedure read_scalar_integer4_attribute
     module procedure read_1d_integer8_array_attribute
     module procedure read_2d_integer8_array_attribute
     module procedure read_3d_integer8_array_attribute
     module procedure read_4d_integer8_array_attribute
     module procedure read_5d_integer8_array_attribute
     module procedure read_6d_integer8_array_attribute
     module procedure read_7d_integer8_array_attribute
     module procedure read_scalar_integer8_attribute
     module procedure read_1d_real_array_attribute
     module procedure read_2d_real_array_attribute
     module procedure read_3d_real_array_attribute
     module procedure read_4d_real_array_attribute
     module procedure read_5d_real_array_attribute
     module procedure read_6d_real_array_attribute
     module procedure read_7d_real_array_attribute
     module procedure read_scalar_real_attribute
     module procedure read_1d_double_array_attribute
     module procedure read_2d_double_array_attribute
     module procedure read_3d_double_array_attribute
     module procedure read_4d_double_array_attribute
     module procedure read_5d_double_array_attribute
     module procedure read_6d_double_array_attribute
     module procedure read_7d_double_array_attribute
     module procedure read_scalar_double_attribute
     module procedure read_1d_string_array_attribute
     module procedure read_2d_string_array_attribute
     module procedure read_3d_string_array_attribute
     module procedure read_4d_string_array_attribute
     module procedure read_5d_string_array_attribute
     module procedure read_6d_string_array_attribute
     module procedure read_7d_string_array_attribute
     module procedure read_scalar_string_attribute
  end interface

  interface hdf5_write_attribute
     module procedure write_1d_integer4_array_attribute
     module procedure write_2d_integer4_array_attribute
     module procedure write_3d_integer4_array_attribute
     module procedure write_4d_integer4_array_attribute
     module procedure write_5d_integer4_array_attribute
     module procedure write_6d_integer4_array_attribute
     module procedure write_7d_integer4_array_attribute
     module procedure write_scalar_integer4_attribute
     module procedure write_1d_integer8_array_attribute
     module procedure write_2d_integer8_array_attribute
     module procedure write_3d_integer8_array_attribute
     module procedure write_4d_integer8_array_attribute
     module procedure write_5d_integer8_array_attribute
     module procedure write_6d_integer8_array_attribute
     module procedure write_7d_integer8_array_attribute
     module procedure write_scalar_integer8_attribute
     module procedure write_1d_real_array_attribute
     module procedure write_2d_real_array_attribute
     module procedure write_3d_real_array_attribute
     module procedure write_4d_real_array_attribute
     module procedure write_5d_real_array_attribute
     module procedure write_6d_real_array_attribute
     module procedure write_7d_real_array_attribute
     module procedure write_scalar_real_attribute
     module procedure write_1d_double_array_attribute
     module procedure write_2d_double_array_attribute
     module procedure write_3d_double_array_attribute
     module procedure write_4d_double_array_attribute
     module procedure write_5d_double_array_attribute
     module procedure write_6d_double_array_attribute
     module procedure write_7d_double_array_attribute
     module procedure write_scalar_double_attribute
     module procedure write_1d_string_array_attribute
     module procedure write_2d_string_array_attribute
     module procedure write_3d_string_array_attribute
     module procedure write_4d_string_array_attribute
     module procedure write_5d_string_array_attribute
     module procedure write_6d_string_array_attribute
     module procedure write_7d_string_array_attribute
     module procedure write_scalar_string_attribute
  end interface
contains

  subroutine define_types
    !
    ! Create type IDs for 4 and 8 byte INTEGER and REAL data
    !
    implicit none
    integer :: i, hdf_err

   call byte_order(i)

    if(i.eq.1)then
       ! Little endian machine
       if(HDF_VERBOSITY.eq.1) THEN
          write(*,*)'[define_types] This machine is little endian.'
       ENDIF
       NATIVE_INTEGER8 = H5T_STD_I64LE
       NATIVE_INTEGER4 = H5T_STD_I32LE
       NATIVE_REAL4    = H5T_IEEE_F32LE
       NATIVE_REAL8    = H5T_IEEE_F64LE
    else
       ! Big endian machine
       if(HDF_VERBOSITY.eq.1) THEN
          write(*,*)'[define_types] This machine is big endian.'
       ENDIF
       NATIVE_INTEGER8 = H5T_STD_I64BE
       NATIVE_INTEGER4 = H5T_STD_I32BE
       NATIVE_REAL4    = H5T_IEEE_F32BE
       NATIVE_REAL8    = H5T_IEEE_F64BE
    end if

    return
  end subroutine define_types



  subroutine hdf5_open_file(handle, fname, readonly)
    !
    ! open an existing file for reading or writing and return an integer
    ! file handle (a default integer, not an integer(hid_t)!)
    !  
    implicit none
    integer                      :: handle
    character(len=*), intent(in) :: fname
    integer(hid_t)               :: hdf_file_id
    integer                      :: hdf_err, hdf_err2
    integer                      :: i
    logical, optional            :: readonly

    if(needInit) then
       needInit = .false.
       fname_table(1:nfilemax) = '*'
       call h5open_f(hdf_err)
       if (HDF_ERROR_SUPPRESS .eq. 0) then
          call h5eset_auto_f(1, hdf_err2)
       else
          !Verbosity = 0, supress HDF5 error messages
          call h5eset_auto_f(0, hdf_err2)
       endif
       if(hdf_err.lt.0)call hdf5_abort( &
            'Unable to initialize HDF5 to open file',fname=fname)
       file_id=-1
       call define_types()
    end if

    if (present(readonly)) then
       if(readonly)then
          if (HDF_VERBOSITY .ge. 1) then
             write(*,*)'[hdf5_open_file] File ',trim(fname),' Opened Read Only'
          endif
          call h5fopen_f (fname,H5F_ACC_RDONLY_F,hdf_file_id,hdf_err)
       else
          if (HDF_VERBOSITY .ge. 1) then
             write(*,*)'[hdf5_open_file] File ',trim(fname), &
                  ' Opened Read/Write'
          endif
          call h5fopen_f (fname,H5F_ACC_RDWR_F,hdf_file_id,hdf_err)
       end if
    else
       if (HDF_VERBOSITY .ge. 1) then
          write(*,*)'[hdf5_open_file] File ',trim(fname),' Opened Read/Write'
       endif
       call h5fopen_f (fname,H5F_ACC_RDWR_F,hdf_file_id,hdf_err)
    endif

    if(hdf_err.lt.0) call hdf5_abort( &
         'Unable to open HDF5 file in open_file()!',fname=fname)

    i=1
    do while(file_id(i).ge.0)
       i=i+1
       if(i.gt.nfilemax) call hdf5_abort( &
            'Too many files open in open_file()!',fname=fname)
    end do
    file_id(i)=hdf_file_id
    nopen=nopen+1

    handle=i

    read_only(handle)=.false.
    if(present(readonly))then
       if(readonly)then
          read_only(handle)=.true.
       endif
    endif

    fname_table(handle) = fname

    return
  end subroutine hdf5_open_file



  subroutine hdf5_create_file(handle, fname)
    !
    ! Create a new file and return an integer
    ! file handle (a default integer, not an integer(hid_t)!)
    !  
    implicit none
    integer :: handle
    character(LEN=*), intent(IN) :: fname
    integer(HID_T) :: hdf_file_id
    integer :: hdf_err, hdf_err2
    integer :: i

    if(needInit)then
       needInit=.false.
       fname_table(1:nfilemax) = '*'
       call h5open_f(hdf_err)
       if (HDF_ERROR_SUPPRESS .gt. 0) then
          call h5eset_auto_f(1, hdf_err2)
       else
          !Verbosity = 0, supress HDF5 error messages
          call h5eset_auto_f(0, hdf_err2)
       endif
       if(hdf_err.lt.0) call hdf5_abort( &
            'Unable to initialize HDF5 in create_file()!',fname=fname)
       file_id=-1
       call define_types
    end if
    call h5fcreate_f(fname, H5F_ACC_TRUNC_F, hdf_file_id, hdf_err, H5P_DEFAULT_F, H5P_DEFAULT_F)

    if(hdf_err.lt.0) call hdf5_abort( &
         'Unable to open HDF5 file in create_file()!',fname=fname)

    if (HDF_VERBOSITY .ge. 1) then
       write(*,*)'[hdf5_create_file] File ',trim(fname),' Created'
    endif

    i=1
    do while(file_id(i).ge.0)
       i=i+1
       if(i.gt.nfilemax) call hdf5_abort( &
            'Too many files open in create_file()!',fname=fname)
    end do
    file_id(i)=hdf_file_id
    nopen=nopen+1
    handle=i

    read_only(handle) = .false.
    fname_table(handle) = fname

    return
  end subroutine hdf5_create_file


  subroutine hdf5_close_file(ifile)
    !
    ! Close a currently open HDF5 file
    !  
    implicit none
    integer, intent(IN) :: ifile
    integer :: hdf_err

    if(ifile.lt.1.or.ifile.gt.nfilemax)call hdf5_abort( &
         'Invalid file handle in close_file')

    if(file_id(ifile).lt.0)call hdf5_abort( &
         'File is not open in close_file()!')

    if (HDF_VERBOSITY .ge. 1) then
       write(*,*)'[hdf5_close_file] Closed file '
    endif

    call h5fclose_f(file_id(ifile), hdf_err) 

    if(hdf_err.lt.0)call hdf5_abort( &
         'Unable to close HDF5 file in close_file()!',fname=fname_table(ifile))

    nopen=nopen-1
    file_id(ifile)=-1

    if(nopen.eq.0) then 
       call h5close_f(hdf_err)
       needInit = .true.
    endif

    fname_table(ifile) = '*'

    return
  end subroutine hdf5_close_file


  subroutine hdf5_get_dimensions(ifile,name,rank,dims)
!
! Routine to find out the rank and dimensions of a dataset
! or attribute.
!
    implicit none
    integer, intent(IN) :: ifile
    character(len=*), intent(IN) :: name
    integer, intent(OUT) :: rank
    integer, dimension(*), intent(out) :: dims

    integer(hid_t) :: dset_id, dspace_id, attr_id, loc_id
    integer :: hdf_err, hdf_err2
    integer(hsize_t), dimension(10) :: hdf_dims,hdf_maxdims
    logical :: isattribute, isgroup
    integer :: nslash
    character(len=200) :: loc_name,attr_name

    if(ifile.lt.1.or.ifile.gt.nfilemax)call hdf5_abort( &
         'Invalid file handle in get_dimensions',name=name)

    if(file_id(ifile).lt.0) call hdf5_abort( &
         'File is not open in get_dimensions()!', &
         name=name)

    ! Try to open 'name' as a dataset then as an attribute
    call h5eset_auto_f(0, hdf_err2)
    call h5dopen_f(file_id(ifile),name,dset_id,hdf_err)
    if(HDF_ERROR_SUPPRESS .eq. 0)call h5eset_auto_f(1, hdf_err2)

    if(hdf_err.lt.0)then
       ! Its not a dataset - try it as an attribute. Need to identify
       ! what it is an attribute of!
       ! Split 'name' into parent object name and attribute name
       nslash=index(name,"/",.true.)
       if(nslash.eq.0)call hdf5_abort( &
            'Invalid dataset/attribute name in get_dimensions()!', &
            fname=fname_table(ifile),name=name)
       loc_name=name(1:nslash-1)
       attr_name=name(nslash+1:len_trim(name))

       ! Try to open loc_name as a group then as a dataset
       isgroup=.false.
       call h5eset_auto_f(0, hdf_err2)
       call h5dopen_f(file_id(ifile),loc_name,loc_id,hdf_err)
       if(HDF_ERROR_SUPPRESS .eq. 0)call h5eset_auto_f(1, hdf_err2)
       if(hdf_err.lt.0)then
          call h5gopen_f(file_id(ifile),loc_name, loc_id, hdf_err)
          if(hdf_err.lt.0)call hdf5_abort( &
               'Unable to open attribute parent object in get_dimensions()!', &
               name=name,fname=fname_table(ifile))
          isgroup=.true.
       end if

       ! Open the attribute and get its dataspace
       call h5aopen_name_f(loc_id,attr_name,attr_id,hdf_err) 
       if(hdf_err.lt.0)call hdf5_abort( &
            'Cannot open datset/attribute in get_dimensions()!', &
            name=name,fname=fname_table(ifile))
       call h5aget_space_f(attr_id, dspace_id, hdf_err)
       isattribute=.true.

    else
       call h5dget_space_f(dset_id, dspace_id, hdf_err)
       isattribute=.false.
    end if

    call h5sget_simple_extent_ndims_f(dspace_id,rank,hdf_err)
    if(rank.gt.0) then
       call h5sget_simple_extent_dims_f(dspace_id,hdf_dims,hdf_maxdims, &
            hdf_err)
    else
       hdf_dims=0
       hdf_maxdims=0
    endif
    call h5sclose_f(dspace_id,hdf_err)

    if(isattribute)then
       call h5aclose_f(attr_id,hdf_err)
       if(isgroup)then
          call h5gclose_f(loc_id,hdf_err)
       else
          call h5dclose_f(loc_id,hdf_err)
       end if
    else
       call h5dclose_f(dset_id,hdf_err)
    end if

    ! Convert hdf_dims from hsize_t to default integer
    dims(1:rank)=INT(hdf_dims(1:rank),KIND(dims))

    return
  end subroutine hdf5_get_dimensions

!
! Stop cleanly and print an error message
!
  subroutine hdf5_abort(message,name,fname)
    implicit none
    character(len=*)           :: message
    character(len=*), optional :: name, fname
    integer                    :: hdf_err

    call h5close_f(hdf_err)
    write(*,*)' '
    write(*,*)'***Abort HDF5 : ',trim(message)
    if (present(name)) then
       write(*,*)'   identifier : ',trim(name)
    endif
    if (present(fname)) then
       write(*,*)'   file name  : ',trim(fname)
    end if
    write(*,*)' '
    stop
  end subroutine hdf5_abort


!
! Function for creating groups:
!
  subroutine hdf5_create_group(ifile,name)
    implicit none
    integer          :: ifile
    character(len=*) :: name
    character, parameter            :: slash='/'
    integer, parameter              :: MAX_NEST = 10
    integer(hid_t)                  :: id_chain(MAX_NEST)
    integer(hid_t)                  :: group_id, loc_id
    integer, parameter              :: LEN_STR = 256

    character(len=LEN_STR)          :: thestring,currentgroup
    integer          :: i,n_nest, starti, endi, hdf_err

    if(ifile.lt.1.or.ifile.gt.nfilemax)call hdf5_abort( &
         'Invalid file handle in create_group',name=name)

    if(read_only(ifile))call hdf5_abort( &
         "Attempt to create group in read only file!", &
         name=name,fname=fname_table(ifile));

    thestring = name
    i = index(thestring,slash)
    if (i .eq. 1) then
       !strip leading slash
       thestring = thestring(2:LEN_STR)
       i = index(thestring,slash)
    endif

    !Ensure there is a trailing / on the end:
    i = index(thestring,slash,.true.)
    if (i .lt. len_trim(thestring)) then
       thestring(len_trim(thestring)+1:len_trim(thestring)+1) = '/'
    endif
    i = index(thestring,slash)

    loc_id = file_id(ifile)

    starti = 1
    endi   = i
    
    n_nest = 0
    do while(i > 0)
       !Get the name of the current group:
       currentgroup = thestring(1:i-1)
       
       !Cut the current group from the start of the string:
       thestring = thestring(i+1:LEN_STR)
       
       !Update position of the first /
       i = index(thestring,slash)
       
       !Create groups:
       call h5eset_auto_f(0, hdf_err)
       call h5gcreate_f(loc_id, currentgroup, group_id, hdf_err, 0_size_t)
       if (hdf_err .eq. 0 .and. HDF_VERBOSITY .ge. 1) then
          write(*,*)	'[hdf5_create_group] Creating group ',trim(currentgroup)
       endif
       if(HDF_ERROR_SUPPRESS.eq.0)call h5eset_auto_f(1, hdf_err)
       
       call h5gopen_f(loc_id, currentgroup, group_id, hdf_err)
       if(hdf_err.lt.0) then
          call hdf5_abort('Unable to open group in create_group()!', &
               name=name,fname=fname_table(ifile))
       endif
       
       n_nest = n_nest + 1
       id_chain(n_nest) = group_id
       loc_id = group_id
    enddo
    
    !Close the nest of groups:
    do i=n_nest,1,-1
       call h5gclose_f(id_chain(i),hdf_err)
       if(hdf_err .lt. 0) call hdf5_abort( &
            'Unable to close group in write_data()!', &
            name=name,fname=fname_table(ifile))
    enddo
    
  end subroutine hdf5_create_group


!
! Reading and writing subroutines
!
#include "read_dataset_preprocessor.inc"

#include "read_attribute_preprocessor.inc"

#include "write_dataset_preprocessor.inc"
 
#include "write_attribute_preprocessor.inc"


end module hdf5_wrapper
