  subroutine SUB_NAME (ifile,name,arr,start,count)
!
! Read an n-D array dataset from an open file. This assumes some
! preprocessor variables have been set - see hdf_wrapper.F90.
!
! Try to read all data first as a dataset, and failing that as an
! attribute
!
    implicit none
    character(LEN=*), intent(IN) :: name 
    integer, dimension(:), intent(IN), optional :: start,count
    integer,intent(IN) :: ifile
#ifdef SCALAR
    ARR_TYPE :: arr
#else
    ARR_TYPE , dimension ARRAY_DIM :: arr
#endif
    integer :: hdf_err
    integer(HID_T) :: dset_id,dtype_id,dspace_id,memtype_id,memspace_id
    integer(HSSIZE_T), dimension(10) :: hdfstart
    integer(HSIZE_T),  dimension(10) :: hdfcount,maxdims,filedims,memdims,pad_dims,read_dims
    integer :: select_rank
    integer :: class, rank, i, j
    character, parameter            :: slash='/'
    integer(SIZE_T) :: size, nativesize

    ! Check inputs
    if(present(start).and.(.not.present(count)))call hdf5_abort( &
         'If parameter start is present count must be present in read_dataset')
    if(present(count).and.(.not.present(start)))call hdf5_abort( &
         'If parameter count is present start must be present in read_dataset')
    if(ifile.lt.1.or.ifile.gt.nfilemax)call hdf5_abort( &
         'Invalid file handle in read_data',name=name)
    if(file_id(ifile).lt.0) &
         call hdf5_abort('File is not open in read_data()!', &
         name=name)

    ! Determine datatype for array in memory
#ifndef STRING
    call h5tcopy_f( NATIVE_TYPE, memtype_id, hdf_err)
    if(hdf_err.lt.0)call hdf5_abort( &
         'Unable to open datatype in read_data()!', &
         name=name,fname=fname_table(ifile))
#else
    call h5tcopy_f(H5T_NATIVE_CHARACTER, memtype_id, hdf_err)
    call h5tset_size_f(memtype_id, INT(len(arr),KIND=SIZE_T), hdf_err) 
#endif

    ! Open datset and get size
    call h5dopen_f(file_id(ifile),name, dset_id, hdf_err)
    if (hdf_err .lt. 0) then
       call hdf5_abort('Unable to open dataset in read_dataset', &
            name=name,fname=fname_table(ifile))
    endif
    call h5dget_type_f(dset_id, dtype_id, hdf_err)
    call h5tget_size_f(dtype_id, size, hdf_err)
    call h5tget_class_f(dtype_id, class, hdf_err) 
    call h5tget_size_f(memtype_id, nativesize, hdf_err) 
    call h5dget_space_f(dset_id, dspace_id, hdf_err) 
    call h5sget_simple_extent_ndims_f(dspace_id, rank, hdf_err) 
    call h5sget_simple_extent_dims_f(dspace_id, filedims, maxdims, hdf_err) 

    ! Check we can read this type of data into this type of array
    if(class.ne. DATA_CLASS ) then
       call hdf5_abort('Dataset is incompatible type in read_data()', &
            name=name,fname=fname_table(ifile))
	 endif						  
#ifndef STRING
    if(size.ne.nativesize) then
       call hdf5_abort &
            ('Precision in file doesn''t match precision of array in read_data()', &
            name=name,fname=fname_table(ifile))
    endif
#else
    if(size.gt.nativesize) then
       call hdf5_abort &
            ('String in file doesn''t fit in string in memory in read_data()',&
            name=name,fname=fname_table(ifile))
    endif
#endif

    ! Make sure start and count have enough dimensions if present
    if(present(start))then
       if(lbound(start,1).ne.1.or.ubound(start,1).lt. rank ) call hdf5_abort( &
            'start parameter has incorrect dimensions in read_data', &
            name=name,fname=fname_table(ifile))
       if(lbound(count,1).ne.1.or.ubound(count,1).lt. rank ) call hdf5_abort( &
            'count parameter has incorrect dimensions in read_data', &
            name=name,fname=fname_table(ifile))
    endif

    ! Get effective rank of selection in the dataset (ie ignore dimensions
    ! with size 1)
    select_rank = 0
    if(.not.present(start))then
       do i=1,rank,1
          if(filedims(i).gt.1)select_rank=select_rank+1
       end do
    else
       do i=1,rank,1
          if(count(i).gt.1)select_rank=select_rank+1
       end do
    endif
    !if(select_rank.ne. NDIMS )call hdf5_abort( &
    !     "Rank of selection does not match rank of data array in read_data", &
    !     name=name,fname=fname_table(ifile))

    ! Make sure selection fits in array in file
    if(present(start))then
       do i=1,rank,1
          if(start(i).lt.1)call hdf5_abort( &
               "All elements of start must be positive in read_data", &
               name=name,fname=fname_table(ifile))
          if(count(i).lt.1)call hdf5_abort( &
               "All elements of count must be positive in read_data", &
               name=name,fname=fname_table(ifile))
          if(start(i)+count(i)-1.gt.filedims(i))call hdf5_abort( &
               "Selected elements are outside dataset in read_data", &
               name=name,fname=fname_table(ifile))
       end do
    end if

    ! Get size of output array
#ifndef SCALAR
    do i=1, NDIMS ,1
       memdims(i)=ubound(arr,i)-lbound(arr,i)+1
    end do
#endif

    ! Make sure selection fits in the output array
    if(select_rank.gt.0)then
       if(present(start))then
          j = 1
          do i=1, NDIMS ,1
             do while(count(j).eq.1)
                j=j+1
             end do
             if(count(j).gt.memdims(i))call hdf5_abort( &
                  "Selection doesn't fit in output array in read_data", &
                  name=name,fname=fname_table(ifile))
             j=j+1
          end do
       else
          j = 1
          do i=1, NDIMS ,1
             do while(filedims(j).eq.1)
                j=j+1
             end do
             if(filedims(j).gt.memdims(i))call hdf5_abort( &
                  "Dataset doesn't fit in output array in read_data", &
                  name=name,fname=fname_table(ifile))
             j=j+1
          end do
       endif
    endif

    ! Create the memory dataspace - this needs to have the same rank as the
    ! file/selection dataspace so it may need some padding with 1-element
    ! dimensions if the dataset or selection has 1-element dimensions.
    ! This means the user doesn't have to pass silly things like arr(1:n,1:1)
    ! to read 1 1D slice of a 2D dataset into a 1D array.
    if( NDIMS .gt.0)then
       if(present(start))then
          read_dims(1:rank) = count(1:rank)
       else
          read_dims(1:rank) = filedims(1:rank)
       endif

       j=1
       do i=1,rank,1
          if(read_dims(i).gt.1)then
             pad_dims(i) = memdims(j)
             j=j+1
          else
             pad_dims(i) = 1
          endif
       end do

       maxdims=pad_dims
       memdims=pad_dims

       call h5screate_simple_f(rank,memdims,memspace_id,hdf_err,maxdims) 
    else
       ! Scalar data so use a scalar memory dataspace
       call h5screate_f(H5S_SCALAR_F, memspace_id, hdf_err)
       if(present(start))then
          do i=1,rank,1
             if(count(i).ne.1)call hdf5_abort( &
                  "Can only read one element into a scalar variable in read_data", &
                  name=name,fname=fname_table(ifile))
          end do
       endif
    endif

    ! Select a region in the memory dataspace corresponding to the
    ! file dataspace, or the selection if start/count are specified
#ifndef SCALAR
    hdfstart(1:rank)=0
    hdfcount(1:rank)=filedims(1:rank)
    if(present(start))then
       hdfcount(1:rank) = count(1:rank)
    endif
    call h5sselect_hyperslab_f(memspace_id,H5S_SELECT_SET_F,hdfstart, &
    hdfcount,hdf_err)
#endif

    ! Select a region in the file dataspace corresponding to
    ! the selection
    if(present(start))then
       hdfstart(1:rank) = start(1:rank)-1
       hdfcount(1:rank) = count(1:rank)
       call h5sselect_hyperslab_f(dspace_id,H5S_SELECT_SET_F,hdfstart, &
            hdfcount,hdf_err)
    endif

    call read_hdf5_dataset(dset_id, memtype_id , memspace_id, &
         dspace_id,arr,hdf_err)

    if(hdf_err.lt.0) then
       call hdf5_abort('Unable to read dataset in read_data()', &
            name=name,fname=fname_table(ifile))
    endif

    if (HDF_VERBOSITY .ge. 1) then
       write(*,*)'[read_dataset] Read dataset ',trim(name) 
    endif

    call h5dclose_f(dset_id,hdf_err)
    call h5tclose_f(dtype_id,hdf_err)
    call h5tclose_f(memtype_id,hdf_err)
    call h5sclose_f(dspace_id,hdf_err)
    call h5sclose_f(memspace_id,hdf_err)

    return
  end subroutine SUB_NAME
