;*****************************************************************
;*                       PINOCCHI0  V2.2                         *
;*  (PINpointing Orbit-Crossing Collapsed HIerarchical Objects)  *
;*****************************************************************
;
;This code was written by
;Pierluigi Monaco, Tom Theuns, Giuliano Taffoni
;Copyright (C) 2008
;
;web page: http//adlibitum.oats.inaf.it/monaco/Homepage/Pinocchio/
;
;This program is free software; you can redistribute it and/or modify
;it under the terms of the GNU General Public License as published by
;the Free Software Foundation; either version 2 of the License, or
;(at your option) any later version.
;
;This program is distributed in the hope that it will be useful,
;but WITHOUT ANY WARRANTY; without even the implied warranty of
;MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;GNU General Public License for more details.
;
;You should have received a copy of the GNU General Public License
;along with this program; if not, write to the Free Software
;Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


pro generate_density,seed,Lx,Ly,Lz

; generate potential
phi = double(randomu(seed,Lx,Ly,Lz))
phi = phi-mean(phi,/double)

;write potential field to file
write_dens,'./','phi.d',phi

; second derivatives
ddx  = shift(phi,-1,0,0) + shift(phi,1,0,0)  -2.*phi
ddy  = shift(phi,0,-1,0) + shift(phi,0,1,0)  -2.*phi
ddz  = shift(phi,0,0,-1) + shift(phi,0,0,1)  -2.*phi

; density
rho  = ddx + ddy + ddz

; write density field by plane to file
density_file = 'Density_g1.d'
write_dens,'Data/',density_file,rho

return
end


pro idl_test

; read potential
read_dens,'./','phi.d',phi
dimen = size(phi,/dimen,/L64)
Lx=dimen[0] & Ly=dimen[1] & Lz=dimen[2]

; PART 1 - compute derivatives of idl potential

; first derivative
dx   = (shift(phi,-1,0,0) - shift(phi,1,0,0)) / 2.0d0
dy   = (shift(phi,0,-1,0) - shift(phi,0,1,0)) / 2.0d0
dz   = (shift(phi,0,0,-1) - shift(phi,0,0,1)) / 2.0d0

; second derivatives
ddx  = shift(phi,-1,0,0) + shift(phi,1,0,0)  -2.*phi
ddy  = shift(phi,0,-1,0) + shift(phi,0,1,0)  -2.*phi
ddz  = shift(phi,0,0,-1) + shift(phi,0,0,1)  -2.*phi
dxdy = 0.5*(shift(phi,1,1,0) + shift(phi,-1,-1,0) - shift(phi,1,-1,0) - shift(phi,-1,1,0))
dxdz = 0.5*(shift(phi,1,0,1) + shift(phi,-1,0,-1) - shift(phi,1,0,-1) - shift(phi,-1,0,1))
dydz = 0.5*(shift(phi,0,1,1) + shift(phi,0,-1,-1) - shift(phi,0,1,-1) - shift(phi,0,-1,1))

; obtain density by differencing potential
rho  = ddx + ddy + ddz

; PART 2 - read pinocchio results after compiling with idl test

; warning: take care using correct precision.
dir = './tmp_local/'
ddx_fft  = read_planes(Lz,dir,'save_g1_d2_c1_')
ddy_fft  = read_planes(Lz,dir,'save_g1_d2_c2_')
ddz_fft  = read_planes(Lz,dir,'save_g1_d2_c3_')
dxdy_fft = read_planes(Lz,dir,'save_g1_d2_c4_') 
dxdz_fft = read_planes(Lz,dir,'save_g1_d2_c5_') 
dydz_fft = read_planes(Lz,dir,'save_g1_d2_c6_') 
;
dx_fft   =  read_planes(Lz,dir,'save_g1_d1_c1_') 
dy_fft   =  read_planes(Lz,dir,'save_g1_d1_c2_') 
dz_fft   =  read_planes(Lz,dir,'save_g1_d1_c3_') 
;

; PART 3 - compare
norm = stdev(ddx)
print,' ddx  ',min(abs(ddx -ddx_fft ))/norm,max(abs(ddx -ddx_fft ))/norm
print,' ddy  ',min(abs(ddy -ddy_fft ))/norm,max(abs(ddy -ddy_fft ))/norm
print,' ddz  ',min(abs(ddz -ddz_fft ))/norm,max(abs(ddz -ddz_fft ))/norm
print,' dxdy ',min(abs(dxdy-dxdy_fft))/norm,max(abs(dxdy-dxdy_fft))/norm
print,' dxdz ',min(abs(dxdz-dxdz_fft))/norm,max(abs(dxdz-dxdz_fft))/norm
print,' dydz ',min(abs(dydz-dydz_fft))/norm,max(abs(dydz-dydz_fft))/norm
;
norm = stdev(dx)
print,' dx   ',min(abs(dx-dx_fft)/norm),max(abs(dx-dx_fft)/norm)
print,' dy   ',min(abs(dy-dy_fft)/norm),max(abs(dy-dy_fft)/norm)
print,' dz   ',min(abs(dz-dz_fft)/norm),max(abs(dz-dz_fft)/norm)

return
end

