#!/bin/sh
# *****************************************************************
# *                       PINOCCHI0  V2.2                         *
# *  (PINpointing Orbit-Crossing Collapsed HIerarchical Objects)  *
# *****************************************************************
# 
# This code was written by
# Pierluigi Monaco, Tom Theuns, Giuliano Taffoni
# Copyright (C) 2008
# 
# web page: http//adlibitum.oats.inaf.it/monaco/Homepage/Pinocchio/
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
#########################################################################
#
# RUNFILE FOR IDL TEST OF PINOCCHIO - V2.2
#
#########################################################################

#
# define all the variables
#

run=idltst                 # kernel name of the run *** MUST BE 6 CHAR LONG ***

# you may want to try the test with many combinations of grid lengths
# and in the parallel or serial versions

Lgrid_x=80                 # grid length in the x direction
Lgrid_y=60                 # grid length in the y direction
Lgrid_z=50                 # grid length in the z direction
nproc=1                    # number of processors
parallel=0                 # parallel run if =1, serial run if =0
seed=3337322               # seed for random number generator

# the cosmological parameters are not relevant for this test

Omega0=0.3                 # Omega0
OmegaLambda=0.7            # Omega_Lambda at z=0 (NB a flat Universe is assumed)
PowerSpectrum=-10.         # Power spectrum (see documentation)
Sigma8=0.8                 # Sigma_8
Hubble100=0.7              # Hubble constant in units of 100 km/s/Mpc
BoxSize=100.               # Box size in the x-direction (Mpc, not Mpc/h)

#
# create the fmax.info file 
#

echo "$run                       " >  fmax.info

#
# generate the $run.info file
#

echo "$BoxSize                    " >  $run.info
echo "$Omega0                     " >> $run.info
echo "$OmegaLambda                " >> $run.info
echo "$PowerSpectrum              " >> $run.info
echo "$Sigma8                     " >> $run.info
echo "$Hubble100                  " >> $run.info
echo "$Lgrid_x $Lgrid_y $Lgrid_z  " >> $run.info


#
# create the directory tree
#

echo "--------------------------------------------"
echo  Making the directory tree for the run $run
echo "--------------------------------------------"
echo

mkdir $run              2> error.log 
mkdir $run/Data/        2>> error.log 
mkdir $run/tmp_local/   2>> error.log 
mv $run.info fmax.info  $run/  2>> error.log 
cd $run
ln -s ../../idl/* .

#
# run idl to generate the initial density
#

\rm -f idl_startup
echo ".run read_planes                 " >  idl_startup
echo ".run write_planes                " >> idl_startup
echo ".run write_dens                  " >> idl_startup
echo ".run readfield                   " >> idl_startup
echo ".run test_macros                 " >> idl_startup
echo "seed = long($seed)               " >> idl_startup
echo "Lx   = long($Lgrid_x)            " >> idl_startup
echo "Ly   = long($Lgrid_y)            " >> idl_startup
echo "Lz   = long($Lgrid_z)            " >> idl_startup
echo "generate_density,seed,Lx,Ly,Lz   " >> idl_startup
echo "exit                             " >> idl_startup


echo "-----------------------------------------"
echo  Stating idl to generate the density field
echo  idl will run the following commands:
cat idl_startup
echo "-----------------------------------------"
echo

idl idl_startup

#
# run fmax
#

if [ $parallel -eq 1 ]
then

    echo "-----------------------------------------------"
    echo  Calling parallel fmax-2.2.x with $nproc processors
    echo "-----------------------------------------------"
    echo

    mpirun -np $nproc ../fmax-2.2.x > fmax.log  2>&1
else

    echo "-------------------------"
    echo  Calling serial fmax-2.2.x 
    echo "-------------------------"
    echo

    ../fmax-2.2.x > fmax.log  2>&1
fi

#
# run idl again to perform the test
#

\rm -f idl_startup
echo ".run read_planes                 " >  idl_startup
echo ".run write_planes                " >> idl_startup
echo ".run write_dens                  " >> idl_startup
echo ".run readfield                   " >> idl_startup
echo ".run test_macros                 " >> idl_startup
echo "idl_test                         " >> idl_startup
echo "exit                             " >> idl_startup

echo "------------------------------------"
echo  Stating idl to perform the idl_test
echo  idl will run the following commands:
cat idl_startup
echo "------------------------------------"
echo

idl idl_startup

