#!/bin/sh
# *****************************************************************
# *                       PINOCCHI0  V2.2                         *
# *  (PINpointing Orbit-Crossing Collapsed HIerarchical Objects)  *
# *****************************************************************
# 
# This code was written by
# Pierluigi Monaco, Tom Theuns, Giuliano Taffoni
# Copyright (C) 2008
# 
# web page: http//adlibitum.oats.inaf.it/monaco/Homepage/Pinocchio/
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
#########################################################################
#
# This script runs the pinocchio-1.0 lcdm00 example with both the old (1.0)
# and new (2.2) versions of Pinocchio
#
#########################################################################


nout=4                     # number of intermediate outputs
zfin=0.00                  # final redshift
z1=4.00                    # 
z2=3.00                    # intermediate redshifts
z3=2.00                    #
z4=1.00                    #

nproc=1                    # number of processors
parallel=1                 # parallel run if =1, serial run if =0
hdf5=0                     # if =1 the fmax-2.2 files are in HDF5 format

oldrun='lcdm00'
newrun='oldnew'

ln -s ../../Pinocchio-1.0/rundir/$oldrun .
mkdir $newrun
mkdir $newrun/Data
mkdir $newrun/tmp_local
mkdir $newrun/tmp_global
cp $oldrun/$oldrun.info $newrun/$newrun.info
ln -s ../idl/* .

#
# the old2new code trasforms a file from the 1.0 format to thwe 2.2 one;
# the new format depends on the precompiler option used 
# (unformatted fortran or HDF5)
#

# density
./old2new.x <<EOF
$oldrun
$oldrun/Data/lcdm00.dl
$newrun/Data/Density_g1.d
0
EOF

## Fmax
#./old2new.x <<EOF
#$oldrun
#$oldrun/Data/Fmax.ell.$oldrun.out
#$newrun/Data/Fmax.$oldrun.d
#0
#EOF
#
##vmax-x
#./old2new.x <<EOF
#$oldrun
#$oldrun/Data/vmxx.ell.$oldrun.out
#$newrun/Data/Velx.$oldrun.d
#1
#EOF
#
##vmax-y
#./old2new.x <<EOF
#$oldrun
#$oldrun/Data/vmxy.ell.$oldrun.out
#$newrun/Data/Vely.$oldrun.d
#1
#EOF
#
##vmax-z
#./old2new.x <<EOF
#$oldrun
#$oldrun/Data/vmxz.ell.$oldrun.out
#$newrun/Data/Velz.$oldrun.d
#1
#EOF

#
# Pinocchio-2.2 run with the same initial conditions
#

cd $newrun

echo "$newrun"         > fmax.info

echo "$newrun         ! run                              " >  fragment.info
echo "$newrun         ! kernel                           " >> fragment.info
echo "$zfin           ! z fin                            " >> fragment.info
echo "$nout           ! # outputs                        " >> fragment.info
echo "$z1 $z2 $z3 $z4 ! z out                            " >> fragment.info
echo "0               ! parameter flag (0=default)       " >> fragment.info
echo "4               ! out_flag                         " >> fragment.info
echo "1               ! stop_flag  (0=run again, 1=stop) " >> fragment.info
echo "0               ! limit flag (0=no limits)         " >> fragment.info

if [ $parallel -eq 1 ]
then
    mpirun -np $nproc ../fmax-2.2.x 
else
    ../fmax-2.2.x
fi

../fragment-2.2.x

cd ..

if [ $hdf5 -eq 1 ]
then
    echo ".run readfield                                                        " >  idl_startup
    echo "oldfield = readfield('$oldrun/Data/Fmax.ell.$oldrun.out',128,128,128,/old)" >> idl_startup
    echo "newfield = readfield('$newrun/Data/Fmax.$newrun.d',Lx,Ly,Lz,/slice,/hdf5)    " >> idl_startup
    echo "plot,oldfield,newfield,psym=3,xrange=[-11,15],yrange=[-11,15],/xstyle,/ystyle" >> idl_startup
else
    echo ".run readfield                                                        " >  idl_startup
    echo "oldfield = readfield('$oldrun/Data/Fmax.ell.$oldrun.out',128,128,128,/old)" >> idl_startup
    echo "newfield = readfield('$newrun/Data/Fmax.$newrun.d',Lx,Ly,Lz,/slice)    " >> idl_startup
    echo "plot,oldfield,newfield,psym=3,xrange=[-11,15],yrange=[-11,15],/xstyle,/ystyle" >> idl_startup
fi

idl idl_startup
