#!/bin/sh
# *****************************************************************
# *                       PINOCCHI0  V2.2                         *
# *  (PINpointing Orbit-Crossing Collapsed HIerarchical Objects)  *
# *****************************************************************
# 
# This code was written by
# Pierluigi Monaco, Tom Theuns, Giuliano Taffoni
# Copyright (C) 2008
# 
# web page: http//adlibitum.oats.inaf.it/monaco/Homepage/Pinocchio/
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
#########################################################################
#
# RUNFILE FOR PINOCCHIO - V2.2

#########################################################################

#
# Here all the variables are defined
#

run=lcdm99                 # kernel name of the run *** MUST BE 6 CHAR LONG ***
Lgrid_x=80                 # grid length in the x direction
Lgrid_y=60                 # grid length in the y direction
Lgrid_z=50                 # grid length in the z direction
Omega0=0.3                 # Omega0
OmegaLambda=0.7            # Omega_Lambda at z=0 (NB a flat Universe is assumed)
PowerSpectrum=-10.         # Power spectrum (see documentation)
Sigma8=0.8                 # Sigma_8
Hubble100=0.7              # Hubble constant in units of 100 km/s/Mpc
BoxSize=100.               # Box size in the x-direction (Mpc, not Mpc/h)
seed=3337322               # seed for random number generator
nout=4                     # number of intermediate outputs
zfin=0.00                  # final redshift
z1=4.00                    # 
z2=3.00                    # intermediate redshifts
z3=2.00                    #
z4=1.00                    #

nproc=1                    # number of processors
parallel=0                 # parallel run if =1, serial run if =0

# Other options - see the DOCUMENTATION for details; if you don't known what
# they are, then you likely don't need them!
#final_redshift=0.
#limit_in_delta=0.

#
# Here it creates the fmax.info file 
# Uncomment the keywords that you want to give to fmax
# (see the DOCUMENTATION for details)
#

echo "$run                       " >  fmax.info
echo "GENERATE_DENSITY_FIELD     " >> fmax.info
echo "$seed                      " >> fmax.info
#echo "ALLOC_VERBOSE              " >> fmax.info
#echo "FINAL_REDSHIFT             " >> fmax.info
#echo "$final_redshift            " >> fmax.info
#echo "LIMIT_IN_DELTA             " >> fmax.info
#echo "$limit_in_delta            " >> fmax.info
#echo "USE_SCRATCH_FOR_DERIVATIVES" >> fmax.info
#echo "USE_SCRATCH_FOR_ALL        " >> fmax.info
#echo "OUTPUT_RMAX                " >> fmax.info

#
# Here it creates the $run.info file 
#

echo "$BoxSize                    " >  $run.info
echo "$Omega0                     " >> $run.info
echo "$OmegaLambda                " >> $run.info
echo "$PowerSpectrum              " >> $run.info
echo "$Sigma8                     " >> $run.info
echo "$Hubble100                  " >> $run.info
echo "$Lgrid_x $Lgrid_y $Lgrid_z  " >> $run.info

echo $Hubble100 $Omega0 $BoxSize $Lgrid_x > tmp.file.1
awk '{print log(2.778e11*$1**2*$2*($3/$4)**3)/log(10)}' < tmp.file.1 > tmp.file.2

cat tmp.file.2 >> $run.info

rm -f tmp.file.1 tmp.file.2

#
# Here it creates the file fragment.info
#

echo "$run            ! run                              " >  fragment.info
echo "$run            ! kernel                           " >> fragment.info
echo "$zfin           ! z fin                            " >> fragment.info
echo "$nout           ! # outputs                        " >> fragment.info
echo "$z1 $z2 $z3 $z4 ! z out                            " >> fragment.info
echo "0               ! parameter flag (0=default)       " >> fragment.info
echo "4               ! out_flag                         " >> fragment.info
echo "1               ! stop_flag  (0=run again, 1=stop) " >> fragment.info
echo "0               ! limit flag (0=no limits)         " >> fragment.info

#######################
# Here the run begins #
#######################


#
# Step 1: creates the directory and the .info file
#

mkdir $run              2> error.log 
mkdir $run/Data/        2>> error.log 
mkdir $run/tmp_local/   2>> error.log 
mv $run.info fmax.info fragment.info $run/  2>> error.log 
cd $run

#
# Step 2: runs fmax
#

if [ $parallel -eq 1 ]
then
    mpirun -np $nproc ../fmax-2.2.x > fmax.log  2>&1
else
    ../fmax-2.2.x > fmax.log  2>&1
fi

#
# Step 3: runs fragment
#

../fragment-2.2.x > fragment.log 2>&1

# end

cd ../

