!*****************************************************************
!*                       PINOCCHI0  V2.2                         *
!*  (PINpointing Orbit-Crossing Collapsed HIerarchical Objects)  *
!*****************************************************************
!
!This code was written by
!Pierluigi Monaco, Tom Theuns, Giuliano Taffoni
!Copyright (C) 2008
!
!web page: http//adlibitum.oats.inaf.it/monaco/Homepage/Pinocchio/
!
!This program is free software; you can redistribute it and/or modify
!it under the terms of the GNU General Public License as published by
!the Free Software Foundation; either version 2 of the License, or
!(at your option) any later version.
!
!This program is distributed in the hope that it will be useful,
!but WITHOUT ANY WARRANTY; without even the implied warranty of
!MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!GNU General Public License for more details.
!
!You should have received a copy of the GNU General Public License
!along with this program; if not, write to the Free Software
!Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


! This file contains subroutines for the random numbers of fmax-2.2

#ifdef SPRNG
module sprng
  use numbers
  integer(kind=singleI) ::  streamnum, nstreams, sprng_seed
end module sprng
#endif

#ifndef SPRNG
! insert own random number generator here. Ran3 is added as an example, don't use it.
integer(kind=singleI) function init_random(seed,grid,streamnum)
! initialize random sequence
  use numbers
  implicit none
  integer(kind=singleI), intent(in)    :: seed, grid
  integer(kind=singleI), intent(inout) :: streamnum
  real(kind=doubleR)                   :: r, random

! restart random number generator
  init_random = -abs(seed + grid + streamnum*23)  ! pathetic way for generating indepdent streams
  r           = random(init_random)

  return
end function init_random


real(kind=doubleR) function random(seed)
! Random number generator ran3
  use numbers
  implicit none
  integer(kind=singleI), intent(inout):: seed

!     parameters
  integer(kind=singleI),parameter:: m1=2147483563,a1=40014,q1=53668 &
       ,r1=12211,m2=2147483399,a2=40692,q2=52774,r2=3791 &
       ,m3=m1-1,div=32,med=1+m3/div
  real(kind=doubleR),parameter::small=1.2e-7,rea_1=1./m1,rea_2=1.-small

  integer(kind=singleI),save::seed_save=327104,iv(div)=0,k=0
  integer(kind=singleI)::i,j

     
  if (seed.eq.0) stop 'SEED = 0 in RANDOM'
  if (seed.lt.0) then
!    initialization
     seed=-seed
     seed_save=seed
     do i=div+8,1,-1
        j=seed/q1
        seed=a1*(seed-j*q1)-j*r1
        if (seed.lt.0) then
           seed=seed+m1
        endif
        if (i.le.div) iv(i)=seed
     enddo
     k=iv(1)
  endif
!     updates seed
  j=seed/q1
  seed=a1*(seed-j*q1)-j*r1
  if (seed.lt.0) then
     seed=seed+m1
  endif
!     updates saved seed
  j=seed_save/q2
  seed_save=a2*(seed_save-j*q2)-j*r2
  if (seed_save.lt.0) then
     seed_save=seed_save+m2
  endif
  i=1+k/med
  k=iv(i)-seed_save
  iv(i)=seed
  if (k.lt.1) k=k+m3
!     calculates random number
  random=min(rea_1*k,rea_2)
  return
end function random
#endif
