!*****************************************************************
!*                       PINOCCHI0  V2.2                         *
!*  (PINpointing Orbit-Crossing Collapsed HIerarchical Objects)  *
!*****************************************************************
!
!This code was written by
!Pierluigi Monaco, Tom Theuns, Giuliano Taffoni
!Copyright (C) 2008
!
!web page: http//adlibitum.oats.inaf.it/monaco/Homepage/Pinocchio/
!
!This program is free software; you can redistribute it and/or modify
!it under the terms of the GNU General Public License as published by
!the Free Software Foundation; either version 2 of the License, or
!(at your option) any later version.
!
!This program is distributed in the hope that it will be useful,
!but WITHOUT ANY WARRANTY; without even the implied warranty of
!MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!GNU General Public License for more details.
!
!You should have received a copy of the GNU General Public License
!along with this program; if not, write to the Free Software
!Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


! This code reads a pinocchio-1.0 file and writes it in the pinocchio-2.* format
! NB it is a serial code! please run it on one processor only!

program old2new

  use numbers
  use pinocchio_mpi
  use pinocchio_common
  use fftw_variables, ONLY: nx_fftw, ny_fftw, nz_fftw
  implicit none

! local variables
  character(len=120)::oldfile,newfile
  integer(kind=singleI),dimension(3)::Lgrid_s
  real(kind=singleR),allocatable,dimension(:,:,:)::field
  logical::velfield

! starting MPI
#ifdef MPI

  write(*,*) 'NB: This is a serial code!!!'
  write(*,*) 'Run it without -DMPI or with one processor only'

  call mpi_init(ierr)
  call mpi_comm_rank(mpi_comm_world, proc_id, ierr)
  call mpi_comm_size(mpi_comm_world, nproc, ierr)
#else
  nproc   = 1
  proc_id = 0
#endif

  if (nproc/=1) call abort_code ('Please run with a single processor')

! cosmological info
  write(*,*) 'run'
  read (*,'(a)') run
  call set_run_old2new(Lgrid_s)

  nx_fftw = Lgrid_s(1) ;  ny_fftw = Lgrid_s(2) ;  nz_fftw = Lgrid_s(3)

  print *,nx_fftw,ny_fftw,nz_fftw

! filenames
  write(*,*) 'Name of old pinocchio-1.0 file'
  read (*,'(a)')oldfile
  write(*,*) 'Name of new pinocchio-2.* file'
  read (*,'(a)')newfile
  write(*,*) 'give 1 if it is a velocity field'
  read (*,*) velfield

! reads oldfile
  allocate (field(Lgrid_s(1),Lgrid_s(2),Lgrid_s(3)))
  open (10,file=trim(oldfile),status='old',form='unformatted')
  read (10)field
  close(10)
  write(*,*) 'Read old file '//trim(oldfile)

  if (velfield) field = field / InterPartDist

! writes newfile
  call write_field(Lgrid_s,newfile,field)
  write(*,*) 'Written new file '//trim(newfile)

  stop
end program old2new


subroutine set_run_old2new(Lgrid_s)

  use pinocchio_common
  implicit none

  integer(kind=singleI),intent(out),dimension(3)::Lgrid_s

! local variables
  logical :: cubic=.false., success=.false.
  integer(kind=singleI) :: Lgrid1


  do while(.not.success)
     open  (unit=10,file=run//'/'//run//'.info',status='old',err=40)
     read(10,*)BoxSize(1)
     read(10,*)Omega0
     read(10,*)OmegaLambda
     read(10,*)PowerSpectrum
     read(10,*)Sigma8
     read(10,*)Hubble100
     if(cubic)then
        read(10,*,err=41) Lgrid1
        success = .true.
        Lgrid_s(:)  = Lgrid1
     else
        read(10,*,err=41) Lgrid_s(:)
        success  = .true.
     endif
     
41   close(10)
     if(success) then
        InterPartDist = BoxSize(1)/dble(Lgrid_s(1))
        BoxSize(:)    = InterPartDist*Lgrid_s(:)
     endif
     cubic = .true.
  enddo

  return
40 call abort_code('file '//run//'.info not found')

end subroutine set_run_old2new


subroutine write_field(Lgrid_s,file,field)

! Drives the writing of a field into a pinocchio-2.* file

  use pinocchio_mpi
  use pinocchio_common
  use tmp_plane
  implicit none

! external variables
  integer(kind=singleI),dimension(3)    ,intent(in) :: Lgrid_s
  character(*)                          ,intent(in) :: file
  real(kind=singleR),dimension(Lgrid_s(1),Lgrid_s(2),Lgrid_s(3)),intent(in) :: field

! local variables
  integer(kind=singleI)    :: local_z, global_z, unit


#ifndef HDF5
#ifdef ASCII_WRITE
  call message('writing ascii (debug)')
#else
  call message('writing binary')
#endif
#else
  call message('writing in HDF5 format')
#endif


! allocates the tmp plane
  allocate (plane(Lgrid_s(1),Lgrid_s(2)))

! open the file and write the headers
  unit=1
  call open_file_to_write(0,unit,file)

! write data plane by plane
  do local_z=0,Lgrid_s(3)-1
     plane = field(:,:,local_z+1)

! write the plane
     call write_plane(unit,local_z)

  enddo
  call close_file(unit)

! exit
  return
20 call message(' error writing output file '//trim(file))
  call abort_code(' write_fmax_fields ')
end subroutine write_field

