!*****************************************************************
!*                       PINOCCHI0  V2.2                         *
!*  (PINpointing Orbit-Crossing Collapsed HIerarchical Objects)  *
!*****************************************************************
!
!This code was written by
!Pierluigi Monaco, Tom Theuns, Giuliano Taffoni
!Copyright (C) 2008
!
!web page: http//adlibitum.oats.inaf.it/monaco/Homepage/Pinocchio/
!
!This program is free software; you can redistribute it and/or modify
!it under the terms of the GNU General Public License as published by
!the Free Software Foundation; either version 2 of the License, or
!(at your option) any later version.
!
!This program is distributed in the hope that it will be useful,
!but WITHOUT ANY WARRANTY; without even the implied warranty of
!MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!GNU General Public License for more details.
!
!You should have received a copy of the GNU General Public License
!along with this program; if not, write to the Free Software
!Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


! This file contains the main program of the fmax-2.2 code

program onlysmoothing

  use numbers
  use fmax_smoothing_radii ; use fmax_grids
  use pinocchio_mpi        ; use fmax_run
  use pinocchio_common
  implicit none

! local variables
  integer(kind=singleI):: ismooth, igrid, ipos, ix, iy, iz
  character(len=120)    :: info

  character(len=24):: fdate
  external fdate
#ifdef MPI
  real(kind=doubleR)   :: cputime(0:10)
#endif

! NB MEMTEST runs a memory test on the machine and then stops
!    IDLTEST runs a test on the derivatives of the deformation tensor and then stops


! **************************************************
!                   STARTING MPI
! **************************************************
#ifdef MPI
  call mpi_init(ierr)
  call mpi_comm_rank(mpi_comm_world, proc_id, ierr)
  call mpi_comm_size(mpi_comm_world, nproc, ierr)
#else
  call message ('MPI is not active')
  nproc   = 1
  proc_id = 0
#endif

  call message(' ******* Starting Pinocchio 2.2 ******')
  write(info,'(''['',a24,''] Starting run'')') fdate()
  call message (trim(info))
  write (info,'(''.. mpi started, # processors = '',i6)') nproc
  call message(info)

#ifdef MPI
! timing of the code
  cputime(0) = mpi_wtime()
#endif

! ******************************
! routines to initialize the run
! ******************************
  call initialization()

  if (ngrid>1) call finish ('Only one grid')


! ************************
! CYCLE ON SMOOTHING RADII
! ************************
  do ismooth = 1, nsmooth

     call message ("")
     write(info,'(''['',a24,''] Starting smoothing radius '',i5,'' of '',i5)') fdate(),ismooth,nsmooth
     call message (trim(info))
     write(info,'(''Smoothing radius: '',i3,2(1x,f15.8))') &
          ismooth,Smoothing_Radii(ismooth),sqrt(Smoothing_Variances(ismooth))
     call message (trim(info))

     call compute_density(ismooth,1)

     call compute_only_variance(ismooth,0,0,0)

! *******************************
! END OF CYCLE ON SMOOTHING RADII
! *******************************
     call message ('Smoothing radius completed')
#ifdef MPI
     call mpi_barrier(mpi_comm_world,ierr)
#endif
  enddo

  if (proc_id == 0) then
     open(10,file=run//'.onlysmoothing',status='unknown')
     do ismooth = 1, nsmooth
        write(10,'(3(1x,f15.8))') Smoothing_Radii(ismooth) &
             , Smoothing_Variances(ismooth) &
             , True_Variances(ismooth)
     enddo
     close(10)
     write (*,*) 'Written file '//run//'.onlysmoothing'
  endif

#ifdef MPI
! final cpu time
  cputime(0) = mpi_wtime() - cputime(0)
  write(info,'(''['',a24,''] Finishing, total cputime = '',e12.4)') fdate(),cputime(0)
  call message (trim(info))
#endif

! finalize MPI
  call finish('Fmax done!')

end program onlysmoothing


function fdate ()
  
  character(len=24) :: fdate
  character(len=8)  :: date
  character(len=10) :: time
  character(len=5)  :: timezone
  integer                   :: values(8)
  
  call date_and_time (date, time, timezone, values)
  
  write (fdate, &
       &   "(I2.2,'-',I2.2,'-',I4.4,' ',I2.2,':',I2.2,'.',I2.2,'     ')") &
       &   values(2), values(3), values(1), values(5), values(6), &
       &   values(7)
  
  return
end function fdate
                                                                                                                                     
 
