!*****************************************************************
!*                       PINOCCHI0  V2.2                         *
!*  (PINpointing Orbit-Crossing Collapsed HIerarchical Objects)  *
!*****************************************************************
!
!This code was written by
!Pierluigi Monaco, Tom Theuns, Giuliano Taffoni
!Copyright (C) 2008
!
!web page: http//adlibitum.oats.inaf.it/monaco/Homepage/Pinocchio/
!
!This program is free software; you can redistribute it and/or modify
!it under the terms of the GNU General Public License as published by
!the Free Software Foundation; either version 2 of the License, or
!(at your option) any later version.
!
!This program is distributed in the hope that it will be useful,
!but WITHOUT ANY WARRANTY; without even the implied warranty of
!MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!GNU General Public License for more details.
!
!You should have received a copy of the GNU General Public License
!along with this program; if not, write to the Free Software
!Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

subroutine extract_groups()

! Constructs and orders the list of neighbouring groups of the vertices

  use numbers
  use fragment_mpi
  use GridData
  use GroupData
  implicit none
  integer(kind=singleI) :: ierr,i,id,ng,this_group(6)

! only master does this
  if(MyPE == 0)then
     allocate(GroupIndex(n_total_vertex*GroupsPerVertex),stat=ierr)
     call trace_allocations(ierr,4.*float(n_total_vertex*GroupsPerVertex),.false.,'GroupIndex')
     allocate(GroupName(n_total_vertex*GroupsPerVertex),stat=ierr)
     call trace_allocations(ierr,4.*float(n_total_vertex*GroupsPerVertex),.false.,'GroupName')

! sort on masterID
     ng = 1
     do i=1, n_total_vertex
        GroupName(ng:ng+GroupsPerVertex-1) = GridVertex(i)%groupID(:)
        ng = ng + GroupsPerVertex
     enddo
     call index_sort_i4(n_total_vertex*GroupsPerVertex, GroupName, GroupIndex)
  endif

  return
end subroutine extract_groups


subroutine fetch_next_group_data()

! The master n.0 queries all the other master processors for the list of 
! needed groups.
! NB: the masterID is the ID assigned to the group by the master when it is 
!     created, and runs from first_inode to (NMaster-first_inode)*MaxGroupsOnServer (beyond which 
!     the code aborts).  The Peak field of the GroupDataType gives the globalID
!     of the position of the peak which seeds the group.  Master processors  
!     with MyPE>0 store groups as follows:
!          mygroup = Group%masterID - (GroupPE-first_inode)*MaxGroupsOnServer

  use numbers
  use Fragment_mpi
  use GroupData
  use GridData, only : n_total_vertex, GroupsPerVertex
  implicit none
  integer :: ig, current_group, previous_group
  integer(kind=singleI), allocatable :: n_this_proc(:), n_sent_this_proc(:)
  integer       :: inode, ivertex, isent, istart, inext, ntosend, isend, ierr, ioff, ntorecv, mygroup, i
  logical, save :: first_call = .true.
  integer(kind=singleI), allocatable :: NodesHoc(:), GroupLL(:), GroupNamesThisProc(:)
  type(GroupDataType) :: tmpgroup
#ifdef DEBUG
  character(len=20)   :: fname  
  character(len=31)    :: format_100='(6(e12.4,1x),3(i8,1x),2(i2,1x))'

  write (fname,'(''group.'',i1)') MyPE
  open(unit=20,file=fname,status='unknown')
#endif

  if (my_type == Master) then

     if(first_call)then
        call mpi_create_grouptype()
        first_call = .false.
     endif

     if(MyPE == 0)then
! determine how many groups need sending from Group Server nodes (PEs 1:NMaster-1)
        allocate(n_this_proc(NMaster-1))
        allocate(n_sent_this_proc(NMaster-1))
        call trace_allocations(0,8.*float(NMaster-1),.false.,'n_this_proc')
! linked lists to find groups that belong to each group server
        allocate(NodesHOC(NMaster-1))
        call trace_allocations(0,4.*float(NMaster-1),.false.,'NodesHOC')
        allocate(GroupLL(n_total_vertex*GroupsPerVertex),stat=ierr)
        call trace_allocations(ierr,4.*float(n_total_vertex*GroupsPerVertex),.false.,'GroupLL')

        n_sent_this_proc(:) = 0
        n_this_proc(:)      = 0
        NodesHOC(:)         = 0
        GroupLL(:)          = 0

! PE=0 counts how many groups will be received from each group serving node
        previous_group = -1
        current_group  = GroupName(GroupIndex(1))
        do ig=1, n_total_vertex*GroupsPerVertex ! loop over all groups of all vertices

! Groups are distributed such that Groups 1:MaxGroupsOnServer are on PE=first_inode etc.
! Use linked list to find all groups on a given server node
           current_group    = GroupName(GroupIndex(ig))
           if (current_group .ne. previous_group) then
              previous_group = current_group
              if (current_group > 0) then
                 inode = first_inode + int((current_group-1)/MaxGroupsOnServer) ! group resides on this processor  !  **2G**
                 n_this_proc(inode) = n_this_proc(inode) + 1
! insert into Linked List
                 GroupLL(ig)     = NodesHOC(inode)
                 NodesHOC(inode) = ig
              endif
           endif
        enddo
! total number of groups to receive
        n_master_groups = sum(n_this_proc)

! tell group nodes how many groups it will need to send
        do inode=first_inode, NMaster-1          !  **2G**
           call mpi_send(n_this_proc(inode), 1, mpi_integer, inode, inode, master_comm, ierr)
        enddo
! now send group names to each master node, but maximum MaxGroupsBuffer at a time
! reserve extra space for possibly newly created groups
        n_allocated_groups = n_master_groups + n_total_vertex
        allocate (MasterGroups(n_allocated_groups),stat=ierr)
        call trace_allocations(ierr,float(GroupDataTypeSize*n_allocated_groups),.false.,'MasterGroups')
        do while(sum(n_sent_this_proc) /= sum(n_this_proc))
           do inode=first_inode, NMaster-1         !   **2G**
              if(n_sent_this_proc(inode) /= n_this_proc(inode)) then
! determine next to send
                 isent  = 0
                 istart = NodesHOC(inode)
                 do while (isent .lt. n_sent_this_proc(inode))
                    isent  = isent + 1
                    istart = GroupLL(istart)
                 enddo
! send from istart until no more particles, or MaxGroupsBuffer have been send
                 ntosend = min(MaxGroupsBuffer, n_this_proc(inode)-n_sent_this_proc(inode))
                 allocate(GroupNamesThisProc(ntosend))
                 call trace_allocations(0,4.*float(ntosend),.false.,'GroupNameThisProc')
                 do isend=1, ntosend
                    GroupNamesThisProc(isend) = GroupName(GroupIndex(istart))
                    istart = GroupLL(istart)
                 enddo
                 call mpi_send(GroupNamesThisProc, ntosend, mpi_integer, inode, inode, master_comm, ierr)
                 call mpi_recv(GroupBuffer(1), ntosend*groupdatampisize, mpi_byte, inode, inode, master_comm, mpi_status, ierr)

! NOte: because we used a linked-list to construct the group names, the returned groups are in now sorted
! in *decreasing* ID, we want increasing ID instead

! offset for receiving these groups
                 ioff = sum(n_this_proc(1:inode)) - n_sent_this_proc(inode)
                 do ig=1, ntosend
                    MasterGroups(ioff-ig+1) = GroupBuffer(ig)
                 enddo
                 n_sent_this_proc(inode) = n_sent_this_proc(inode) + ntosend
                 deallocate(GroupNamesThisProc)
                 call trace_allocations(0,-4.*float(ntosend),.false.,'GroupNameThisProc')
              endif
           enddo
        enddo
#ifdef DEBUG
        do i=n_master_groups,1,-1!+n_total_vertex
           write (20,format_100) &
                MasterGroups(i)%ComPos(1),MasterGroups(i)%ComPos(2),MasterGroups(i)%ComPos(3) &
                , MasterGroups(i)%ComVel(1),MasterGroups(i)%ComVel(2),MasterGroups(i)%ComVel(3) &
                , MasterGroups(i)%mass, MasterGroups(i)%MasterID, MasterGroups(i)%Peak &
                , MasterGroups(i)%modified, MasterGroups(i)%alive
        enddo
#endif
        deallocate(n_this_proc)
        deallocate(n_sent_this_proc)
        call trace_allocations(0,-8.*float(NMaster-1),.false.,'n_this_proc')
        deallocate(NodesHOC)
        call trace_allocations(0,-4.*float(NMaster-1),.false.,'NodesHOC')
        deallocate(GroupLL)
        call trace_allocations(0,-4.*float(n_total_vertex*GroupsPerVertex),.false.,'GroupLL')
        deallocate(GroupIndex)
        call trace_allocations(0,-4.*float(n_total_vertex*GroupsPerVertex),.false.,'GroupIndex')
        deallocate(GroupName)
        call trace_allocations(0,-4.*float(n_total_vertex*GroupsPerVertex),.false.,'GroupName')

     else if (GroupPE>=first_inode) then     ! **2G**

! group node receives number of groups from 0
        call mpi_recv(ntosend, 1, mpi_integer, 0, GroupPE, master_comm, mpi_status, ierr)
        isent = 0
        do while(isent /= ntosend)
! receive minimum of MaxGroups, ngroups
           ntorecv = min(MaxGroupsBuffer, ntosend-isent)
           allocate(GroupNamesThisProc(ntorecv))
           call trace_allocations(0,4.*float(ntorecv),.false.,'GroupNameThisProc')
           call mpi_recv(GroupNamesThisProc,ntorecv,mpi_integer,0,GroupPE,master_comm,mpi_status,ierr)
! load and send group properties for these groups
           do ig=1, ntorecv
              mygroup = GroupNamesThisProc(ig) - (GroupPE-first_inode)*MaxGroupsOnServer
              GroupBuffer(ig) = ServerGroups(mygroup)
#ifdef DEBUG
              write (20,format_100) &
                   GroupBuffer(ig)%ComPos(1),GroupBuffer(ig)%ComPos(2),GroupBuffer(ig)%ComPos(3) &
                   , GroupBuffer(ig)%ComVel(1),GroupBuffer(ig)%ComVel(2),GroupBuffer(ig)%ComVel(3) &
                   , GroupBuffer(ig)%mass, GroupBuffer(ig)%MasterID, GroupBuffer(ig)%Peak &
                   , GroupBuffer(ig)%modified, GroupBuffer(ig)%alive
#endif
           enddo
           call mpi_send(GroupBuffer(1), ntorecv*groupdatampisize, mpi_byte, 0, GroupPE, master_comm, ierr)
           isent = isent + ntorecv
           deallocate(GroupNamesThisProc)
           call trace_allocations(0,-4.*float(ntorecv),.false.,'GroupNameThisProc')
       enddo
     endif
  endif

#ifdef DEBUG
  close(20)
#endif

  return
end subroutine fetch_next_group_data


subroutine update_groups()

! After running fragment, masters exchange information on the updated groups

  use numbers
  use fragment_mpi
  use GroupData
  use GridData, only : n_total_vertex
  implicit none
  integer(kind=singleI), allocatable :: n_this_proc(:), n_sent_this_proc(:), HoC(:), LL(:)
  integer(kind=singleI) :: i, inode, ID, igroup, ntosend, nsent, ntorecv, ierr, isent, irecv, ig, mygroup
#ifdef DEBUG
  character(len=20) :: fname
  character(len=31) :: format_100='(6(e12.4,1x),3(i8,1x),2(i2,1x))'

  write (fname,'(''ugroups.'',i1)') MyPE
  open(unit=20,file=fname,status='unknown')
#endif
  if(my_type==Master)then

     if(MyPE == 0)then
        allocate(n_this_proc(NMaster-1)) 
        allocate(n_sent_this_proc(NMaster-1))
        allocate(HOC(NMaster-1))
        allocate(LL(n_master_groups),stat=ierr)
        call trace_allocations(ierr,12.*float(NMaster-1)+4.*n_master_groups,.false.,'LLs for groups')

! count all modified groups for changes of old and new id

! update old id
        HOC(:) = 0 ; LL(:)  = 0
        n_this_proc(:) = 0 ;  n_sent_this_proc(:) = 0
        do i=1, n_master_groups
           if (MasterGroups(i)%modified) then
              ID  = MasterGroups(i)%masterID
                 inode = first_inode+ int((ID-1)/MaxGroupsOnServer) ! group resides on this processor  !   **2G**
                 n_this_proc(inode) = n_this_proc(inode) + 1
                 LL(i)      = HOC(inode)
                 HOC(inode) = i
           endif
        enddo

! let all group servers know how may groups need updating
        do inode=first_inode, NMaster-1   !  **2G**
!           write(*,'("Proc. 0 sending ",i9," groups to proc. ",i3)')n_this_proc(inode),inode
           call mpi_send(n_this_proc(inode), 1, mpi_integer, inode, inode, master_comm, ierr)
        enddo
! now send changed groups in chuncks of MaxGroupBuffer at a time
        do inode=first_inode, NMaster-1   !  **2G**
           igroup = HOC(inode)
           do while(n_sent_this_proc(inode) /= n_this_proc(inode))
              ntosend = 0
              do while(n_sent_this_proc(inode) /= n_this_proc(inode)  .and. ntosend .lt. MaxGroupsBuffer)
                 ntosend  = ntosend + 1
                 GroupBuffer(ntosend) = MasterGroups(igroup)
                 if(igroup > 0) igroup = LL(igroup)
                 n_sent_this_proc(inode) = n_sent_this_proc(inode) + 1
              enddo
              call mpi_send(GroupBuffer, ntosend*GroupDataMPIsize, mpi_byte, inode, inode, master_comm, ierr)
#ifdef DEBUG
              do i=1,ntosend
                 write (20,format_100) &
                      GroupBuffer(i)%ComPos(1),GroupBuffer(i)%ComPos(2),GroupBuffer(i)%ComPos(3) &
                      , GroupBuffer(i)%ComVel(1),GroupBuffer(i)%ComVel(2),GroupBuffer(i)%ComVel(3) &
                      , GroupBuffer(i)%mass, GroupBuffer(i)%MasterID, GroupBuffer(i)%Peak &
                      , GroupBuffer(i)%modified, GroupBuffer(i)%alive
              enddo
#endif
           enddo
        enddo

        deallocate(n_this_proc)
        deallocate(n_sent_this_proc)
        deallocate(HOC)
        deallocate(LL)
        call trace_allocations(0,-12.*float(NMaster-1)-4.*n_master_groups,.false.,'LLs for groups')
        
        deallocate(MasterGroups)
        call trace_allocations(0,-float(GroupDataTypeSize*n_allocated_groups),.false.,'MasterGroups')


     else if (GroupPE>=first_inode) then    !   **2G**
! GroupServer node

! update of OldID
        call mpi_recv(ntosend, 1, mpi_integer, 0, GroupPE, master_comm, mpi_status, ierr)
!        write(*,'("Proc ",i3," receiving ",i9," groups")')GroupPE,ntosend
        isent = 0
! receive ntosend values in chuncks of size MaxGroupBuffer at a time
        do while(isent /= ntosend)
           ntorecv = min(MaxGroupsBuffer, ntosend-isent)
           call mpi_recv(GroupBuffer(1),ntorecv*groupdatampisize,mpi_byte,0,GroupPE,master_comm,mpi_status,ierr)
#ifdef DEBUG
           do i=1,ntorecv
              write (20,format_100) &
                   GroupBuffer(i)%ComPos(1),GroupBuffer(i)%ComPos(2),GroupBuffer(i)%ComPos(3) &
                   , GroupBuffer(i)%ComVel(1),GroupBuffer(i)%ComVel(2),GroupBuffer(i)%ComVel(3) &
                   , GroupBuffer(i)%mass, GroupBuffer(i)%MasterID, GroupBuffer(i)%Peak &
                   , GroupBuffer(i)%modified, GroupBuffer(i)%alive
           enddo
#endif
! update group properties now
           do irecv=1,ntorecv
              mygroup = GroupBuffer(irecv)%masterID - (GroupPE-first_inode)*MaxGroupsOnServer
              ngroups_on_server = max(ngroups_on_server,mygroup)
              ServerGroups(mygroup) = GroupBuffer(irecv)
              ServerGroups(mygroup)%modified = .false.
           enddo
           isent = isent + ntorecv
        enddo
     endif
  endif
#ifdef DEBUG
  close(20)
#endif
  return
end subroutine update_groups


function locate_group(this_group)

! Given MasterGroups, sorted on MasterGroups%ID, find where "this_group" fits
! bisection

  use numbers
  use GroupData, only : MasterGroups, n_master_groups
  implicit none
  integer(kind=singleI), intent(in) :: this_group
  integer(kind=singleI) :: locate_group
  integer group, ilow, ihigh, lname, hname, name
  

  ilow  = 1
  lname = MasterGroups(ilow)%masterID
  if(this_group == lname)then
     locate_group = ilow
     return
  endif
  ihigh = n_master_groups
  hname = MasterGroups(ihigh)%masterID
  if(this_group == hname)then
     locate_group = ihigh
     return
  endif
  locate_group = -1
! fool proof
  if(ihigh == ilow)then
     if(lname /= this_group)then
        write (*,*) ' ERROR: looking for group: ',this_group
        do ilow=1,n_master_groups
           write (*,*) ilow, MasterGroups(ilow)%masterID
        enddo
        call abort_code(' locate_value: error: this_group not found')
     else
        locate_group = ilow
        return
     endif
  endif

  group = 0.5*(ilow+ihigh)
  name  = MasterGroups(group)%masterID
  do while (ilow /= ihigh-1)
     if(this_group > name)then
        ilow = group
     else if (this_group < name) then
        ihigh = group
     else
        locate_group = group
        return
     endif
     group = 0.5*(ilow+ihigh)
     name  = MasterGroups(group)%masterID
  enddo
  hname = MasterGroups(ihigh)%masterID
  if (this_group /= hname) then
     write (*,*) ' ERROR 2: looking for group: ',this_group
     write (*,*) ilow,lname
     write (*,*) ihigh,hname
     do ilow=1,n_master_groups
        write (*,*) ilow, MasterGroups(ilow)%masterID
     enddo
     call abort_code(" locate_value: this_group not found")
  else
     locate_group = ihigh
     return
  endif

  return
end function locate_group


subroutine init_group(group, masterID, Peak, Mass, Posx, Posy, Posz, Velx, Vely, Velz, alive, modified)
  use numbers
  use GroupData, only: GroupDataType
  use grid, only : Lgrid_global
  implicit none
  type(GroupDataType), intent(out) :: Group
  integer(kind=singleI), intent(in) :: Mass, masterID
  real(kind=singleR), intent(in) :: Posx, Posy, Posz, Velx, Vely, Velz
  logical,intent(in) :: alive, modified
  integer(kind=doubleI),intent(in)::Peak

  Group%masterID = masterID
  Group%Peak     = Peak
  Group%Mass     = Mass
  Group%z_appear = -1.
  Group%z_merge  = -1.
  Group%CoMPos(:) = (/Posx, Posy, Posz/)
  Group%CoMVel(:) = (/Velx, Vely, Velz/)
  Group%modified  = modified
  Group%alive     = alive
end subroutine init_group
