!*****************************************************************
!*                       PINOCCHI0  V2.2                         *
!*  (PINpointing Orbit-Crossing Collapsed HIerarchical Objects)  *
!*****************************************************************
!
!This code was written by
!Pierluigi Monaco, Tom Theuns, Giuliano Taffoni
!Copyright (C) 2008
!
!web page: http//adlibitum.oats.inaf.it/monaco/Homepage/Pinocchio/
!
!This program is free software; you can redistribute it and/or modify
!it under the terms of the GNU General Public License as published by
!the Free Software Foundation; either version 2 of the License, or
!(at your option) any later version.
!
!This program is distributed in the hope that it will be useful,
!but WITHOUT ANY WARRANTY; without even the implied warranty of
!MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!GNU General Public License for more details.
!
!You should have received a copy of the GNU General Public License
!along with this program; if not, write to the Free Software
!Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

subroutine fetch_next_grid_data(current_fmax, nfmax)

! Does sending and receiving of a block of gridpoints and filaments

  use numbers
  use fragment_mpi
  use GridData
  use FilamentData
  use grid
  use scratch_io
  implicit none
  real(kind=singleR), intent(in)      :: current_fmax, nfmax
  integer(kind=singleI), allocatable  :: n_to_recv(:), nfilaments_this_plane(:)  !, n_recv(:)
  character(len=24),external          :: fdate
  integer(kind=singleI)               :: n_data, my_n_to_send, isend, ierr, tag, iplane_local, iplane_global &
       , nreceived, n_this_plane
  integer(kind=singleI)               :: index_current_fmax, index_next_fmax, find_index
  integer(kind=singleI)               :: index2D,i,j,k,ii,jj,kk,kp,ineigh,periodic,nmax
  integer(kind=singleI)               :: unit, ioff, my_filaments, fil_counter, tmp_counter, n_to_receive
! saved variables
  logical, save                       :: first_call=.true.
  integer, save                       :: count_routine=0
#ifdef FETCH
  character(len=20)                   :: fetch
  integer, parameter                  :: funit = 30
#endif

#ifdef DEBUG
  character(len=20)                   :: fname
  character(len=25)                   :: format_200 = '(4(e12.4,1x),4(i8,1x),i2)' &
                                        ,format_100 = '(3(e12.4,1x),4(i8,1x),i2)' !,  format_101 = '(i4,3(e12.4,1x),4(i8,1x),i2)'


  write (fname,'(''fetch-vertex.'',i1)') MyPE
  open(unit=20,file=fname,status='unknown')
  write (fname,'(''fetch-filament.'',i1)') MyPE
  open(unit=21,file=fname,status='unknown')
  count_routine = count_routine + 1
#endif
#ifdef FETCH
  write (fetch,'(''fetch.'',i3.3)') MyPE
  if(first_call) then
     open(unit=funit,file=fetch,status='unknown')
  else
     open(unit=funit,file=fetch,status='old',position='append')
  endif
  write (funit,'("[",a24,"] +++ starting ")') fdate()
  call flush(funit)
#endif

! create MPI type corresponding to GridVertex and Filament
  if (first_call) then
     call mpi_create_gridvertex()
     call mpi_create_filament()
     first_call = .false.
  endif

  n_data = LGrid(1)*LGrid(2)
  nmax   = MaxVertexOnMaster / Lgrid_global(3) +1000

! Processor 0
  if(MyPE == 0)then

! receive grid points
     allocate(n_to_recv(NServer),stat=ierr)
     call trace_allocations(ierr,4.*float(Nserver),.false.,'n_to_recv')

! query server processors for the number of vertices to receive
     do tag=0, NServer-1
        call mpi_recv(my_n_to_send, 1, mpi_integer, NMaster+tag, tag, mpi_comm_world, mpi_status, ierr)
        n_to_recv(tag+1) = my_n_to_send
     enddo

     write(*,'("[",a24,"]   +++ Proc. 0 will receive vertices:",99(1x,i9))') fdate(),n_to_recv

#ifdef FETCH
     write (funit,'("[",a24,"]   +++ Proc. 0 will receive vertices:",99(1x,i9))') fdate(),n_to_recv
     call flush(funit)
#endif

     n_total_vertex = sum(n_to_recv)
     allocate (GridVertex(n_total_vertex),stat=ierr)
     call trace_allocations(ierr,float(GridVertexTypeSize*n_total_vertex),.false.,'GridVertex')

     nreceived = 0
     do tag=0,nserver-1
        call mpi_recv(GridVertex(nreceived+1), n_to_recv(tag+1)*gridvertexsize, mpi_byte, NMaster+tag, tag, &
             mpi_comm_world, mpi_status, ierr)
        nreceived = nreceived +  n_to_recv(tag+1)

!     write(*,'("[",a24,"]   +++ Proc. 0 received",i9," vertices from server ",i2)') fdate(),n_to_recv(tag+1),Nmaster+tag

#ifdef FETCH
     write (funit,'("[",a24,"]   +++ Proc. 0 received",i9," vertices from server ",i2)') fdate(),n_to_recv(tag+1),Nmaster+tag
     call flush(funit)
#endif     

     enddo
#ifdef DEBUG
     do i=1, n_total_vertex
        write (20,format_200) &
             GridVertex(i)%fmax,GridVertex(i)%velx,GridVertex(i)%vely,GridVertex(i)%velz &
             , GridVertex(i)%i,GridVertex(i)%j,GridVertex(i)%k,GridVertex(i)%mygroupID !&
!             , GridVertex(i)%modified
     enddo
#endif
! now receive filaments
     n_to_recv(:)    = 0
     do tag=0, NServer-1
        call mpi_recv(my_n_to_send, 1, mpi_integer, NMaster+tag, tag, mpi_comm_world, mpi_status, ierr)
        n_to_recv(tag+1) = my_n_to_send
     enddo

     n_filaments = sum(n_to_recv)
     allocate(MasterFilaments(n_filaments),stat=ierr)
     call trace_allocations(ierr,float(FilamentTypeSize*n_filaments),.false.,'MasterFilaments')

     if (n_filaments > 0) then

        write(*,'("[",a24,"]   +++ Proc. 0 will receive filaments:",99(1x,i9))') fdate(),n_to_recv

        MasterFilaments(:)%MyGroupID = 0
        MasterFilaments(:)%GlobalID = 0
        MasterFilaments(:)%alive    = .false.
        nreceived = 0
        do tag=0,nserver-1
           n_to_receive = n_to_recv(tag+1)
           if (n_to_receive > 0) then
!!$              call mpi_recv(MasterFilaments(nreceived+1), n_to_receive, FilamentDataMPIType, NMaster+tag, tag &
!!$                   , mpi_comm_world, mpi_status, ierr)
              call mpi_recv(MasterFilaments(nreceived+1), n_to_receive*FilamentDataMPISize, mpi_byte, NMaster+tag, &
                   tag, mpi_comm_world, mpi_status, ierr)
              
!!$              call mpi_recv(MasterFilaments(1), n_to_receive, FilamentDataMPIType, NMaster+tag, tag &
!!$                   , mpi_comm_world, mpi_status, ierr)

              nreceived = nreceived + n_to_recv(tag+1)
!              write(*,'("[",a24,"]   +++ Proc. 0 received",i9," filaments from server ",i2)') fdate(),n_to_recv(tag+1),Nmaster+tag 
              call flush(6)
#ifdef FETCH
              write (funit,'("[",a24,"]   +++ Proc. 0 received",i9," filaments from server ",i2)') fdate(),n_to_recv(tag+1),Nmaster+tag 
              call flush(funit)
#endif     
           endif
        enddo

#ifdef DEBUG
        do i=1,n_filaments
           call get_coordinates(MasterFilaments(i)%GlobalID,ii,jj,kk)
           write(21,format_100)  &
                MasterFilaments(i)%velx,MasterFilaments(i)%vely,MasterFilaments(i)%velz &
!                ,MasterFilaments(i)%i,MasterFilaments(i)%j,MasterFilaments(i)%k &
                ,ii,jj,kk &
                ,MasterFilaments(i)%GlobalID  &
                ,MasterFilaments(i)%alive
        enddo
#endif

     endif

     deallocate(n_to_recv)
     call trace_allocations(0,-4.*float(Nserver),.false.,'n_to_recv')
#ifdef FETCH
     write (funit,'("[",a24,"]   +++ done ")') fdate()
     call flush(funit)
#endif     
     
  else

! Server processors
     if(my_type == server)then

        allocate(GridVertex(nmax*nplanes(GroupPE+1)),stat=ierr)
        call trace_allocations(ierr,float(GridVertexTypeSize*nmax*nplanes(GroupPE+1)),.false.,'GridVertex')
        allocate(indexx(n_data),stat=ierr)
        call trace_allocations(ierr,4.*float(n_data),.false.,'indexx')
        allocate(plane1D(n_data),stat=ierr)
        call trace_allocations(ierr,4.*float(n_data),.false.,'plane1D')
#ifdef FETCH
     write (funit,'("[",a24,"]   +++ done allocations ")') fdate()
     call flush(funit)
#endif     

! read all planes
        my_n_to_send = 0
        my_filaments = 0
        do iplane_local=1, nplanes(GroupPE+1)

           iplane_global = iplane_local + plane_start(GroupPE+1)
#ifdef STORE_FMAX
           indexx = stored_index(:,iplane_local)
           plane1D = stored_fmax(:,iplane_local)
#else
           call read_scratch (iplane_global + index_offset, n_data, indexx)
           call read_scratch (iplane_global + fmax_offset, n_data, plane1D)
#ifdef FETCH
           write (funit,'("[",a24,"]   read index+fmax ",i4," of ",i4)') fdate(),iplane_local,nplanes(groupPE+1)
           call flush(funit)
#endif     
#endif
           index_current_fmax = find_index(n_data, current_fmax, indexx, plane1D)  ! NB: This gives the location BEFORE the point at which Fmax <= current_fmax
           index_next_fmax    = find_index(n_data, nfmax       , indexx, plane1D)
           n_this_plane       = index_next_fmax - index_current_fmax 

! load values of GridVertex
           if(n_this_plane > 0)then
! store fmax
              do isend = 1, n_this_plane
                 GridVertex(my_n_to_send + isend)%fmax = plane1D(indexx(index_current_fmax + isend))
              enddo
! load and store velocities
#ifdef STORE_VEL
              plane1D=stored_vel(1,:,iplane_local)
#else
              call read_scratch (iplane_global + velx_offset, n_data, plane1D)
#ifdef FETCH
              write (funit,'("[",a24,"]   read velx_offset ",i4," of ",i4)') fdate(),iplane_local,nplanes(groupPE+1)
              call flush(funit)
#endif     
#endif
              do isend = 1, n_this_plane
                 GridVertex(my_n_to_send + isend)%velx = plane1D(indexx(index_current_fmax + isend))
              enddo
#ifdef STORE_VEL
              plane1D=stored_vel(2,:,iplane_local)
#else
              call read_scratch (iplane_global + vely_offset, n_data, plane1D)
#ifdef FETCH
              write (funit,'("[",a24,"]   read vely_offset ",i4," of ",i4)') fdate(),iplane_local,nplanes(groupPE+1)
              call flush(funit)
#endif     
#endif
              do isend = 1, n_this_plane
                 GridVertex(my_n_to_send + isend)%vely = plane1D(indexx(index_current_fmax + isend))
              enddo
#ifdef STORE_VEL
              plane1D=stored_vel(3,:,iplane_local)
#else
              call read_scratch (iplane_global + velz_offset, n_data, plane1D)
#ifdef FETCH
              write (funit,'("[",a24,"]   read velz_offset ",i4," of ",i4)') fdate(),iplane_local,nplanes(groupPE+1)
              call flush(funit)
#endif     
#endif
              do isend = 1, n_this_plane
                 GridVertex(my_n_to_send + isend)%velz = plane1D(indexx(index_current_fmax + isend))
              enddo

              do isend = 1, n_this_plane
                 index2D = indexx(index_current_fmax + isend)
! determine 3D coordinates (i,j,k) corresponding to this index
                 call D1_2_D2_coordinates(index2D,i,j)
                 GridVertex(my_n_to_send + isend)%MyGroupID     = server_GroupID(i,j,iplane_local)
!!!                 GridVertex(my_n_to_send + isend)%OldGroupID(0) = server_GroupID(i,j,iplane_local)

                 do ineigh=1,GroupsPerVertex
                    ii = periodic(i + GridOffset(ineigh)%i_offset, LGrid_global(1))
                    jj = periodic(j + GridOffset(ineigh)%j_offset, LGrid_global(2))
                    kk = iplane_local + GridOffset(ineigh)%k_offset

! If any of the neighbours are a filament, label the ID with is_filament (<0)
! so we can load them and send them later
                    if(server_GroupID(ii,jj,kk) < 0) then
                       GridVertex(my_n_to_send + isend)%GroupID(ineigh)    = Is_Filament
!!!                       GridVertex(my_n_to_send + isend)%OldGroupID(ineigh) = Is_Filament
                       if(server_GroupID(ii,jj,kk) == Is_Filament)then
                          server_GroupID(ii,jj,kk) = 2*Is_Filament ! label as such, to avoid loading the same filament more than once.
                          my_filaments = my_filaments + 1
                       endif
                    else
                       GridVertex(my_n_to_send + isend)%GroupID(ineigh)    = server_GroupID(ii,jj,kk)
!!!                       GridVertex(my_n_to_send + isend)%OldGroupID(ineigh) = server_GroupID(ii,jj,kk)
                    endif
                 enddo
! 3D coordinates of this vertex
                 GridVertex(my_n_to_send + isend)%i = i
                 GridVertex(my_n_to_send + isend)%j = j
                 GridVertex(my_n_to_send + isend)%k = iplane_global
              enddo
              my_n_to_send = my_n_to_send + n_this_plane
           endif
        enddo

!        write (*,'("  Server ",i3," will send ",i9," vertices")')MyPE,my_n_to_send

! send vertices
#ifdef FETCH
        write (funit,'("[",a24,"]   sending ",i4)') fdate(),my_n_to_send
        call flush(funit)
#endif     
        call mpi_send(my_n_to_send, 1, mpi_integer, 0, GroupPE, mpi_comm_world, ierr)
!!!        call mpi_send(GridVertex, my_n_to_send, GridVertexMPIType, 0, grouppe, mpi_comm_world, ierr)
        call mpi_send(GridVertex, my_n_to_send*gridvertexsize, mpi_byte, 0, grouppe, mpi_comm_world, ierr)
#ifdef FETCH
        write (funit,'("[",a24,"]   done sending ",i4)') fdate(),my_n_to_send
        call flush(funit)
#endif     
#ifdef DEBUG
        do i=1, my_n_to_send
           write (20,format_200) &
                GridVertex(i)%fmax,GridVertex(i)%velx,GridVertex(i)%vely,GridVertex(i)%velz &
                , GridVertex(i)%i,GridVertex(i)%j,GridVertex(i)%k,GridVertex(i)%mygroupID !&
!                , GridVertex(i)%modified
        enddo
#endif
        deallocate(GridVertex)
        call trace_allocations(0,-float(GridVertexTypeSize*nmax*nplanes(GroupPE+1)),.false.,'GridVertex')
        deallocate(indexx)
        call trace_allocations(0,-4.*float(n_data),.false.,'indexx')

#ifdef FETCH
        write (funit,'("[",a24,"]   +++ done allocations ")') fdate()
        call flush(funit)
#endif     
        
        ! now load and send filaments, if any.
#ifdef FETCH
        write (funit,'("[",a24,"]   +++ filaments to send: ",I4)') fdate(),my_filaments
        call flush(funit)
#endif     
        call mpi_send(my_filaments, 1, mpi_integer, 0, GroupPE, mpi_comm_world, ierr)

!        write (*,'("  Server ",i3," will send ",i9," filaments")')MyPE,my_filaments
#ifdef FETCH
        write (funit,'("[",a24,"]  +++ #  of filaments, I4")') fdate(), my_filaments
        call flush(funit)
#endif     

        if (my_filaments > 0) then

           allocate(ServerFilaments(my_filaments),stat=ierr)
           call trace_allocations(ierr,float(FilamentTypeSize*my_filaments),.false.,'ServerFilaments')
#ifdef FETCH
           write (funit,'("[",a24,"]   +++ done allocations ServerFilaments ")') fdate()
           call flush(funit)
#endif     

           fil_counter=0

           do k=0,LGrid(3)+1                          ! read boundary planes as well
              iplane_global = mod(k + plane_start(GroupPE+1) -1 + LGrid_global(3), LGrid_global(3)) + 1
              kp = periodic(iplane_global, LGrid_global(3))

! do velx
#ifdef STORE_VEL
              plane1D = stored_vel(1,:,k)
#else
              call read_scratch (iplane_global + velx_offset, n_data, plane1D)
#ifdef FETCH
              write (funit,'("[",a24,"]   read velx_offset k",i4," of ",i4)') fdate(),k,Lgrid(3)
              call flush(funit)
#endif     
#endif
              tmp_counter=0
              do j=1,LGrid(2)
                 do i=1,LGrid(1)
                    if(server_GroupID(i,j,k) == 2*Is_Filament)then
                       index2D = i+(j-1)*LGRid(1)
                       tmp_counter=tmp_counter+1
!                       ServerFilaments(fil_counter+tmp_counter)%i = i
!                       ServerFilaments(fil_counter+tmp_counter)%j = j
!                       ServerFilaments(fil_counter+tmp_counter)%k = kp
                       ServerFilaments(fil_counter+tmp_counter)%MyGroupID = is_filament
                       ServerFilaments(fil_counter+tmp_counter)%GlobalID = i + (j-1)*LGrid_global(1) &
                            + (kp-1)*LGrid_global(1)*LGrid_global(2)
                       ServerFilaments(fil_counter+tmp_counter)%alive = .true.
                       ServerFilaments(fil_counter+tmp_counter)%velx  = plane1D(index2D)
                    endif
                 enddo
              enddo

! do vely
#ifdef STORE_VEL
              plane1D = stored_vel(2,:,k)
#else
              call read_scratch (iplane_global + vely_offset, n_data, plane1D)
#ifdef FETCH
              write (funit,'("[",a24,"]   read vely_offset k",i4," of ",i4)') fdate(),k,Lgrid(3)
              call flush(funit)
#endif     
#endif
              tmp_counter = 0
              do j=1,LGrid(2)
                 do i=1,LGrid(1)
                    if(server_GroupID(i,j,k) == 2*Is_Filament)then
                       index2D = i+(j-1)*LGRid(1)
                       tmp_counter=tmp_counter+1
                       ServerFilaments(fil_counter+tmp_counter)%vely = plane1D(index2D)
                    endif
                 enddo
              enddo
                 
! do velz and reset GroupID
#ifdef STORE_VEL
              plane1D = stored_vel(3,:,k)
#else
              call read_scratch (iplane_global + velz_offset, n_data, plane1D)
#ifdef FETCH
              write (funit,'("[",a24,"]   read velz_offset k",i4," of ",i4)') fdate(),k,Lgrid(3)
              call flush(funit)
#endif     
#endif
              tmp_counter = 0
              do j=1,LGrid(2)
                 do i=1,LGrid(1)
                    if (server_GroupID(i,j,k) == 2*Is_Filament)then
                       server_GroupID(i,j,k) = Is_Filament ! reset
                       index2D = i+(j-1)*LGRid(1)
                       tmp_counter=tmp_counter+1
                       ServerFilaments(fil_counter+tmp_counter)%velz = plane1D(index2D)
                    endif
                 enddo
              enddo
              fil_counter = fil_counter + tmp_counter
           enddo

! check on the number of filaments
#ifdef VERBOSE
           write (*,*) 'Check: ',my_filaments,fil_counter
#endif
           if (fil_counter /= my_filaments) call abort_code('Wrong number of filaments!')

! send filaments

!!!           call mpi_send(ServerFilaments, my_filaments, FilamentDataMPIType, 0, GroupPE, mpi_comm_world, ierr)
#ifdef FETCH
           write (funit,'("[",a24,"]   sending filaments ",i4)') fdate(),my_filaments
           call flush(funit)
#endif     

           call mpi_send(ServerFilaments, my_filaments*FilamentDataMPISize, mpi_byte, 0, GroupPE, mpi_comm_world, ierr)
#ifdef FETCH
           write (funit,'("[",a24,"]   done sending filaments ",i4)') fdate(),my_filaments
           call flush(funit)
#endif     
#ifdef DEBUG
           do i=1,my_filaments
              call get_coordinates(ServerFilaments(i)%GlobalID,ii,jj,kk)
              write(21,format_100) ServerFilaments(i)%velx,ServerFilaments(i)%vely,ServerFilaments(i)%velz, &
!                   ServerFilaments(i)%i,ServerFilaments(i)%j,ServerFilaments(i)%k, &
                   ii,jj,kk, &
                   ServerFilaments(i)%GlobalID,ServerFilaments(i)%alive
           enddo
#endif
#ifdef FETCH
           write (funit,'("[",a24,"]   deallocating filaments ",i4)') fdate(),my_filaments
           call flush(funit)
#endif     
           deallocate(ServerFilaments)
           call trace_allocations(0,-float(FilamentTypeSize*my_filaments),.false.,'ServerFilaments')

!        else
!           write (*,'("  Server ",i3," has no filaments to send")') MyPE
        endif

        deallocate(plane1D)
        call trace_allocations(0,-4.*float(n_data),.false.,'plane1D')

#ifdef FETCH
           write (funit,'("[",a24,"]   done filaments")') fdate()
           call flush(funit)
#endif     

     endif
  endif

#ifdef DEBUG
  close(20)
  close(21)
#endif

  return
end subroutine fetch_next_grid_data


!!!subroutine update_vertices()
!!!
!!!! update GroupIDs of vertices.
!!!
!!!  use numbers
!!!  use fragment_mpi
!!!  use GridData
!!!  use Grid, only : LGrid,server_GroupID, plane_start, nplanes, myserver, Lgrid_global
!!!  use FilamentData, only : is_filament, MasterFilaments, n_filaments, FilamentTypeSize
!!!  use borders
!!!  implicit none
!!!
!!!  integer(kind=singleI), allocatable :: n_this_proc(:), HoC(:,:), LL(:,:)
!!!  integer(kind=singleI), external    :: ipbc
!!!  character(len=24),external         :: fdate
!!!  integer(kind=singleI) :: group, ntosend, ierr, ivertex, iserver, isent, ntorecv, i, j, k, local_plane, n_data
!!!  integer(kind=singleI) :: ii, jj, kk, periodic, ineigh, global_plane, iserver_left, iserver_right, ilist, &
!!!       OldID, NewID, left, right
!!!  type(GridVertexType)  :: ThisVertex
!!!
!!!#ifdef FETCH
!!!  character(len=20)                   :: fetch
!!!  integer, parameter                  :: funit = 30
!!!  logical,save                        :: first_call=.true.
!!!#endif
!!!
!!!#ifdef DEBUG
!!!  character(len=40)     :: fname
!!!  character(len=25)     :: format_200 = '(4(e12.4,1x),4(i8,1x),i3)'
!!!
!!!  write (fname,'(''update.'',i1)') MyPe
!!!#endif
!!!
!!!#ifdef FETCH
!!!  write (fetch,'(''update.'',i3.3)') MyPE
!!!  if(first_call) then
!!!     open(unit=funit,file=fetch,status='unknown')
!!!     first_call=.false.
!!!  else
!!!     open(unit=funit,file=fetch,status='old',position='append')
!!!  endif
!!!  write (funit,'("[",a24,"] +++ starting ")') fdate()
!!!  call flush(funit)
!!!#endif
!!!
!!!  if(MyPE == 0)then
!!! 
!!!     allocate(n_this_proc(NServer))
!!!     allocate(HOC(3,NServer))         ! **!!** THREE LINKING LISTS!
!!!     allocate(LL(3,n_total_vertex),stat=ierr)
!!!     call trace_allocations(ierr,4.*float(4*Nserver+3*n_total_vertex),.false.,'LLs')
!!!     n_this_proc(:) = 0
!!!     HOC(:,:) = 0 ; LL(:,:) = 0
!!!     do ivertex=1, n_total_vertex
!!!! where does this vertex live?
!!!        global_plane = GridVertex(ivertex)%k
!!!        iserver = myserver(GridVertex(ivertex)%k)
!!!        n_this_proc(iserver+1) = n_this_proc(iserver+1) + 1
!!!        LL(1,ivertex) = HOC(1,iserver+1)
!!!        HOC(1,iserver+1)  = ivertex
!!!
!!!! vertices in bordering planes are also communicated
!!!        iserver_left  = myserver(periodic(GridVertex(ivertex)%k-1,Lgrid_global(3)))
!!!        if (iserver_left  /= iserver) then
!!!           n_this_proc(iserver_left +1) = n_this_proc(iserver_left+1) + 1
!!!           LL(2,ivertex) = HOC(2,iserver_left+1)
!!!           HOC(2,iserver_left+1) = ivertex
!!!        endif
!!!        iserver_right = myserver(periodic(GridVertex(ivertex)%k+1,Lgrid_global(3)))
!!!        if (iserver_right /= iserver) then
!!!           n_this_proc(iserver_right+1) = n_this_proc(iserver_right+1) + 1
!!!           LL(3,ivertex) = HOC(3,iserver_right+1)
!!!           HOC(3,iserver_right+1) = ivertex
!!!        endif
!!!     enddo
!!!
!!!#ifdef FETCH
!!!     write(*,'("[",a24,"]   +++ Proc. 0 sending lists",99(1x,i9))') fdate(),n_this_proc
!!!     call flush(funit)
!!!#endif
!!!! let all servers know how many vertices they will receive
!!!     do iserver=0,NServer-1
!!!        call mpi_send(n_this_proc(iserver+1),1,mpi_integer,iserver+Nmaster,iserver+NMaster,mpi_comm_world,ierr)
!!!     enddo
!!!
!!!! now send nodes
!!!#ifdef DEBUG
!!!     open(unit=20,file=fname,status='unknown')
!!!#endif
!!!     do iserver=0,NServer-1
!!!        ntosend = n_this_proc(iserver+1)
!!!        if (ntosend > 0)then
!!!           allocate(GridVertexBuffer(ntosend),stat=ierr)
!!!           call trace_allocations(ierr,float(GridVertexTypeSize*ntosend),.false.,'GridVertexBuffer')
!!!           ilist=1
!!!           ivertex = HOC(ilist,iserver+1)
!!!           loop: do isent=1,ntosend
!!!              GridVertexBuffer(isent) = GridVertex(ivertex)
!!!              if (isent == ntosend) exit loop
!!!              if (LL(ilist,ivertex) > 0) then
!!!                 ivertex = LL(ilist,ivertex)
!!!              else
!!!                 ilist = ilist+1
!!!                 ivertex = HOC(ilist,iserver+1)
!!!                 if (ivertex==0) then
!!!                    if (ilist<3) then 
!!!                       ilist=ilist+1
!!!                       ivertex = HOC(ilist,iserver+1)
!!!                       if (ivertex==0) call abort_code ('error in update_vertices')
!!!                    else
!!!                       call abort_code ('error in update_vertices')
!!!                    endif
!!!                 endif
!!!              endif
!!!           enddo loop
!!!           call mpi_send(GridVertexBuffer(1),ntosend*gridvertexsize,mpi_byte,&
!!!                iserver+Nmaster,iserver+NMaster,mpi_comm_world,ierr)
!!!#ifdef FETCH
!!!           write(*,'("[",a24,"]   +++ Proc. 0 sent list to Proc ",i3)') fdate(),iserver+Nmaster
!!!           call flush(6)
!!!#endif
!!!#ifdef DEBUG
!!!           do i=1,ntosend
!!!              write (20,format_200) &
!!!                   GridVertexBuffer(i)%fmax,GridVertexBuffer(i)%velx,GridVertexBuffer(i)%vely,GridVertexBuffer(i)%velz &
!!!                   , GridVertexBuffer(i)%i,GridVertexBuffer(i)%j,GridVertexBuffer(i)%k,GridVertexBuffer(i)%mygroupID !&
!!!!                   , GridVertexBuffer(i)%modified
!!!           enddo
!!!#endif
!!!           
!!!           deallocate(GridVertexBuffer)
!!!           call trace_allocations(0,-float(GridVertexTypeSize*ntosend),.false.,'GridVertexBuffer')
!!!        endif
!!!     enddo
!!!     deallocate(n_this_proc)
!!!     deallocate(HOC)
!!!     deallocate(LL)
!!!     call trace_allocations(0,-4.*float(4*Nserver+3*n_total_vertex),.false.,'LLs')
!!!     deallocate(GridVertex)
!!!     call trace_allocations(0,-float(GridVertexTypeSize*n_total_vertex),.false.,'GridVertex')
!!!     deallocate(MasterFilaments)
!!!     call trace_allocations(0,-float(FilamentTypeSize*n_filaments),.false.,'MasterFilaments')
!!!     
!!!#ifdef DEBUG
!!!     close(20)
!!!#endif
!!!  else 
!!!     if (my_type == server) then
!!!! receive number to receive
!!!#ifdef DEBUG
!!!        open(unit=20,file=fname,status='unknown')
!!!#endif
!!!        call mpi_recv(ntorecv,1,mpi_integer,0,GroupPE+NMaster,mpi_comm_world,mpi_status,ierr)
!!!#ifdef FETCH
!!!        write(funit,'("[",a24,"]   +++ Proc. ",i3," receiving lists: ",i9)') fdate(),MyPE,ntorecv
!!!        call flush(funit)
!!!#endif
!!!
!!!        if(ntorecv > 0)then
!!!           allocate(GridVertexBuffer(ntorecv),stat=ierr)
!!!           call trace_allocations(ierr,float(GridVertexTypeSize*ntorecv),.false.,'GridVertexBuffer')
!!!!!$        call mpi_recv(GridVertexBuffer(1),ntorecv,GridVertexMPIType,0,GroupPE+NMaster,mpi_comm_world,mpi_status,ierr)
!!!           call mpi_recv(GridVertexBuffer(1),ntorecv*gridvertexsize,mpi_byte,0,GroupPE+NMaster,mpi_comm_world,mpi_status,ierr)
!!!#ifdef FETCH
!!!           write(funit,'("[",a24,"]   +++ Proc. ",i3," received lists")') fdate(),MyPE
!!!           call flush(funit)
!!!#endif
!!!#ifdef DEBUG
!!!           do i=1,ntorecv
!!!              write (20,format_200) &
!!!                   GridVertexBuffer(i)%fmax,GridVertexBuffer(i)%velx,GridVertexBuffer(i)%vely,GridVertexBuffer(i)%velz &
!!!                   , GridVertexBuffer(i)%i,GridVertexBuffer(i)%j,GridVertexBuffer(i)%k,GridVertexBuffer(i)%mygroupID !&
!!!!                   , GridVertexBuffer(i)%modified
!!!           enddo
!!!           close(20)
!!!#endif
!!!
!!!! update GroupIDs
!!!! borders
!!!           n_border_left  = 0
!!!           n_border_right = 0
!!!           n_border_max   = 20*Lgrid(1)*Lgrid(2) + 1000
!!!           allocate(border_left(n_border_max),stat=ierr)
!!!           call trace_allocations(ierr,4.*float(n_border_max),.false.,'border_left')
!!!           border_left=0
!!!           allocate(border_right(n_border_max),stat=ierr)
!!!           call trace_allocations(ierr,4.*float(n_border_max),.false.,'border_right')
!!!           border_right=0
!!!
!!!           do ivertex=1, ntorecv
!!!
!!!              ThisVertex  = GridVertexBuffer(ivertex)
!!!              i           = ThisVertex%i
!!!              j           = ThisVertex%j
!!!! PBCs on the global box are taken into account
!!!              if (myserver(ThisVertex%k)==GroupPE .or. Nserver==1) then
!!!                 local_plane = ThisVertex%k - plane_start(GroupPE+1)
!!!              else 
!!!                 if (ThisVertex%k==Lgrid_global(3)) then
!!!                    local_plane = 0
!!!                 else if (ThisVertex%k==1) then
!!!                    local_plane = Lgrid(3)+1
!!!                 else
!!!                    local_plane = ThisVertex%k - plane_start(GroupPE+1)
!!!                 endif
!!!              endif
!!!
!!!              OldID = server_GroupID(i,j,local_plane)
!!!              call update_ID(ThisVertex%MyGroupID,OldID,i,j,local_plane,.true.)
!!!
!!!! neighbours
!!!              do ineigh=1, GroupsPerVertex
!!!                 ii = periodic(i  + GridOffset(ineigh)%i_offset,LGrid_global(1))
!!!                 jj = periodic(j  + GridOffset(ineigh)%j_offset,LGrid_global(2))
!!!                 kk = local_plane + GridOffset(ineigh)%k_offset
!!!                 if (kk>=0 .and. kk<=Lgrid(3)+1) then
!!!                    OldID=server_GroupID(ii,jj,kk)
!!!                    call update_ID(ThisVertex%GroupID(ineigh),OldID,ii,jj,kk,.true.)
!!!                 endif
!!!              enddo
!!!           enddo
!!!           deallocate(GridVertexBuffer)
!!!           call trace_allocations(0,-float(GridVertexTypeSize*ntorecv),.false.,'GridVertexBuffer')
!!!        endif
!!!
!!!#ifdef FETCH
!!!        write(funit,'("[",a24,"]   +++ First update done, now sending/receiving borders: "9(1x,i9))') fdate(),n_border_left,n_border_right
!!!        call flush(funit)
!!!#endif
!!!
!!!! sends and receives changed GroupIDs from neighbours
!!!        left  = mod( GroupPE - 1 + NServer, NServer)
!!!        right = mod( GroupPE + 1 + NServer, NServer)
!!!
!!!! send to left, receive from right
!!!        call mpi_sendrecv(n_border_left, 1, mpi_integer, left, GroupPE &
!!!             ,n_border_from_right, 1, mpi_integer, right, right, server_comm, mpi_status, ierr)
!!!        allocate(border_from_right(1:n_border_left),stat=ierr)
!!!        call trace_allocations(ierr,4.*float(n_border_left),.false.,'border_from_right')
!!!        border_from_right(:) = border_left(1:n_border_left)
!!!        deallocate (border_left)
!!!        call trace_allocations(0,-4.*float(n_border_max),.false.,'border_left')
!!!        allocate (border_left(1:n_border_left),stat=ierr)
!!!        call trace_allocations(ierr,4.*float(n_border_left),.false.,'border_left')
!!!        border_left = border_from_right
!!!        deallocate(border_from_right)
!!!        call trace_allocations(0,-4.*float(n_border_left),.false.,'border_from_right')
!!!        allocate(border_from_right(1:n_border_from_right),stat=ierr)
!!!        call trace_allocations(ierr,4.*float(n_border_from_right),.false.,'border_from_right')
!!!        call mpi_sendrecv(border_left, n_border_left, mpi_integer, left, GroupPE &
!!!             , border_from_right, n_border_from_right, mpi_integer, right, right, server_comm, mpi_status, ierr)
!!!
!!!! send to right, receive from left
!!!        call mpi_sendrecv(n_border_right, 1, mpi_integer, right, GroupPE &
!!!             ,n_border_from_left, 1, mpi_integer, left, left, server_comm, mpi_status, ierr)
!!!        allocate(border_from_left(1:n_border_right),stat=ierr)
!!!        call trace_allocations(ierr,4.*float(n_border_right),.false.,'border_from_left')
!!!        border_from_left(:) = border_right(1:n_border_right)
!!!        deallocate (border_right)
!!!        call trace_allocations(0,-4.*float(n_border_max),.false.,'border_right')
!!!        allocate (border_right(1:n_border_right),stat=ierr)
!!!        call trace_allocations(ierr,4.*float(n_border_right),.false.,'border_right')
!!!        border_right = border_from_left
!!!        deallocate(border_from_left)
!!!        call trace_allocations(0,-4.*float(n_border_right),.false.,'border_from_left')
!!!        allocate(border_from_left(1:n_border_from_left),stat=ierr)
!!!        call trace_allocations(ierr,4.*float(n_border_from_left),.false.,'border_from_left')
!!!        call mpi_sendrecv(border_right, n_border_right, mpi_integer, right, GroupPE &
!!!             , border_from_left, n_border_from_left, mpi_integer, left, left, server_comm, mpi_status, ierr)
!!!
!!!#ifdef FETCH
!!!        write(funit,'("[",a24,"]   +++ Received borders: "9(1x,i9))') fdate(),n_border_from_left,n_border_from_right
!!!        call flush(funit)
!!!#endif
!!!
!!!        do ivertex=1,n_border_from_left/5
!!!           OldID = border_from_left((ivertex-1)*5+1)
!!!           NewID = border_from_left((ivertex-1)*5+2)
!!!           i     = border_from_left((ivertex-1)*5+3)
!!!           j     = border_from_left((ivertex-1)*5+4)
!!!           kk    = border_from_left((ivertex  )*5  )
!!!           k     = ipbc(kk - plane_start(GroupPE+1),Lgrid_global(3))
!!!           if (k<-1) k=k+Lgrid_global(3)
!!!           if (k>Lgrid(3)+2) k=k-Lgrid_global(3)
!!!           call update_ID(NewID,OldID,i,j,k,.false.)
!!!        enddo
!!!
!!!#ifdef FETCH
!!!        write(funit,'("[",a24,"]   +++ updated from left"))') fdate()
!!!        call flush(funit)
!!!#endif
!!!
!!!        do ivertex=1,n_border_from_right/5
!!!           OldID = border_from_right((ivertex-1)*5+1)
!!!           NewID = border_from_right((ivertex-1)*5+2)
!!!           i     = border_from_right((ivertex-1)*5+3)
!!!           j     = border_from_right((ivertex-1)*5+4)
!!!           kk    = border_from_right((ivertex  )*5  )
!!!           k     = ipbc(kk - plane_start(GroupPE+1),Lgrid_global(3))
!!!           if (k<-1) k=k+Lgrid_global(3)
!!!           if (k>Lgrid(3)+2) k=k-Lgrid_global(3)
!!!           call update_ID(NewID,OldID,i,j,k,.false.)
!!!        enddo
!!!
!!!#ifdef UPDATE
!!!        do local_plane = 0, Lgrid(3)+1
!!!           global_plane = ipbc(local_plane + plane_start(GroupPE+1),Lgrid_global(3))
!!!           if (global_plane<10) then
!!!              write(fname,'("server_GroupID.",i1,".00",i1)')MyPE,global_plane
!!!           else if (global_plane<100) then
!!!              write(fname,'("server_GroupID.",i1,".0",i2)')MyPE,global_plane
!!!           else
!!!              write(fname,'("server_GroupID.",i1,".",i3)')MyPE,global_plane
!!!           endif
!!!           open(23,file=trim(fname),status='unknown')
!!!           do j=1,Lgrid(2)
!!!              do i=1,Lgrid(1)
!!!                 write(23,'(4(1x,i9)))')i,j,global_plane,server_GroupID(i,j,local_plane)
!!!              enddo
!!!           enddo
!!!           close(23)
!!!        enddo
!!!#endif
!!!
!!!        deallocate (border_left)
!!!        call trace_allocations(0,-4.*float(n_border_left),.false.,'border_left')
!!!        deallocate (border_right)
!!!        call trace_allocations(0,-4.*float(n_border_right),.false.,'border_right')
!!!        deallocate (border_from_left)
!!!        call trace_allocations(0,-4.*float(n_border_from_left),.false.,'border_from_left')
!!!        deallocate (border_from_right)
!!!        call trace_allocations(0,-4.*float(n_border_from_right),.false.,'border_from_right')
!!!
!!!     endif
!!!  endif
!!!#ifdef DEBUG
!!!  close(20)
!!!#endif
!!!  return
!!!end subroutine update_vertices
!!!
!!!
!!!subroutine update_ID(NewID,OldID,i,j,k,border)
!!!
!!!  use numbers
!!!  use Grid, only : LGrid, server_GroupID
!!!  use fragment_mpi
!!!  implicit none
!!!  integer(kind=singleI),intent(in)::NewID,OldID,i,j,k
!!!  logical,intent(in) :: border
!!!  integer(kind=singleI),external::ipbc
!!!
!!!  integer(kind=singleI)::iloop,ix,iy,iz,iix,iiy,count
!!!  logical::change
!!!
!!!  if (k>=0 .and. k<=Lgrid(3)+1) server_GroupID(i,j,k)=NewID
!!!  if (border .and. OldID/=NewID) call fix_border(OldID,NewID,i,j,k)
!!!
!!!! If the group has changed, propagate the change to all neighbours
!!!  if (NewID > 0 .and. OldID > 0 .and. OldID/=NewID) then
!!!     count=0
!!!     loop: do iloop = 1,maxval(Lgrid)
!!!        change = .false.
!!!        z: do iz=max(0,k-iloop),min(k+iloop,Lgrid(3)+1)
!!!           y: do iiy=j-iloop,j+iloop
!!!              iy = ipbc(iiy,Lgrid(2))
!!!              x: do iix=i-iloop,i+iloop
!!!                 ix = ipbc(iix,Lgrid(1))
!!!                 if (iix/=i-iloop .and. iix/=i+iloop .and. iiy/=j-iloop .and. iiy/=j+iloop .and. iz/=k-iloop &
!!!                      .and. iz/=k+iloop) cycle x
!!!                 if (server_GroupID(ix,iy,iz)==OldID) then
!!!                    change=.true.
!!!                    server_GroupID(ix,iy,iz)=NewID
!!!                    if (border) call fix_border(OldID,NewID,ix,iy,iz)
!!!                 endif
!!!              enddo x
!!!           enddo y
!!!        enddo z
!!!        if (.not.change) count=count+1
!!!        if (count==3) exit loop
!!!     enddo loop
!!!  endif
!!!
!!!  return
!!!end subroutine update_ID


subroutine update_vertices_new()

! update GroupIDs of vertices.

  use numbers
  use fragment_mpi
  use GridData
  use Grid, only : LGrid,server_GroupID, plane_start, nplanes, myserver, Lgrid_global
  use FilamentData, only : is_filament, MasterFilaments, n_filaments, FilamentTypeSize
  use borders
  implicit none

  integer(kind=singleI), allocatable :: n_this_proc(:), HoC(:,:), LL(:,:)
  integer(kind=singleI), external :: ipbc
  character(len=24),external          :: fdate
  integer(kind=singleI) :: group, ntosend, ierr, ivertex, iserver, isent, ntorecv, i, j, k, local_plane, n_data
  integer(kind=singleI) :: ii, jj, kk, periodic, ineigh, global_plane, iserver_left, iserver_right, ilist, &
       OldID, NewID, left, right, n_border_left_old, n_border_right_old, n_border,n_border_total, nn
  type(GridVertexType)  :: ThisVertex
  logical               :: transmit_borders
#ifdef FETCH
  character(len=20)                   :: fetch
  integer, parameter                  :: funit = 30
  logical,save                        :: first_call=.true.
#endif
#ifdef DEBUG
  character(len=40)     :: fname
  character(len=25)     :: format_200 = '(4(e12.4,1x),4(i8,1x),i3)'

  write (fname,'(''update.'',i1)') MyPe
#endif

#ifdef FETCH
  write (fetch,'(''update.'',i3.3)') MyPE
  if(first_call) then
     open(unit=funit,file=fetch,status='unknown')
     first_call=.false.
  else
     open(unit=funit,file=fetch,status='old',position='append')
  endif
  write (funit,'("[",a24,"] +++ starting ")') fdate()
  call flush(funit)
#endif

  if(MyPE == 0)then
 
     allocate(n_this_proc(NServer))
     allocate(HOC(3,NServer))         ! **!!** THREE LINKING LISTS!
     allocate(LL(3,n_total_vertex),stat=ierr)
     call trace_allocations(ierr,4.*float(4*Nserver+3*n_total_vertex),.false.,'LLs')
     n_this_proc(:) = 0
     HOC(:,:) = 0 ; LL(:,:) = 0
     do ivertex=1, n_total_vertex
! where does this vertex live?
        global_plane = GridVertex(ivertex)%k
        iserver = myserver(GridVertex(ivertex)%k)
        n_this_proc(iserver+1) = n_this_proc(iserver+1) + 1
        LL(1,ivertex) = HOC(1,iserver+1)
        HOC(1,iserver+1)  = ivertex

! vertices in bordering planes are also communicated
        iserver_left  = myserver(periodic(GridVertex(ivertex)%k-1,Lgrid_global(3)))
        if (iserver_left  /= iserver) then
           n_this_proc(iserver_left +1) = n_this_proc(iserver_left+1) + 1
           LL(2,ivertex) = HOC(2,iserver_left+1)
           HOC(2,iserver_left+1) = ivertex
        endif
        iserver_right = myserver(periodic(GridVertex(ivertex)%k+1,Lgrid_global(3)))
        if (iserver_right /= iserver) then
           n_this_proc(iserver_right+1) = n_this_proc(iserver_right+1) + 1
           LL(3,ivertex) = HOC(3,iserver_right+1)
           HOC(3,iserver_right+1) = ivertex
        endif
     enddo

! let all servers know how many vertices they will receive
     do iserver=0,NServer-1
        call mpi_send(n_this_proc(iserver+1),1,mpi_integer,iserver+Nmaster,iserver+NMaster,mpi_comm_world,ierr)
     enddo

! now send nodes
#ifdef DEBUG
     open(unit=20,file=fname,status='unknown')
#endif
     do iserver=0,NServer-1
        ntosend = n_this_proc(iserver+1)
        if (ntosend > 0)then
           allocate(GridVertexBuffer(ntosend),stat=ierr)
           call trace_allocations(ierr,float(GridVertexTypeSize*ntosend),.false.,'GridVertexBuffer')
           ilist=1
           ivertex = HOC(ilist,iserver+1)
           loop: do isent=1,ntosend
              GridVertexBuffer(isent) = GridVertex(ivertex)
              if (isent == ntosend) exit loop
              if (LL(ilist,ivertex) > 0) then
                 ivertex = LL(ilist,ivertex)
              else
                 ilist = ilist+1
                 ivertex = HOC(ilist,iserver+1)
                 if (ivertex==0) then
                    if (ilist<3) then 
                       ilist=ilist+1
                       ivertex = HOC(ilist,iserver+1)
                       if (ivertex==0) call abort_code ('error in update_vertices')
                    else
                       call abort_code ('error in update_vertices')
                    endif
                 endif
              endif
           enddo loop
           call mpi_send(GridVertexBuffer(1),ntosend*gridvertexsize,mpi_byte,&
                iserver+Nmaster,iserver+NMaster,mpi_comm_world,ierr)
#ifdef DEBUG
           do i=1,ntosend
              write (20,format_200) &
                   GridVertexBuffer(i)%fmax,GridVertexBuffer(i)%velx,GridVertexBuffer(i)%vely,GridVertexBuffer(i)%velz &
                   , GridVertexBuffer(i)%i,GridVertexBuffer(i)%j,GridVertexBuffer(i)%k,GridVertexBuffer(i)%mygroupID !&
!                   , GridVertexBuffer(i)%modified
           enddo
#endif
           
           deallocate(GridVertexBuffer)
           call trace_allocations(0,-float(GridVertexTypeSize*ntosend),.false.,'GridVertexBuffer')
        endif
     enddo
     deallocate(n_this_proc)
     deallocate(HOC)
     deallocate(LL)
     call trace_allocations(0,-4.*float(4*Nserver+3*n_total_vertex),.false.,'LLs')
     deallocate(GridVertex)
     call trace_allocations(0,-float(GridVertexTypeSize*n_total_vertex),.false.,'GridVertex')
     deallocate(MasterFilaments)
     call trace_allocations(0,-float(FilamentTypeSize*n_filaments),.false.,'MasterFilaments')
     
#ifdef DEBUG
     close(20)
#endif
  else 
     if (my_type == server) then
! receive number to receive
#ifdef DEBUG
        open(unit=20,file=fname,status='unknown')
#endif
        call mpi_recv(ntorecv,1,mpi_integer,0,GroupPE+NMaster,mpi_comm_world,mpi_status,ierr)

        if(ntorecv > 0)then
           allocate(GridVertexBuffer(ntorecv),stat=ierr)
           call trace_allocations(ierr,float(GridVertexTypeSize*ntorecv),.false.,'GridVertexBuffer')
!!$        call mpi_recv(GridVertexBuffer(1),ntorecv,GridVertexMPIType,0,GroupPE+NMaster,mpi_comm_world,mpi_status,ierr)
           call mpi_recv(GridVertexBuffer(1),ntorecv*gridvertexsize,mpi_byte,0,GroupPE+NMaster,mpi_comm_world,mpi_status,ierr)
#ifdef DEBUG
           do i=1,ntorecv
              write (20,format_200) &
                   GridVertexBuffer(i)%fmax,GridVertexBuffer(i)%velx,GridVertexBuffer(i)%vely,GridVertexBuffer(i)%velz &
                   , GridVertexBuffer(i)%i,GridVertexBuffer(i)%j,GridVertexBuffer(i)%k,GridVertexBuffer(i)%mygroupID !&
!                   , GridVertexBuffer(i)%modified
           enddo
           close(20)
#endif
! update GroupIDs
! borders
           n_border_left  = 0
           n_border_right = 0
           n_border_max   = 20*Lgrid(1)*Lgrid(2)   !5*ntorecv ! what is this?
           allocate(border_left(n_border_max),stat=ierr)
           call trace_allocations(ierr,4.*float(n_border_max),.false.,'border_left')
           border_left=0
           allocate(border_right(n_border_max),stat=ierr)
           call trace_allocations(ierr,4.*float(n_border_max),.false.,'border_right')
           border_right=0

           do ivertex=1, ntorecv
              ThisVertex  = GridVertexBuffer(ivertex)
              i           = ThisVertex%i
              j           = ThisVertex%j
! PBCs on the global box are taken into account
              if (myserver(ThisVertex%k)==GroupPE .or. Nserver==1) then
                 local_plane = ThisVertex%k - plane_start(GroupPE+1)
              else 
                 if (ThisVertex%k==Lgrid_global(3)) then
                    local_plane = 0
                 else if (ThisVertex%k==1) then
                    local_plane = Lgrid(3)+1
                 else
                    local_plane = ThisVertex%k - plane_start(GroupPE+1)
                 endif
              endif

              OldID = server_GroupID(i,j,local_plane)
              call update_ID_new(ThisVertex%MyGroupID,OldID,i,j,local_plane,.true.)

! neighbours
              do ineigh=1, GroupsPerVertex
                 ii = periodic(i  + GridOffset(ineigh)%i_offset,LGrid_global(1))
                 jj = periodic(j  + GridOffset(ineigh)%j_offset,LGrid_global(2))
                 kk = local_plane + GridOffset(ineigh)%k_offset
                 if (kk>=0 .and. kk<=Lgrid(3)+1) then
                    OldID=server_GroupID(ii,jj,kk)
                    call update_ID_new(ThisVertex%GroupID(ineigh),OldID,ii,jj,kk,.true.)
                 endif
              enddo
           enddo
           deallocate(GridVertexBuffer)
           call trace_allocations(0,-float(GridVertexTypeSize*ntorecv),.false.,'GridVertexBuffer')
        endif

! the transmission of borders is iterated until no more data to transmit are present
        n_border_left_old=n_border_max
        n_border_right_old=n_border_max
        transmit_borders=.true.
        do while (transmit_borders)

! sends and receives changed GroupIDs from neighbours
           left  = mod( GroupPE - 1 + NServer, NServer)
           right = mod( GroupPE + 1 + NServer, NServer)

           n_border=n_border_left+n_border_right
#ifdef FETCH
           write(funit,'("[",a24,"]   +++ Server ",i3," will communicate borders: ",3(1x,i9))') fdate(),MyPE,n_border_left,n_border_right,n_border
           call flush(funit)
#endif
           call mpi_reduce(n_border,n_border_total,1,MPI_INTEGER,MPI_SUM,0,server_comm,ierr)
           call mpi_bcast(n_border_total,1,MPI_INTEGER,0,server_comm,ierr)
#ifdef FETCH
           write(funit,'("[",a24,"]   +++ communicating borders: ",i9)') fdate(),n_border_total
           call flush(funit)
#endif

! transmit borders if there are any
           if (n_border_total>0) then

              nn=n_border_total/5
              if (GroupPE==0) write(*,'("[",a24,"]   +++ Communicating borders: ",i9)') fdate(),nn

! send to left, receive from right
              call mpi_sendrecv(n_border_left, 1, mpi_integer, left, GroupPE &
                   ,n_border_from_right, 1, mpi_integer, right, right, server_comm, mpi_status, ierr)
              allocate(border_from_right(1:n_border_left),stat=ierr)
              call trace_allocations(ierr,4.*float(n_border_left),.false.,'border_from_right')
              border_from_right(:) = border_left(1:n_border_left)
              deallocate (border_left)
              call trace_allocations(0,-4.*float(n_border_left_old),.false.,'border_left')
              allocate (border_left(1:n_border_left),stat=ierr)
              call trace_allocations(ierr,4.*float(n_border_left),.false.,'border_left')
              border_left = border_from_right
              deallocate(border_from_right)
              call trace_allocations(0,-4.*float(n_border_left),.false.,'border_from_right')
              allocate(border_from_right(1:n_border_from_right),stat=ierr)
              call trace_allocations(ierr,4.*float(n_border_from_right),.false.,'border_from_right')
              call mpi_sendrecv(border_left, n_border_left, mpi_integer, left, GroupPE &
                   , border_from_right, n_border_from_right, mpi_integer, right, right, server_comm, mpi_status, ierr)
              deallocate(border_left)
              call trace_allocations(0,-4.*float(n_border_left),.false.,'border_left')
              allocate(border_left(n_border_max),stat=ierr)
              call trace_allocations(ierr,4.*float(n_border_max),.false.,'border_left')


! send to right, receive from left
              call mpi_sendrecv(n_border_right, 1, mpi_integer, right, GroupPE &
                   ,n_border_from_left, 1, mpi_integer, left, left, server_comm, mpi_status, ierr)
              allocate(border_from_left(1:n_border_right),stat=ierr)
              call trace_allocations(ierr,4.*float(n_border_right),.false.,'border_from_left')
              border_from_left(:) = border_right(1:n_border_right)
              deallocate (border_right)
              call trace_allocations(0,-4.*float(n_border_right_old),.false.,'border_right')
              allocate (border_right(1:n_border_right),stat=ierr)
              call trace_allocations(ierr,4.*float(n_border_right),.false.,'border_right')
              border_right = border_from_left
              deallocate(border_from_left)
              call trace_allocations(0,-4.*float(n_border_right),.false.,'border_from_left')
              allocate(border_from_left(1:n_border_from_left),stat=ierr)
              call trace_allocations(ierr,4.*float(n_border_from_left),.false.,'border_from_left')
              call mpi_sendrecv(border_right, n_border_right, mpi_integer, right, GroupPE &
                   , border_from_left, n_border_from_left, mpi_integer, left, left, server_comm, mpi_status, ierr)
              deallocate(border_right)
              call trace_allocations(0,-4.*float(n_border_right),.false.,'border_right')
              allocate(border_right(n_border_max),stat=ierr)
              call trace_allocations(ierr,4.*float(n_border_max),.false.,'border_right')

#ifdef FETCH
              write(funit,'("[",a24,"]   +++ Received borders: "9(1x,i9))') fdate(),n_border_from_left,n_border_from_right
              call flush(funit)
#endif

              n_border_left_old=n_border_left
              n_border_right_old=n_border_right
              n_border_left=0
              n_border_right=0

              do ivertex=1,n_border_from_left/5
                 OldID = border_from_left((ivertex-1)*5+1)
                 NewID = border_from_left((ivertex-1)*5+2)
                 i     = border_from_left((ivertex-1)*5+3)
                 j     = border_from_left((ivertex-1)*5+4)
                 kk    = border_from_left((ivertex  )*5  )
                 k     = ipbc(kk - plane_start(GroupPE+1),Lgrid_global(3))
                 if (k<-1) k=k+Lgrid_global(3)
                 if (k>Lgrid(3)+2) k=k-Lgrid_global(3)
                 call update_ID_new(NewID,OldID,i,j,k,.false.)
              enddo

#ifdef FETCH
              write(funit,'("[",a24,"]   +++ updated from left"))') fdate()
              call flush(funit)
#endif

             do ivertex=1,n_border_from_right/5
                 OldID = border_from_right((ivertex-1)*5+1)
                 NewID = border_from_right((ivertex-1)*5+2)
                 i     = border_from_right((ivertex-1)*5+3)
                 j     = border_from_right((ivertex-1)*5+4)
                 kk    = border_from_right((ivertex  )*5  )
                 k     = ipbc(kk - plane_start(GroupPE+1),Lgrid_global(3))
                 if (k<-1) k=k+Lgrid_global(3)
                 if (k>Lgrid(3)+2) k=k-Lgrid_global(3)
                 call update_ID_new(NewID,OldID,i,j,k,.false.)
              enddo

#ifdef FETCH
              write(funit,'("[",a24,"]   +++ updated from right"))') fdate()
              call flush(funit)
#endif

              deallocate (border_from_left)
              call trace_allocations(0,-4.*float(n_border_from_left),.false.,'border_from_left')
              deallocate (border_from_right)
              call trace_allocations(0,-4.*float(n_border_from_right),.false.,'border_from_right')
#ifdef FETCH
              write(funit,'("[",a24,"]   +++ done iteration"))') fdate()
              call flush(funit)
#endif
           else
              transmit_borders=.false.
           endif
        enddo

#ifdef UPDATE
        do local_plane = 0, Lgrid(3)+1
           global_plane = ipbc(local_plane + plane_start(GroupPE+1),Lgrid_global(3))
           if (global_plane<10) then
              write(fname,'("server_GroupID.",i1,".00",i1)')MyPE,global_plane
           else if (global_plane<100) then
              write(fname,'("server_GroupID.",i1,".0",i2)')MyPE,global_plane
           else
              write(fname,'("server_GroupID.",i1,".",i3)')MyPE,global_plane
           endif
           open(23,file=trim(fname),status='unknown')
           do j=1,Lgrid(2)
              do i=1,Lgrid(1)
                 write(23,'(4(1x,i9)))')i,j,global_plane,server_GroupID(i,j,local_plane)
              enddo
           enddo
           close(23)
        enddo
#endif
        deallocate (border_left)
        call trace_allocations(0,-4.*float(n_border_left_old),.false.,'border_left')
        deallocate (border_right)
        call trace_allocations(0,-4.*float(n_border_right_old),.false.,'border_right')

#ifdef FETCH
        write(funit,'("[",a24,"]   +++ done iteration"))') fdate()
        call flush(funit)
#endif


     endif
  endif
#ifdef DEBUG
  close(20)
#endif
  return
end subroutine update_vertices_new


subroutine update_ID_new(NewID,OldID,i,j,k,border)

  use numbers
  use Grid, only : LGrid, server_GroupID
  use fragment_mpi
  implicit none
  integer(kind=singleI),intent(in)::NewID,OldID,i,j,k
  logical,intent(in) :: border
  integer(kind=singleI),external::ipbc

  integer(kind=singleI)::iloop,ix,iy,iz,iix,iiy,count
  logical::change

  if (k>=0 .and. k<=Lgrid(3)+1) server_GroupID(i,j,k)=NewID
  if (border .and. OldID/=NewID) call fix_border(OldID,NewID,i,j,k)

! If the group has changed, propagate the change to all neighbours
  if (NewID > 0 .and. OldID > 0 .and. OldID/=NewID) then
     count=0
     loop: do iloop = 1,maxval(Lgrid)
        change = .false.
        z: do iz=max(0,k-iloop),min(k+iloop,Lgrid(3)+1)
           y: do iiy=j-iloop,j+iloop
              iy = ipbc(iiy,Lgrid(2))
              x: do iix=i-iloop,i+iloop
                 ix = ipbc(iix,Lgrid(1))
                 if (iix/=i-iloop .and. iix/=i+iloop .and. iiy/=j-iloop .and. iiy/=j+iloop .and. iz/=k-iloop &
                      .and. iz/=k+iloop) cycle x
                 if (server_GroupID(ix,iy,iz)==OldID) then
                    change=.true.
                    server_GroupID(ix,iy,iz)=NewID
                    if (border) then
                       call fix_border(OldID,NewID,ix,iy,iz)
                    else
                       if (abs(k-iz)>Lgrid(3)/2) call fix_border(OldID,NewID,ix,iy,iz)
                    endif
                 endif
              enddo x
           enddo y
        enddo z
        if (.not.change) count=count+1
        if (count==3) exit loop
     enddo loop
  endif

  return
end subroutine update_ID_new


subroutine fix_border(OldID,NewID,ix,iy,iz)

  use numbers
  use borders
  use fragment_mpi
  use Grid, only : Lgrid, plane_start, Lgrid_global
  implicit none
  integer(kind=singleI),intent(in)::OldID,NewID,ix,iy,iz
  integer, external:: ipbc

  if (iz<=2) then
     if (n_border_left+5>n_border_max) call abort('TOO MANY BORDER PARTICLES!')
     n_border_left = n_border_left + 1 ; border_left(n_border_left)=OldID
     n_border_left = n_border_left + 1 ; border_left(n_border_left)=NewID
     n_border_left = n_border_left + 1 ; border_left(n_border_left)=ix
     n_border_left = n_border_left + 1 ; border_left(n_border_left)=iy
     n_border_left = n_border_left + 1 ; border_left(n_border_left)=ipbc(iz+plane_start(GroupPE+1),Lgrid_global(3))
  endif
  if (iz>=Lgrid(3)-1) then
     if (n_border_right+5>n_border_max) call abort('TOO MANY BORDER PARTICLES!')
     n_border_right = n_border_right + 1 ; border_right(n_border_right)=OldID
     n_border_right = n_border_right + 1 ; border_right(n_border_right)=NewID
     n_border_right = n_border_right + 1 ; border_right(n_border_right)=ix
     n_border_right = n_border_right + 1 ; border_right(n_border_right)=iy
     n_border_right = n_border_right + 1 ; border_right(n_border_right)=ipbc(iz+plane_start(GroupPE+1),Lgrid_global(3))
  endif

  return
end subroutine fix_border


function locate_vertex(this_vertex, n_total_vertex, VertexName)
  use numbers
  implicit none
  integer(kind=singleI), intent(in) :: n_total_vertex
  integer(kind=doubleI), intent(in) :: this_vertex, VertexName(n_total_vertex)
  integer(kind=doubleI) :: name, lname, hname
  integer(kind=singleI) :: locate_vertex
  integer group, ilow, ihigh

! Given VertexName(1:n_total_vertex), find where "this_vertex" fits
! Note: it does not have to be in this list
! bisection

  ilow  = 1
  lname = VertexName(ilow)
  if(this_vertex == lname) then
     locate_vertex = ilow
     return
  endif
  ihigh = n_total_vertex
  hname = VertexName(ihigh)
  if(this_vertex == hname)then
     locate_vertex = ihigh
     return
  end if
  locate_vertex = -1 ! means vertex not in list

! fool proof
  if(ihigh == ilow)then
     if(lname /= this_vertex)then
        return
     else
        locate_vertex = ilow
        return
     endif
  endif

  group = 0.5*(ilow+ihigh)
  name  = VertexName(group)
  do while (ilow /= ihigh-1)
     if(this_vertex > name)then
        ilow = group
     else if (this_vertex < name) then
        ihigh = group
     else
        locate_vertex = group
        return
     endif
     group = 0.5*(ilow+ihigh)
     name  = VertexName(group)
  enddo
  hname = VertexName(ihigh)
  if(this_vertex /= hname)then
     return
  else
     locate_vertex = ihigh
     return
  endif

  return
end function locate_vertex


function find_index(n,value,index,data)

! determine find_index such that 
! data(index(find_index)) > value >= data(index(find_index+1))

  use numbers
  implicit none
  integer(kind=singleI),intent(in) :: n, index(n)
  real(kind=singleR), intent(in)   :: value, data(n)
  integer(kind=singleI) :: find_index

  find_index = 0
  if(value < 0)then
     return
  else
     do while(data(index(find_index+1)) > value .and. find_index < n)
        find_index = find_index + 1
     enddo
  endif
  return
end function find_index


real(kind=singleR) function next_fmax(iout,nodata)

! computes the next fmax-value at the end of the data block

  use numbers
  use fragment_mpi
  use grid
  use scratch_io
  use runtime_parameters
  use GridData, only : MaxVertexOnMaster
  use fast_fmax_module
  implicit none
  integer(kind=singleI),intent(in) :: iout
  logical,intent(out) :: nodata

! local variables
  real(kind=singleR)    :: tmp, next_fmax1
  real(kind=singleR),save :: largest=0.
  integer(kind=singleI) :: tag
  integer(kind=singleI) :: n_data, ierr, nmax
  integer(kind=singleI) :: iplane_global, iplane_local, indx
  integer(kind=singleI) :: index_current_fmax, index_next_fmax, find_index, iserver
  logical,save :: firsttime=.true.

  next_fmax = 0.

! new, fast version that uses the cumulative distributions for each server
  if (MyPE==0) then
     if (fmax_out(iout)>largest_fmax) then
! checks whether there are fmax vertices to import
        nodata    = .true.
        next_fmax = largest_fmax
     else
        nodata=.false.
! compute the next fmax value for each server
        do iserver=1,Nserver
! find the number of vertices for current fmax value by interpolation
           fbin = (Flargest-log(current_fmax))/binsize
           fbin = max(fbin,0.)
           ibin = int(fbin)+1
           fbin = fbin - (ibin-1.)
           if (ibin==1) then
              n_current_fmax=0
           else
              n_current_fmax = int(float(fast_fmax(iserver,ibin-1))*(1.-fbin)+float(fast_fmax(iserver,ibin))*fbin)
           endif
! estimate the next fmax
           n_next_fmax = n_current_fmax + int(float(MaxVertexOnMaster)*0.95/float(Nserver))
! are there enough vertices?
           if (n_next_fmax>fast_fmax(iserver,nbin)) then
              next_fmax1 = exp(Fsmallest)
           else
! find next fmax for the server by interpolating the cumulative distribution
              do while (fast_fmax(iserver,ibin)<n_next_fmax)
                 ibin=ibin+1
              enddo
              rbin = float(n_next_fmax-fast_fmax(iserver,ibin-1))/float(fast_fmax(iserver,ibin)-fast_fmax(iserver,ibin-1))
              next_fmax1 = Flargest-(ibin-1.+rbin)*binsize
              next_fmax1 = exp(next_fmax1)
              next_fmax  = max(next_fmax,next_fmax1)
           endif
        enddo
! next fmax is the largest value for all the servers
        next_fmax = max(next_fmax,fmax_out(iout))
     endif
  endif

!!!! old version
!!!
!!!  nmax      = int(MaxVertexOnMaster / Lgrid_global(3))  ! **!!** Sbagliato: NServer)
!!!
!!!  if(my_type == master)then
!!!     if(myPE == 0)then
!!!        do tag=0, NServer-1
!!!! receive value from each server node
!!!           if (firsttime) then
!!!              call mpi_recv(tmp, 1, mpi_real, NMaster+tag, tag, mpi_comm_world, mpi_status, ierr)
!!!              largest=max(largest,tmp)
!!!           endif
!!!           call mpi_recv(tmp, 1, mpi_real, NMaster+tag, tag, mpi_comm_world, mpi_status, ierr)
!!!           next_fmax = max(next_fmax, tmp)
!!!        enddo
!!!
!!!! values of fmax need to be smaller than fmax_out(iout)
!!!        next_fmax = max(fmax_out(iout),next_fmax)
!!!
!!!        nodata = .false.
!!!        if (largest < next_fmax) nodata = .true.
!!!
!!!        firsttime=.false.
!!!     endif
!!!  else
!!!
!!!! each processor determines next_fmax such that there are approximately nmax particles that collapse
!!!! between current_fmax and next_fmax
!!!
!!!     n_data = Lgrid(1)*Lgrid(2)
!!!     allocate(plane1D(n_data))
!!!     allocate(indexx(n_data))
!!!     call trace_allocations(0,8.*float(n_data),.false.,'plane1D, indexx')
!!!     tmp     = -10.
!!!     largest = 1.e12
!!!     do iplane_local=1, nplanes(GroupPE+1)
!!!
!!!        iplane_global = iplane_local + plane_start(GroupPE+1)
!!!        call read_scratch (iplane_global + fmax_offset, n_data, plane1D)
!!!        call read_scratch (iplane_global + index_offset, n_data, indexx)
!!!
!!!        index_current_fmax = find_index(n_data, current_fmax, indexx, plane1D)      
!!!        index_next_fmax    = min(index_current_fmax + nmax , n_data)
!!!!        tmp = plane1D(indexx(index_next_fmax))
!!!        tmp = max(tmp,plane1D(indexx(index_next_fmax)))
!!!        largest = min(largest,plane1D(indexx(index_current_fmax+1)))
!!!
!!!     enddo
!!!
!!!     next_fmax = tmp
!!!
!!!     Deallocate(indexx)
!!!     deallocate(plane1d)
!!!     call trace_allocations(0,-8.*float(n_data),.false.,'plane1D, indexx')
!!!
!!!     if (firsttime) then
!!!        call mpi_send(largest, 1, mpi_real, 0, GroupPE, mpi_comm_world, ierr)
!!!        firsttime = .false.
!!!     endif
!!!
!!!     call mpi_send(next_fmax, 1, mpi_real, 0, GroupPE, mpi_comm_world, ierr)
!!!  endif



  call mpi_bcast(next_fmax, 1, mpi_real, 0, mpi_comm_world, ierr)
  call mpi_bcast(nodata, 1, mpi_logical, 0, mpi_comm_world, ierr)

  return
end function next_fmax


subroutine extract_filaments()

! sort filament data and removes doubles

  use numbers
  use FilamentData
  use fragment_mpi
  use grid, only : Lgrid_global
  implicit none
  integer(kind=singleI), allocatable :: indexx(:)
  integer(kind=doubleI), allocatable :: filament_id(:)
  type(FilamentType), allocatable :: FilamentCopy(:)
  integer(kind=singleI) :: i, nf, n_filaments_new, ierr, nshift1, nshift2
  integer(kind=doubleI) :: p_index, half
  character(len=24), external:: fdate

  if(MyPE == 0 .and. n_filaments > 0)then

! sort filaments based on global ID
     allocate(indexx(n_filaments),stat=ierr)
     call trace_allocations(ierr,4*float(n_filaments),.false.,'indexx')
     allocate(filament_id(n_filaments),stat=ierr)
     call trace_allocations(ierr,8*float(n_filaments),.false.,'filament_id')
     allocate(filamentcopy(n_filaments),stat=ierr)
     call trace_allocations(ierr,FilamentTypeSize*float(n_filaments),.false.,'filamentcopy')

     half = Lgrid_global(1)*Lgrid_global(2)
     half = half*Lgrid_global(3)/2
     nshift1=0
     do while (MasterFilaments(nshift1+1)%GlobalID>half)
        nshift1=nshift1+1
     enddo
     nshift2=0
     do while (MasterFilaments(n_filaments-nshift2)%GlobalID<half)
        nshift2=nshift2+1
     enddo
     
     FilamentCopy(1:nshift2) = MasterFilaments(n_filaments-nshift2+1:n_filaments)
     FilamentCopy(nshift2+1:n_filaments-nshift1) = MasterFilaments(nshift1+1:n_filaments-nshift2)
     FilamentCopy(n_filaments-nshift1+1:n_filaments) = MasterFilaments(1:nshift1)

     Filament_id(1:nshift2) = MasterFilaments(n_filaments-nshift2+1:n_filaments)%GlobalID
     Filament_id(nshift2+1:n_filaments-nshift1) = MasterFilaments(nshift1+1:n_filaments-nshift2)%GlobalID
     Filament_id(n_filaments-nshift1+1:n_filaments) = MasterFilaments(1:nshift1)%GlobalID     

!     filament_id(1:n_filaments) = MasterFilaments(1:n_filaments)%GlobalID
!     filamentcopy(1:n_filaments)= MasterFilaments(1:n_filaments)
     deallocate(MasterFilaments)
     call trace_allocations(0,-float(FilamentTypeSize*n_filaments),.false.,'MasterFilaments')

!     write(*,'("[",a24,"]    +++ allocations")') fdate()
     call index_sort_i8(n_filaments, filament_id, indexx)
!     write(*,'("[",a24,"]    +++ sorting done")') fdate()

     p_index = 0
     n_filaments_new = 0
     do i=1, n_filaments
        if (FilamentCopy(indexx(i))%GlobalID /= p_index)then
           n_filaments_new = n_filaments_new + 1
           p_index = FilamentCopy(indexx(i))%GlobalID
        endif
     enddo

!     write(*,'("[",a24,"]    +++ cleaning done")') fdate()

     allocate(MasterFilaments(n_filaments_new),stat=ierr)
     call trace_allocations(ierr,float(FilamentTypeSize*n_filaments_new),.false.,'MasterFilaments')

!     write(*,'("[",a24,"]    +++ reallocation done")') fdate()

     nf = 0
     p_index = 0
     do i=1, n_filaments
        if(FilamentCopy(indexx(i))%GlobalID /= p_index)then
           nf = nf + 1
           MasterFilaments(nf) = FilamentCopy(indexx(i))
           p_index             = FilamentCopy(indexx(i))%GlobalID
        endif
     enddo

!     write(*,'("[",a24,"]    +++ copy done")') fdate()

     deallocate(indexx)
     call trace_allocations(0,-4*float(n_filaments),.false.,'indexx')
     deallocate(filament_id)
     call trace_allocations(0,-8*float(n_filaments),.false.,'filament_id')
     deallocate(FilamentCopy)
     call trace_allocations(0,-float(FilamentTypeSize*n_filaments),.false.,'filamentcopy')
     n_filaments = n_filaments_new

  endif

  return
end subroutine extract_filaments


subroutine pointers_to_neighbours()

! Constructs the pointers from vertices to neighbour vertices and filaments,
! and the pointers from filaments to vertices
! NB: This is only called by Processor 0

  use numbers
  use GridData !, only : GridOffset, GridVertex, n_total_vertex, GroupsPerVertex &
!       , InverseNeighbour, VertexPointer, 
  use FilamentData!, only : n_filaments, is_filament, MasterFilaments, FilamentPointer
  use Grid, only : Lgrid_global
  implicit none
  integer(kind=singleI),external::periodic, locate_vertex

  integer(kind=singleI), allocatable :: VertexPtr(:)
  integer(kind=doubleI), allocatable :: VertexName(:), VertexCopy(:)
  integer(kind=doubleI), allocatable :: FilamentName(:)
  integer(kind=doubleI) :: GlobalID 
  integer(kind=singleI) :: nvertex, ivertex, i, j, k, my_vertex, ineigh, ii, jj, kk, index, igroup, m, ifilament, ierr

! make set of pointers for each current GridVertex to neighbouring vertices (if any)
  allocate(VertexPointer(n_total_vertex),stat=ierr)
  call trace_allocations(ierr,float(VertexPointerTypeSize*n_total_vertex),.false.,'VertexPointer')
  allocate(VertexCopy(n_total_vertex),stat=ierr)
  call trace_allocations(ierr,8.*float(n_total_vertex),.false.,'VertexCopy')
  allocate(VertexName(n_total_vertex),stat=ierr)
  call trace_allocations(ierr,8.*float(n_total_vertex),.false.,'VertexName')
  allocate(VertexPtr(n_total_vertex),stat=ierr)
  call trace_allocations(ierr,4.*float(n_total_vertex),.false.,'VertexPtr')

! initialize VertexPointer and copy global IDs of all vertices to VertexName
  do ivertex = 1, n_total_vertex
     i = GridVertex(ivertex)%i
     j = GridVertex(ivertex)%j
     k = GridVertex(ivertex)%k
     GlobalID = i + (j-1)*LGrid_global(1) + (k-1)*LGrid_global(1)*LGrid_global(2)
! test
!     call get_coordinates(GlobalID,ii,jj,kk)
!     print *,i,ii,j,jj,k,kk
     VertexPointer(ivertex)%GlobalID = GlobalID
     VertexName(ivertex) = GlobalID
  enddo
! sort vertices by global ID
  call index_sort_i8(n_total_vertex,VertexName,VertexPtr)
  VertexCopy(:) = VertexName
! make VertexName the sorted list of global vertex indices
! and VertexPtr a pointer to where in GridVertex (or VertexPointer) the corresponding data live 
  do i=1,n_total_vertex
     VertexName(i) = VertexCopy(VertexPtr(i))
  enddo
  deallocate(VertexCopy)
  call trace_allocations(0,-8.*float(n_total_vertex),.false.,'VertexCopy')

! find where in the list each neighbour of a vertex lives, and point to that
  do ivertex = 1, n_total_vertex
     VertexPointer(ivertex)%Neighbour(:) = 0
     do ineigh = 1, GroupsPerVertex
! find neighbouring vertex (if it exists)
        ii = periodic(GridVertex(ivertex)%i + GridOffset(ineigh)%i_offset, LGrid_global(1))
        jj = periodic(GridVertex(ivertex)%j + GridOffset(ineigh)%j_offset, LGrid_global(2))
        kk = periodic(GridVertex(ivertex)%k + GridOffset(ineigh)%k_offset, LGrid_global(3))
        GlobalID = ii + (jj-1)*LGrid_global(1)+(kk-1)*LGrid_global(1)*LGrid_global(2)
        nvertex = locate_vertex(GlobalID, n_total_vertex, VertexName)
        if (nvertex > 0) then
           VertexPointer(ivertex)%Neighbour(ineigh) = VertexPtr(nvertex)
        endif
     enddo
  enddo
  deallocate(VertexName)
  call trace_allocations(0,-8.*float(n_total_vertex),.false.,'VertexName')
  deallocate(VertexPtr)
  call trace_allocations(0,-4.*float(n_total_vertex),.false.,'VertexPtr')

! now make pointers to Filaments (if any)
  allocate(FilamentName(n_filaments),stat=ierr)
  call trace_allocations(ierr,8.*float(n_filaments),.false.,'FilamentName')
  do i=1, n_filaments
     FilamentName(i) = MasterFilaments(i)%GlobalID
  enddo

  allocate (FilamentPointer(n_filaments),stat=ierr)
  call trace_allocations(ierr,float(FilamentPointerTypeSize*n_filaments),.false.,'FilamentPointer')
  do ifilament=1,n_filaments
     FilamentPointer(ifilament)%NeighbourVertex(:) = 0
     FilamentPointer(ifilament)%position(:)        = 0
  enddo

  do index = 1, n_total_vertex
     ivertex    = index
     call get_coordinates(VertexPointer(index)%GlobalID,i,j,k)
     VertexPointer(index)%Filament(:) = 0

! check we have the right one

     do ineigh = 1, GroupsPerVertex
        igroup   = GridVertex(ivertex)%GroupID(ineigh) !GroupID(ineigh)
        if (igroup == Is_Filament) then ! denotes filament
           ii = periodic(i + GridOffset(ineigh)%i_offset,LGrid_global(1))
           jj = periodic(j + GridOffset(ineigh)%j_offset,LGrid_global(2))
           kk = periodic(k + GridOffset(ineigh)%k_offset,LGrid_global(3))
           GlobalID = ii + (jj-1)*LGrid_global(1) + (kk-1)*LGrid_global(1)*LGrid_global(2)
           ifilament  = locate_vertex(GlobalID, n_filaments, FilamentName)
           if(ifilament < 0)then
              write (*,*) ' error: filament not found '
              write (*,*) ' vertex = ',i,j,k
              write (*,*) ' neighbour = ',ii,jj,kk,ineigh
              write (*,*) ' GlobalID = ',GlobalID
              call abort_code(' Filament not found ')
           endif
           VertexPointer(index)%Filament(ineigh) = ifilament
           m=1
           do while (m<GroupsPerVertex .and. FilamentPointer(ifilament)%NeighbourVertex(m)>0)
              m=m+1
           enddo
           if (FilamentPointer(ifilament)%NeighbourVertex(m)>0) then
              write(*,*) 'ERROR! FilamentPointer stack full',m
              write(*,'(9(1x,i12))')MasterFilaments(ifilament)%GlobalID
              write(*,'(9(1x,i9))')FilamentPointer(ifilament)%NeighbourVertex
              write(*,'(9(1x,i9))')FilamentPointer(ifilament)%position
              call abort_code('done')
           endif
           FilamentPointer(ifilament)%NeighbourVertex(m)=ivertex
           FilamentPointer(ifilament)%position(m)=ineigh
        endif
     enddo
  enddo
  deallocate(FilamentName)
  call trace_allocations(0,-8.*float(n_filaments),.false.,'FilamentName')

  return
end subroutine pointers_to_neighbours


subroutine get_coordinates(ID,i,j,k)

  use numbers
  use grid, only : Lgrid_global
  implicit none
  integer(kind=doubleI),intent(in)::ID
  integer(kind=singleI),intent(out)::i,j,k

  integer(kind=singleI)::kk
  integer(kind=singleI),save::Lgrid12
  logical::firsttime=.true.

  if (firsttime) then
     Lgrid12=Lgrid_global(1)*Lgrid_global(2)
     firsttime=.false.
  endif

  k=1+(ID-1)/Lgrid12
  kk=ID-1-(k-1)*Lgrid12
  j=1+kk/Lgrid_global(1)
  i=1+(kk-(j-1)*Lgrid_global(1))

  return
end subroutine get_coordinates


subroutine write_GroupIDs()

  use numbers
  use fragment_mpi
  use grid
  use runtime_parameters, only: string_length
  implicit none
  integer(kind=singleI)::i
  character(len=2)::label
  character(len=string_length)::filename

  if (my_type == server) then

     write(label,'(i2)')MyPE
     if (MyPE<10) label(1:1)='0'
     filename='server_GroupIDs.'//label
#ifdef ASCII_OUT
     open(1,file=filename,status='unknown')
     write(1,*)Lgrid(1),Lgrid(2),Lgrid(3)
     do i=1,Lgrid(3)
        write(1,'(i9)')i
        write(1,'(10(1x,i9))')server_GroupID(:,:,i)
     enddo
     close(1)
#endif
#ifdef UNFORM_OUT
     open(1,file=filename,status='unknown',form='unformatted')
     write(1)Lgrid(1),Lgrid(2),Lgrid(3)
     do i=1,Lgrid(3)
        write(1)i
        write(1)server_GroupID(:,:,i)
     enddo
     close(1)
#endif

  endif
  return
end subroutine write_Groupids


function locate_element(element, nlist, list, indices)
  use numbers
  implicit none
  integer(kind=singleI) :: locate_element
  integer(kind=singleI), intent(in) :: nlist, element, list(nlist), indices(nlist)
  integer(kind=singleI) :: name, lname, hname, group, ilow, ihigh

! Given list(1:nlist) and its ordered indices, indices(1:nlist), find where "element" is
! Note: it does not have to be in this list
! bisection

  ilow  = 1
  lname = list(indices(ilow))
  if (element == lname) then
     locate_element = ilow
     return
  endif
  ihigh = nlist
  hname = list(indices(ihigh))
  if (element == hname) then
     locate_element = ihigh
     return
  end if
  locate_element = -1 ! means vertex not in list

! fool proof
  if (ihigh == ilow) then
     if (lname /= element) then
        return
     else
        locate_element = ilow
        return
     endif
  endif

  group = 0.5*(ilow+ihigh)
  name  = list(indices(group))
  do while (ilow /= ihigh-1)
     if (element > name) then
        ilow = group
     else if (element < name) then
        ihigh = group
     else
        locate_element = group
        return
     endif
     group = 0.5*(ilow+ihigh)
     name  = list(indices(group))
  enddo
  hname = list(indices(ihigh))
  if (element /= hname) then
     return
  else
     locate_element = ihigh
     return
  endif

  return
end function locate_element
