!*****************************************************************
!*                       PINOCCHI0  V2.2                         *
!*  (PINpointing Orbit-Crossing Collapsed HIerarchical Objects)  *
!*****************************************************************
!
!This code was written by
!Pierluigi Monaco, Tom Theuns, Giuliano Taffoni
!Copyright (C) 2008
!
!web page: http//adlibitum.oats.inaf.it/monaco/Homepage/Pinocchio/
!
!This program is free software; you can redistribute it and/or modify
!it under the terms of the GNU General Public License as published by
!the Free Software Foundation; either version 2 of the License, or
!(at your option) any later version.
!
!This program is distributed in the hope that it will be useful,
!but WITHOUT ANY WARRANTY; without even the implied warranty of
!MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!GNU General Public License for more details.
!
!You should have received a copy of the GNU General Public License
!along with this program; if not, write to the Free Software
!Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

subroutine mpi_create_gridvertex()
  use numbers
  use GridData
  use fragment_mpi
  implicit none
  integer  :: ierr
!  integer(kind=singleI), parameter :: ntypes = 3
  integer(kind=singleI), parameter :: ntypes = 2
  integer(kind=singleI)  :: mpi_types(ntypes), blockcounts(ntypes), offsets(ntypes), sizes(ntypes)
  integer(kind=singleI)  :: itype

!  mpi_types   = (/mpi_real,mpi_integer,mpi_logical/)
!  blockcounts = (/4,4+GroupsPerVertex,4/)
  mpi_types   = (/mpi_real,mpi_integer/)
  blockcounts = (/4,4+GroupsPerVertex/)
  offsets(:)  = 0
  !
  do itype=1,ntypes
     call mpi_type_extent(mpi_types(itype), sizes(itype), ierr)
  enddo
  !
  do itype=2,ntypes
     offsets(itype) = offsets(itype-1)+blockcounts(itype-1)
  enddo
  !     
  call mpi_type_struct(ntypes, blockcounts, offsets, mpi_types, GridVertexMPIType, ierr)
  call mpi_type_commit(GridVertexMPIType, ierr)
  call mpi_type_size(GridVertexMPIType,gridvertexSize,ierr)

  return
end subroutine mpi_create_gridvertex


subroutine mpi_create_filament()
  use numbers
  use FilamentData
  use fragment_mpi
  implicit none
  integer  :: ierr, size_real
  integer(kind=singleI), parameter :: ntypes = 4
  integer(kind=singleI)  :: mpi_types(ntypes), blockcounts(ntypes), offsets(ntypes), sizes(ntypes)
  integer(kind=singleI)  :: itype

!  mpi_types   = (/mpi_integer,mpi_real,mpi_integer,mpi_logical/)
!  blockcounts = (/2,3,5,4/)  ! note declare 4 logicals = 1 byte to avoid alignment errors

  mpi_types   = (/mpi_real,mpi_integer,mpi_integer,mpi_logical/)
  blockcounts = (/3,1,2,4/)
  offsets(:)  = 0
  !
  do itype=1,ntypes
     call mpi_type_extent(mpi_types(itype), sizes(itype), ierr)
  enddo
  !
  do itype=2,ntypes
     offsets(itype) = offsets(itype-1)+blockcounts(itype-1)
  enddo
  !     
  call mpi_type_struct(ntypes, blockcounts, offsets, mpi_types, FilamentDataMPIType, ierr)
  call mpi_type_commit(FilamentDataMPIType, ierr)
  call mpi_type_size(FilamentDataMPIType,FilamentDataMPISize,ierr)
  !
  return
end subroutine mpi_create_filament


subroutine mpi_create_grouptype()
  use numbers
  use GroupData
  use fragment_mpi
  implicit none
  integer  :: ierr, size_real
  integer(kind=singleI), parameter :: ntypes = 4
  integer(kind=singleI)  :: mpi_types(ntypes), blockcounts(ntypes), offsets(ntypes), sizes(ntypes)
  integer(kind=singleI)  :: itype

  mpi_types   = (/mpi_real,mpi_integer,mpi_integer,mpi_logical/)
  blockcounts = (/8,2,2,4/)
  offsets(:)  = 0
  !
  do itype=1,ntypes
     call mpi_type_extent(mpi_types(itype), sizes(itype), ierr)
  enddo
  !
  do itype=2,ntypes
     offsets(itype) = offsets(itype-1)+blockcounts(itype-1)
  enddo
  !     
  call mpi_type_struct(ntypes, blockcounts, offsets, mpi_types, GroupDataMPIType, ierr)
  call mpi_type_commit(GroupDataMPIType, ierr)
  call mpi_type_size(GroupDataMPIType,GroupDataMPISize,ierr)
  !
  return
end subroutine mpi_create_grouptype


subroutine message(messg)
  use fragment_mpi
  implicit none
  !     
  character(*) :: messg
  
  if(MyPE .eq. 0) then 
     write (*,*) messg
     call flush(6)
  endif
end subroutine message
      

subroutine abort_code(messg)
  use fragment_mpi
  implicit none
  integer      :: ierr
  character(*) :: messg
  write (*,'(a)') messg
  write (*,*) ' mpi_abort called by processor: ',MyPE
  call mpi_abort(mpi_comm_world,ierr)
  call mpi_finalize(' .... stop ')
  stop
end subroutine abort_code


subroutine finish(messg)
  use fragment_mpi
  implicit none
  character(*) :: messg
  integer      :: ierr
  call message(messg)
  call mpi_finalize(ierr)
  stop
end subroutine finish


