!*****************************************************************
!*                       PINOCCHI0  V2.2                         *
!*  (PINpointing Orbit-Crossing Collapsed HIerarchical Objects)  *
!*****************************************************************
!
!This code was written by
!Pierluigi Monaco, Tom Theuns, Giuliano Taffoni
!Copyright (C) 2008
!
!web page: http//adlibitum.oats.inaf.it/monaco/Homepage/Pinocchio/
!
!This program is free software; you can redistribute it and/or modify
!it under the terms of the GNU General Public License as published by
!the Free Software Foundation; either version 2 of the License, or
!(at your option) any later version.
!
!This program is distributed in the hope that it will be useful,
!but WITHOUT ANY WARRANTY; without even the implied warranty of
!MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!GNU General Public License for more details.
!
!You should have received a copy of the GNU General Public License
!along with this program; if not, write to the Free Software
!Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


program fragment_parallel

!

  use numbers
  use fragment_mpi
  use grid
  use runtime_parameters

  use FilamentData, only: MasterFilaments, n_filaments, is_filament
  use GridData, only: n_total_vertex, GridVertex
  use GroupData, only: n_master_groups, MasterGroups

  implicit none
  integer(kind=singleI):: ierr, num_to_recv, this_pe
  real(kind=singleR),external :: next_fmax,inverse_grow
  character(len=24), external:: fdate

! local variables
  real(kind=singleR) :: start, step, stop, nfmax, zz, cputime, steptime
  integer(kind=singleI) :: iout, counter
  logical :: nodata
  character(len=120)::info

! initialize MPI
  call mpi_init(ierr)
  call mpi_comm_rank(mpi_comm_world, MyPE, ierr)
  call mpi_comm_size(mpi_comm_world, NumPEs, ierr)

#if !defined(ASCII_OUT) && !defined (UNFORM_OUT) && !defined (HDF5_OUT)
  call finish('please define output in preprocessing')
#endif

  cputime = mpi_wtime()

  if (MyPE==0) write(*,'("[",a24,"] Starting run")') fdate()

! initialize run paramaters
  call init_run()
  
! divide processors and domain
! PE = 0: master processor that does fragment calculations
! PE = first_inode,NMaster-1: processors store group data
! PE = NMaster - NumPEs-1: processors that store Grid data
! PEs 0-NMaster-1 are part of master_comm
! PEs NMaster - NumPEs-1 are part of server_comm
  call divide_domain()

! set off_sets for writing temporary variables
  call set_offset()

! load and distribute the data. Allocate Grid data
  
  if (MyPE==0) write(*,'("[",a24,"] ***loading data")') fdate()
  call load_data()

  counter = 1
  iout=1

#ifdef RECOVER
  call recover_run(current_fmax,counter)
  if (MyPE==0) then
     do while (fmax_out(iout)>=current_fmax)
        write(*,*) 'SKIPPING OUTPUT AT z = ',z_out(iout)
        iout=iout+1
     enddo
  endif
  call mpi_bcast(iout,1,mpi_integer,0,mpi_comm_world,ierr)
#else
  open(merger_unit,file='list_of_mergers.dat',status='unknown',form='unformatted')
#endif


  steptime = mpi_wtime()
! evolution loop
  do while (current_fmax > final_fmax)
! next value of fmax to load vertices up to

     if (MyPE==0) write(*,'("[",a24,"] ***calling next_fmax")') fdate()
     nfmax = next_fmax(iout,nodata)

     do while (nodata)
       iout=iout+1
       nfmax = next_fmax(iout,nodata)
    enddo

     if(MyPE == 0)then
        zz=inverse_grow(1./nfmax)
        write (*,'(" *********************** ")')
        write (*,'(" step number ",i4)')counter
        write (*,'(" current fmax = ",f8.4)')min(current_fmax,999.)
        write (*,'(" next fmax    = ",f8.4)')nfmax
        write (*,'(" redshift     = ",f8.4)')zz
        write (*,'(" *********************** ")')
     endif

!     call mpi_barrier(mpi_comm_world, ierr)
     if (MyPE==0) write(*,'("[",a24,"] ***calling fetch_next_grid_data")') fdate()
! master receives values up to nfmax from all servers
     call fetch_next_grid_data(current_fmax, nfmax)

!     call mpi_barrier(mpi_comm_world, ierr)
     if (MyPE==0) write(*,'("[",a24,"] ***calling extract_filaments")') fdate()
! sort filament data by global id and remove duplicate filament data
     call extract_filaments()

!     call mpi_barrier(mpi_comm_world, ierr)
     if (MyPE==0) write(*,'("[",a24,"] ***calling extract_groups")') fdate()
! master makes list of all groups for which we need to collect data
     call extract_groups()

!     call mpi_barrier(mpi_comm_world, ierr)
     if (MyPE==0) write(*,'("[",a24,"] ***calling fetch_next_group_data")') fdate()
! master processor needs to get data for all received groups from group servers
     call fetch_next_group_data()

!     call mpi_barrier(mpi_comm_world, ierr)
     if (MyPE==0) write(*,'("[",a24,"] ***calling fragment")') fdate()
! run fragement on current data
     call fragment()

!     call mpi_barrier(mpi_comm_world, ierr)
     if (MyPE==0) write(*,'("[",a24,"] ***calling update_groups")') fdate()
! update group data on GroupServers
     call update_groups()

!     call mpi_barrier(mpi_comm_world, ierr)
     if (MyPE==0) write(*,'("[",a24,"] ***calling update_vertices")') fdate()
! update Group IDs on vertices
     call update_vertices_new()

!     call mpi_barrier(mpi_comm_world,ierr)
     if (MyPE==0) write(*,'("[",a24,"] ***calling output")') fdate()
! output?
     call output (iout,nfmax)

!     call mpi_barrier(mpi_comm_world,ierr)

     current_fmax = nfmax

#ifdef RECOVER
     if (MyPE==0) write(*,'("[",a24,"] ***calling write_recover_files")') fdate()
     call write_recover_files(current_fmax,counter)
#endif

#ifdef UPDATE
     if (MyPE==0) then
        print *,'Step ',counter
        read *
     endif
#endif

     call trace_allocations(0,0.,.true.,'')

     call mpi_barrier(mpi_comm_world, ierr)

     steptime = mpi_wtime() - steptime
     if (MyPE==0) then
        write(*,'("[",a24,"] ***end of step ",i4)') fdate(),counter
        write(*,'("cputime required = ",e12.4,2x,i4)')steptime,counter
     endif
     counter=counter+1
     steptime = mpi_wtime()
  enddo

  if (MyPE==0) close(merger_unit)

! write the mergers.info file
  call write_mergers_info()

#ifdef WRITE_GROUPIDS
! write the GroupID matrices
  if (MyPE==0) write(*,'("[",a24,"] ***calling write_GroupIDs")') fdate()
  call write_GroupIDs()
#endif

! finalize mpi and stop
  cputime = mpi_wtime() - cputime
  write(info,'(''['',a24,''] Finishing, total cputime = '',e12.4)') fdate(),cputime
  call message(trim(info))
  call finish('')

end program fragment_parallel


subroutine fragment()

  use numbers
  use Fragment_mpi
  use GridData, only : GridOffset, GridVertex, n_total_vertex, GroupsPerVertex, InverseNeighbour&
       , VertexPointer, VertexPointerTypeSize
  use GroupData, only : GroupDataType, MasterGroups, n_master_groups, n_allocated_groups, NewGroupName, MaxGroupsOnServer &
       , first_inode
  use FilamentData, only : n_filaments, MasterFilaments, Is_Filament, FilamentPointer, FilamentPointerTypeSize
  use Grid, only : Lgrid_global
  implicit none
  integer(kind=singleI),external :: locate_group
  real(kind=singleR), allocatable, dimension(:) :: fmax_vector, mass_vector
  integer(kind=singleI), allocatable, dimension(:) :: index_fmax, GroupPointer, index_mass
  integer(kind=singleI) :: vertex, igroup
  logical,dimension(GroupsPerVertex,GroupsPerVertex)::merge
  character(len=24),external :: fdate

! scalar properties of grid point
  integer(kind=singleI) :: GroupID(GroupsPerVertex)
  integer(kind=singleI),dimension(GroupsPerVertex) :: GroupID_cleaned,GroupID_cleaned2
  integer(kind=singleI) :: p_group, c_group, ngroups, ig
  integer(kind=singleI) :: ineigh, ivertex, nvertex, ineigh2, nfilament
  integer(kind=singleI) :: i,j,k,ii,jj,kk,periodic,new_group_id,index,ierr, ig1, ig2, ig3, large, small
  integer(kind=singleI) :: GroupResult,itype, neigh_groups, neigh_filaments, neighbour
  integer(kind=singleI) :: nmerge, group
  real(kind=singleR) :: fmax, velx, vely, velz, distance, radius, best_ratio, ratio, xpos,ypos,zpos
  type(GroupDataType) :: NewGroup
  logical :: merge_flag
  integer(kind=doubleI) :: Peak

  character(len=4),dimension(4)::fate = (/ 'peak', 'accr', 'fila', 'acfi' /)
  integer::ifate,pippo,pippo2
  character(len=*),parameter::formato='(3(1x,i6),1x,f12.8,3(1x,f10.4),1x,a4,2(1x,i8))'


! This routine mimics what the orginal fragment does. The main loop is labeled fmax_loop
! The data structures at that point are:
! GridVertex(1:n_total_vertex),    which is sorted such that the highest fmax value comes first
! MasterGroups(1:n_master_groups)  group data for these groups

! The group pointers in GridVertex%GroupID are no longer the name of the group, but point to the corresponding group in array MasterGroups 
! Therefore the name of the group of the first neighbour of  GridVertex(1) is thus MasterGroups(GridVertex(1)%GroupID(1))%ID

! VertexPointer(1:n_total_vertex)
! This structure allows access to the neighbouring vertex points of a given vertex. That is for nptr = VertexPointer(1)%Neighbour(1),
! nptr < 1: that neighbouring vertex is not in the GridVertex list
! nptr >=1: that neighbouring vertex is GridVertex(nptr)
! Use this structure to determine whether a neighbouring vertex of the current vertex belongs to a group or not

! *important* : use is_filament to indicate filament group

  if(MyPE == 0)then ! only master processes

! sort current fmax values
     write(*,'("[",a24,"]   +++ sorting vertices")') fdate()
     allocate(fmax_vector(n_total_vertex),stat=ierr)
     call trace_allocations(ierr,4.*float(n_total_vertex),.false.,'fmax_vector')
     allocate(index_fmax(n_total_vertex),stat=ierr)
     call trace_allocations(ierr,4.*float(n_total_vertex),.false.,'index_fmax')
     do i=1, n_total_vertex
        fmax_vector(i) = -GridVertex(i)%fmax ! want in decreasing order
     enddo
     call index_sort_r4(n_total_vertex, fmax_vector, index_fmax)
     deallocate(fmax_vector)
     call trace_allocations(0,-4.*float(n_total_vertex),.false.,'fmax_vector')


! loop over GroupIDs
! GroupID = is_filament denotes a filament. Find the vertex in FilamentName array, and have VertexPointer%Filament point to it
! Filament()%alive = .false. if filament has previously been accreted onto a group
! GroupID > 0   denotes a group. Find the group and have GroupID point to it.

! NB: changes the GroupID from global to local IDs
     write(*,'("[",a24,"]   +++ locating groups")') fdate()
     do ivertex = 1, n_total_vertex
        GroupID(:) = GridVertex(ivertex)%GroupID(:)
        do ineigh = 1, GroupsPerVertex
           igroup   = GroupID(ineigh)
           if(igroup > 0) then ! denotes group
              GroupID(ineigh) = locate_group(igroup)
           endif
        enddo
        GridVertex(ivertex)%GroupID(:) = GroupID(:)
     enddo


! constructs the pointers to the neighbouring vertices
     write(*,'("[",a24,"]   +++ building pointers")') fdate()
     call pointers_to_neighbours()

! pointers for merged groups
     allocate(GroupPointer(n_allocated_groups),stat=ierr)
     call trace_allocations(ierr,4.*float(n_allocated_groups),.false.,'GroupPointer')
     do i=1,n_allocated_groups
        GroupPointer(i)=i
     enddo

     write(*,'(26x,"   --- Number of vertices:  ",i9)') n_total_vertex
     write(*,'(26x,"   --- Number of filaments: ",i9)') n_filaments
     write(*,'(26x,"   --- Number of groups:    ",i9)') n_master_groups

     write(*,'("[",a24,"]   +++ starting loop")') fdate()

!**********************************************************************
!             loop over fmax values in decreasing order
!**********************************************************************
     fmax_loop: do index=1, n_total_vertex

#ifdef CHECK_POINTERS
! check dei puntatori
        write(*,'("[",a24,"]   +++ checking pointers")') fdate()
        do vertex=1,n_total_vertex
           do ineigh=1,GroupsPerVertex
              nvertex=VertexPointer(vertex)%Neighbour(ineigh)
              if (nvertex>0) then
                 if (VertexPointer(nvertex)%Neighbour(InverseNeighbour(ineigh)) /= vertex) &
                     write(*,*) 'CP1:',vertex,ineigh,nvertex,VertexPointer(nvertex)%Neighbour(InverseNeighbour(ineigh))
              endif
              
              nfilament=VertexPointer(vertex)%Filament(ineigh)
              if (nfilament>0) then
                 ineigh2=1
                 do while (FilamentPointer(nfilament)%NeighbourVertex(ineigh2)/=vertex .and. ineigh2<GroupsPerVertex)
                    ineigh2=ineigh2+1
                 enddo
                 if (FilamentPointer(nfilament)%NeighbourVertex(ineigh2)/=vertex) then
                    write(*,*) 'CP2:',vertex,ineigh,nfilament,FilamentPointer(nfilament)%NeighbourVertex(:)
                 endif
              endif
           enddo
        enddo

        do nfilament=1,n_filaments
           if (MasterFilaments(nfilament)%alive) then
              do ineigh2=1,GroupsPerVertex
                 nvertex=FilamentPointer(nfilament)%NeighbourVertex(ineigh2)
                 ineigh =FilamentPointer(nfilament)%position(ineigh2)
                 if (nvertex>0) then
                    if (GridVertex(nvertex)%groupID(ineigh)/=is_filament) then
                       write(*,'(''CP3:'',1x,f8.4,99(1x,i6))')GridVertex(index_fmax(index))%fmax,nfilament &
                            ,ineigh2,nvertex,ineigh,GridVertex(nvertex)%GroupID
                    endif
                    if (VertexPointer(nvertex)%Filament(ineigh)/=nfilament) then
                       write(*,*) 'CP4:',nfilament,ineigh2,nvertex,ineigh,VertexPointer(nvertex)%Filament
                    endif
                 endif
              enddo
           endif
        enddo
#endif

! current point
        vertex = index_fmax(index)
        fmax = GridVertex(vertex)%fmax             ! it's fmax value
        velx = GridVertex(vertex)%velx             ! it's x-velocity
        vely = GridVertex(vertex)%vely             ! it's y-velocity
        velz = GridVertex(vertex)%velz             ! it's z-velocity
        do i=1,GroupsPerVertex
           if (GridVertex(vertex)%GroupID(i)>0) then
              group=GroupPointer(GridVertex(vertex)%GroupID(i))
              do while (GroupPointer(group)/=group)
                 group=GroupPointer(group)
              enddo
              GroupID(i) = group
           else
              GroupID(i) = GridVertex(vertex)%GroupID(i)
           endif
        enddo
        i = GridVertex(vertex)%i
        j = GridVertex(vertex)%j
        k = GridVertex(vertex)%k

! check whether original groups are still alive
        do ineigh=1, GroupsPerVertex
           ig = GroupID(ineigh)
           if(ig > 0)then
              if (.not.MasterGroups(ig)%alive) then
                 print ('(''ERROR in group IDs! '',f12.6,9(1x,i6))'),fmax,ig,i,j,k,ineigh
                 print *,MasterGroups(ig)%masterID,MasterGroups(ig)%Peak,MasterGroups(ig)%CoMPos
              endif
           endif
        enddo

! Number of neighbouring groups and filaments
        call number_of_neighbours(GroupID,GroupID_cleaned,neigh_groups,neigh_filaments)

        
#ifdef VERBOSE
        write(2,'(''Particle: '',1x,f10.5,11(1x,i6))')fmax,i,j,k,neigh_groups,neigh_filaments,GroupID_cleaned
#endif

! address the fate of the particle
        if (neigh_groups == 0 .and. neigh_filaments == 0) then 
! **********************************************************************
!                       FIRST CASE: PEAK
! **********************************************************************

           ifate = 1

! create new group
           NewGroupName    = NewGroupName    + 1
           n_master_groups = n_master_groups + 1
           GroupResult     = n_master_groups
           Peak       = i+(j-1)*Lgrid_global(1)+(k-1)*Lgrid_global(1)*Lgrid_global(2)
           xpos = i-0.5 ; ypos = j-0.5 ; zpos = k-0.5
           call init_group(NewGroup &  ! group structure
                ,NewGroupName       &  ! masterID
                ,Peak               &  ! Peak
                ,1                  &  ! Mass
                , xpos, ypos, zpos  &  ! CoMPos
                , velx, vely, velz  &  ! CoMVel
                ,.true.             &  ! modified
                ,.true.)               ! alive
! store group
           MasterGroups(n_master_groups) = NewGroup
! vertex points to new group
!           GridVertex(vertex)%modified  = .true.
           GridVertex(vertex)%MyGroupID = GroupResult

#ifdef VERBOSE
           write(2,'(''New Group '',1x,i6,1x,i10)')NewGroupName,Peak
#endif

        else if (neigh_groups == 1) then
! **********************************************************************
!                       SECOND CASE: 1 GROUP
! **********************************************************************

#ifdef VERBOSE
           write(2,'(''Checking particle '',3(1x,i6),4(1x,f12.6))')i,j,k,fmax,velx,vely,velz
           write(2,'(''versus group: '',1x,i6,1x,i10)')GroupID_cleaned(1),MasterGroups(GroupID_cleaned(1))%Peak
#endif

           call condition_for_accretion(i,j,k &
                ,velx,vely,velz,fmax,GroupID_cleaned(1),distance,radius)

#ifdef VERBOSE
           write(2,*)'Condition:',distance,radius
#endif

           if (distance < radius) then

! accretion on a group
              ifate = 2

!              accretion_counter = accretion_counter + 1
!              accreted = .true.

#ifdef VERBOSE
              write(2,'(''Accreting particle on group'')')
#endif

              GroupResult                  = GroupID_cleaned(1)
              call accretion(GroupResult,i,j,k,fmax,velx,vely,velz)
!              GridVertex(vertex)%modified  = .true.
              GridVertex(vertex)%MyGroupID = GroupResult
           else
! becomes filament
              ifate = 3

#ifdef VERBOSE
              write(2,'(''New filament'')')
#endif

!              GridVertex(vertex)%modified  = .true.
              GroupResult                  = Is_Filament
              GridVertex(vertex)%MyGroupID = GroupResult
           endif

#ifdef VERBOSE
           write(2,*)
#endif

! **********************************************************************
!                       THIRD CASE: >1 GROUP
! **********************************************************************
        else if (neigh_groups > 1) then

#ifdef VERBOSE
           write(2,'(''More than 1 neighbour'')')
#endif

! compute the accretion condition for all groups and decide which one to accrete on
           best_ratio  = (10.*Lgrid_global(1))**2
           GroupResult = 0
           do ig1=1,neigh_groups
#ifdef VERBOSE
              write(2,'(''  Checking particle '',3(1x,i6),4(1x,f12.6))')i,j,k,fmax,velx,vely,velz
              write(2,'(''  versus group: '',1x,i6,1x,i10)')GroupID_cleaned(ig1),MasterGroups(GroupID_cleaned(ig1))%Peak
#endif
               call condition_for_accretion(i,j,k &
                   ,velx,vely,velz,fmax,GroupID_cleaned(ig1),distance,radius)
              ratio=distance/radius
              if (ratio < 1..and. ratio < best_ratio) then
                 best_ratio  = ratio
                 GroupResult = GroupID_cleaned(ig1)
              endif
#ifdef VERBOSE
              write(2,*)'  Condition:',distance,radius,best_ratio
#endif
          enddo

           if (GroupResult > 0) then
!              acctetion_counter = accretion_counter+1
!              accreted = .true.
              call accretion(GroupResult,i,j,k,fmax,velx,vely,velz)
#ifdef VERBOSE
              write(2,'(''  Accreting particle on group '',i6)')GroupResult
#endif
!              GridVertex(vertex)%modified  = .true.
              GridVertex(vertex)%MyGroupID = GroupResult

              ifate=2

           endif

! Then checks whether the groups must be merged together
           nmerge=0
           merge(:,:) = .false.
           do ig1=1,neigh_groups
              do ig2=1,ig1-1
                 call condition_for_merging(Fmax,GroupID_cleaned(ig1),GroupID_cleaned(ig2),merge_flag)
                 merge(ig1,ig2)=merge_flag
                 merge(ig2,ig1)=merge_flag
                 if (merge_flag) nmerge=nmerge+1
              enddo
           enddo

!           if (nmerge>1 .and. GroupResult>0) nwei1=nwei1+1


! Merge the selected groups

           if (nmerge>0) then

!              nmeff=0

              allocate (mass_vector(neigh_groups))
              allocate (index_mass(neigh_groups))

              do ig1=1,neigh_groups
                 mass_vector(ig1) = -MasterGroups(GroupID_cleaned(ig1))%Mass !&
!                      +MasterGroups(GroupID_cleaned(ig1))%CoMPos(1)/(2.&
!                      *Lgrid_global(1)) +1.e-4*MasterGroups(GroupID_cleaned(ig1))%CoMVel(3)
!  **!!** TAKE OUT THE RUBBISH!!!


              enddo
              call index_sort_r4(neigh_groups,mass_vector,index_mass)

              do ig2=1,neigh_groups
                 do ig1=1,ig2-1!neigh_groups
                    if (merge(index_mass(ig1),index_mass(ig2))) then
                       large = GroupID_cleaned(index_mass(ig1))
                       small = GroupID_cleaned(index_mass(ig2))
                       if (MasterGroups(large)%Mass<MasterGroups(small)%Mass) then
                          large = GroupID_cleaned(index_mass(ig2))
                          small = GroupID_cleaned(index_mass(ig1))
                       endif
                       if (large==small) cycle
#ifdef VERBOSE
              write(2,'(''  Merging group '',4(1x,i8),'' with '',4(1x,i8))')ig1,large,MasterGroups(large)%Peak &
                   ,MasterGroups(large)%Mass ,ig2,small,MasterGroups(small)%Peak,MasterGroups(small)%Mass
#endif
                       call merge_groups(large,small,Fmax)
                       if (GroupResult==small) GroupResult = large
                       do ig3=1,neigh_groups
                          if (GroupID_cleaned(ig3)==small) GroupID_cleaned(ig3)=large
                       enddo
                       GroupPointer(small)=large
                       merge(index_mass(ig1),index_mass(ig2))=.false.
                       merge(index_mass(ig2),index_mass(ig1))=.false.
!                       nmeff=nmeff+1

                    endif
                 enddo
              enddo

              deallocate (mass_vector)
              deallocate (index_mass)

!              if (nmeff>0) nm(nmeff)=nm(nmeff)+1
           endif

! If the particle is not accreted to a group, try again, else make it a filament
           if (GroupResult==0) then

#ifdef VERBOSE
              write(2,'(''  Trying again to accrete the particle'')')
#endif

              call number_of_neighbours(GroupID_cleaned,GroupID_cleaned2,neigh_groups,neigh_filaments)

              best_ratio  = (10.*Lgrid_global(1))**2
              do ig1=1,neigh_groups
#ifdef VERBOSE
                 write(2,'(''  Checking particle '',3(1x,i6),4(1x,f12.6))')i,j,k,fmax,velx,vely,velz
                 write(2,'(''  versus group: '',1x,i6,1x,i10)')GroupID_cleaned(ig1),MasterGroups(GroupID_cleaned(ig1))%Peak
#endif
                call condition_for_accretion(i,j,k &
                      ,velx,vely,velz,fmax,GroupID_cleaned2(ig1),distance,radius)
#ifdef VERBOSE
                write(2,*)'  Condition:',distance,radius,best_ratio
#endif
                 ratio=distance/radius
                 if (ratio < 1..and. ratio < best_ratio) then
                    best_ratio  = ratio
                    GroupResult = GroupID_cleaned2(ig1)
                 endif

              enddo

              if (GroupResult > 0) then
!                 acctetion_counter = accretion_counter+1
!                 accreted = .true.
                 call accretion(GroupResult,i,j,k,fmax,velx,vely,velz)
#ifdef VERBOSE
                 write(2,'(''  Accreting particle on group '',i6)')GroupResult
#endif
!                 GridVertex(vertex)%modified  = .true.
                 GridVertex(vertex)%MyGroupID = GroupResult

                 ifate=2

              else
! becomes filament

                 ifate=3

!                 GridVertex(vertex)%modified  = .true.
                 GroupResult                  = Is_Filament
                 GridVertex(vertex)%MyGroupID = GroupResult

#ifdef VERBOSE
                 write(2,'(''  New filament'')')
#endif
!                 if (nmerge.gt.1) nwei2=nwei2+1
              endif
           endif

#ifdef VERBOSE
           write(2,*)
#endif
       else
! **********************************************************************
!                       FOUTH CASE: FILAMENT
! **********************************************************************

           ifate = 3

!           GridVertex(vertex)%modified  = .true.
           GroupResult                  = Is_Filament
           GridVertex(vertex)%MyGroupID = GroupResult
        endif

! update groupID in neighbouring vertices
        do ineigh=1, GroupsPerVertex
           nvertex = VertexPointer(vertex)%Neighbour(ineigh)
           if (nvertex > 0) then
              GridVertex(nvertex)%GroupID(InverseNeighbour(ineigh)) =  GroupResult
           endif
        enddo


! accrete neighbouring filaments to GroupResult 
! NB: here is where the filaments are needed
        if (GroupResult > 0) then
           do ineigh=1, GroupsPerVertex
              nvertex   = VertexPointer(vertex)%Neighbour(ineigh)
              if (nvertex>0) then 
                 if (GridVertex(nvertex)%MyGroupID/=is_filament) nvertex=0
              endif
              nfilament = VertexPointer(vertex)%Filament(ineigh)
              if (nvertex > 0 .and. nfilament > 0) then
                 call abort ('inconsistency in the pointers')

              else if (nvertex > 0) then
! in this case the filament to accrete is a grid vertex
                 GridVertex(nvertex)%MyGroupID = GroupResult
!                 GridVertex(nvertex)%modified  = .true.
                 call accretion(GroupResult,GridVertex(nvertex)%i,GridVertex(nvertex)%j,GridVertex(nvertex)%k &
                      ,fmax,GridVertex(nvertex)%velx,GridVertex(nvertex)%vely,GridVertex(nvertex)%velz)

#ifdef VERBOSE
                 write(32,formato) &
                      GridVertex(nvertex)%i,GridVertex(nvertex)%j,GridVertex(nvertex)%k &
                      ,fmax,GridVertex(nvertex)%velx,GridVertex(nvertex)%vely,GridVertex(nvertex)%velz &
                      ,fate(4),MasterGroups(GroupResult)%Peak,MasterGroups(GroupResult)%Mass
#endif

! update groupID in neighbouring vertices
                 do ineigh2=1, GroupsPerVertex
                    if (VertexPointer(nvertex)%Neighbour(ineigh2) > 0) then
                       GridVertex(VertexPointer(nvertex)%Neighbour(ineigh2))%groupID(InverseNeighbour(ineigh2)) = GroupResult
                    endif
                 enddo

              else if (nfilament > 0) then
                 if (MasterFilaments(nfilament)%alive) then

                    MasterFilaments(nfilament)%MyGroupID = GroupResult
                    MasterFilaments(nfilament)%alive = .false.
! accrete the filament on the group
                    call get_coordinates(MasterFilaments(nfilament)%GlobalID,ii,jj,kk)
                    call accretion(GroupResult,ii,jj,kk &
                         ,fmax,MasterFilaments(nfilament)%velx,MasterFilaments(nfilament)%vely,MasterFilaments(nfilament)%velz)

#ifdef VERBOSE
                    write(32,formato) &
!                         MasterFilaments(nfilament)%i,MasterFilaments(nfilament)%j,MasterFilaments(nfilament)%k &
                         ii,jj,kk &
                         ,fmax,Masterfilaments(nfilament)%velx,MasterFilaments(nfilament)%vely,MasterFilaments(nfilament)%velz&
                         ,fate(4),MasterGroups(GroupResult)%Peak,MasterGroups(GroupResult)%Mass
#endif

! is this filament a neighbour of other vertices?
                    do ineigh2=1, GroupsPerVertex
                       nvertex   = FilamentPointer(nfilament)%NeighbourVertex(ineigh2)
                       neighbour = FilamentPointer(nfilament)%position(ineigh2)
                       if (nvertex>0) then
                          GridVertex(nvertex)%groupID(neighbour) = GroupResult
                       endif
                    enddo
                 endif
              endif
           enddo
        endif


#ifdef VERBOSE
        if (GroupResult>0) then
           pippo=MasterGroups(GroupResult)%Peak
           pippo2=MasterGroups(GroupResult)%Mass
        else
           pippo=-10
           pippo2=0
        endif
        write(32,formato) &
             GridVertex(vertex)%i,GridVertex(vertex)%j,GridVertex(vertex)%k &
!             ,0,GroupID &
             ,GridVertex(vertex)%fmax,GridVertex(vertex)%velx,GridVertex(vertex)%vely,GridVertex(vertex)%velz &
             ,fate(ifate),pippo,pippo2
#endif

     enddo fmax_loop

! test memory requirement
     if(NewGroupName > (NMaster-first_inode)*MaxGroupsOnServer)then
        write (*,*) ' error: I formed: ',NewGroupName,' groups, there is only space for :',(NMaster-first_inode)*MaxGroupsOnServer
        call abort_code(' too many groups ')
     endif

! reset GroupIDs
     do vertex=1, n_total_vertex
        ig = GridVertex(vertex)%MyGroupID
        if (ig > 0) then
           group=GroupPointer(ig)
           do while (GroupPointer(group)/=group)
              group=GroupPointer(group)
           enddo
           GridVertex(vertex)%MyGroupID = MasterGroups(group)%masterID
        endif
        do ineigh=1,GroupsPerVertex
           ig = GridVertex(vertex)%GroupID(ineigh)
           if (ig > 0) then
              group=GroupPointer(ig)
              do while (GroupPointer(group)/=group)
                 group=GroupPointer(group)
              enddo
              GridVertex(vertex)%GroupID(ineigh) = MasterGroups(group)%masterID
           endif
        enddo
     enddo

     write(*,'(29x,"--- Total number of groups",i9)')NewGroupName

! deallocate pointers
     deallocate(index_fmax)
     call trace_allocations(0,-4.*float(n_total_vertex),.false.,'index_fmax')
     deallocate(VertexPointer)
     call trace_allocations(0,-float(VertexPointerTypeSize*n_total_vertex),.false.,'VertexPointer')
     deallocate(FilamentPointer)
     call trace_allocations(0,-float(FilamentPointerTypeSize*n_filaments),.false.,'FilamentPointer')
     deallocate(GroupPointer)
     call trace_allocations(0,-4.*float(n_allocated_groups),.false.,'GroupPointer')

  endif

  return
end subroutine fragment


!!$function locate_value(value ,list, nlist)
!!$  use numbers
!!$  implicit none
!!$  integer(kind=singleI), intent(in) :: value, nlist, list(nlist)
!!$  integer(kind=singleI) :: locate_value
!!$  integer group, ilow, ihigh, lname, hname, name
!!$
!!$  ! Given sorted list(1:nlist), return index where value is
!!$  ! bisection
!!$
!!$  ilow  = 1
!!$  lname = list(ilow)
!!$  ihigh = nlist
!!$  hname = list(ihigh)
!!$  ! fool proof
!!$  if(ihigh == ilow)then
!!$     if(lname /= value)then
!!$        call abort_code(' locate_value: error: value not found')
!!$     else
!!$        locate_group = ilow
!!$        return
!!$     endif
!!$  endif
!!$  !
!!$  group = 0.5*(ilow+ihigh)
!!$  name  = list(group)
!!$  do while (name /= value .and. ilow /= ihigh-1)
!!$     if(value > name)then
!!$        ilow = group
!!$     else if (value <= name) then
!!$        ihigh = group
!!$     endif
!!$     group = 0.5*(ilow+ihigh)
!!$     name  = list(group)
!!$  enddo
!!$  hname = list(ihigh)
!!$  if(value /= hname)then
!!$     call abort_code(" locate_value: value not found")
!!$  else
!!$     locate_group = ihigh
!!$     return
!!$  endif
!!$  !
!!$end function locate_value


subroutine set_offset()

!

  use numbers
  use grid
  use GridData
  implicit none
  integer :: n

  fmax_offset  = 30
  index_offset = fmax_offset  + LGrid_global(3)
  velx_offset  = index_offset + LGrid_global(3)
  vely_offset  = velx_offset  + LGrid_global(3)
  velz_offset  = vely_offset  + LGrid_global(3)

  do n=1,6
     GridOffset(n)%i_offset = 0
     GridOffset(n)%j_offset = 0
     GridOffset(n)%k_offset = 0
  enddo

  GridOffset(1)%i_offset = -1
  GridOffset(2)%i_offset = 1
  GridOffset(3)%j_offset = -1
  GridOffset(4)%j_offset = 1
  GridOffset(5)%k_offset = -1
  GridOffset(6)%k_offset = 1

  InverseNeighbour(1) = 2
  InverseNeighbour(2) = 1
  InverseNeighbour(3) = 4
  InverseNeighbour(4) = 3
  InverseNeighbour(5) = 6
  InverseNeighbour(6) = 5

  return
end subroutine set_offset


subroutine divide_domain

! Divides the domain of processors into master and server processors

  use numbers
  use fragment_mpi
  use grid
  implicit none
  integer :: thisPE, ierr, this_comm
  integer :: iplane, min_planes, tot_planes, nrem, tag, iserver
  real(kind=singleR) :: value

! set NServer: these will be used to store the grid data
! The others will do calculations

! **!!** bisognera` trovare un modo per settare il numero di master

!  NServer = max(NumPEs - 3,1)  ! at least two servers and three masters
!  NMaster = NumPEs - NServer
  NServer = NumPEs - NMaster

  if(MyPE == 0) &
       write (*,*) ' # of masters = ',NMaster,' # of servers = ',NServer
  if(NServer .ge. NumPEs) call abort_code(' NServer > NumPEs')
  if(NMaster <= 1)        call abort_code(' Need at least two masters')
  if(Nserver <= 1)        call abort_code(' Need at least two servers')

  if(MyPE .le. NMaster-1)then
     my_type  = master
  else
     my_type  = server
  endif
  call MPI_Comm_split(mpi_comm_world, my_type, master, master_comm, ierr)
  call MPI_Comm_split(mpi_comm_world, my_type, server, server_comm, ierr)

  if(MyPE == 0)then
     call mpi_comm_size(master_comm, this_comm, ierr)
#ifdef VERBOSE
     write (*,*) ' # of compute nodes = ',this_comm
#endif
  endif
  if(MyPE == NMaster)then
     call mpi_comm_size(server_comm, this_comm, ierr)
#ifdef VERBOSE
     write (*,*) ' # of data nodes = ',this_comm
#endif
  endif

! sets the communication pattern for the master and server sets of processors
  GroupPE = -1
  if (my_type == master) then
     call mpi_comm_rank(master_comm, GroupPE, ierr)
     call mpi_comm_size(master_comm, this_comm, ierr)
  else if (my_type == server) then
     call mpi_comm_rank(server_comm, GroupPE, ierr)     
     call mpi_comm_size(server_comm, this_comm, ierr)
  else
     call abort_code(" error in divide_domain")
  endif

  call mpi_barrier(mpi_comm_world, ierr)

!!$  ! intercomm: does not seem to work?
!!$  if(mype == 0)then
!!$     call mpi_intercomm_create(master_comm, 0, mpi_comm_world, NMaster+1, tag, inter_comm, ierr)
!!$  else if (mype == NMaster+1) then
!!$     call mpi_intercomm_create(server_comm, 0, mpi_comm_world, 0, tag, inter_comm, ierr)
!!$  endif
!!$
!!$  call mpi_barrier(mpi_comm_world, ierr)

! divide grids over servers along z
  min_planes = LGrid_global(3) / NServer  ! average number of planes per processor
  allocate(nplanes(NServer))          ! number of planes on given server
  allocate(plane_start(NServer))      ! global start index of plane (starts from zero)
  allocate(myserver(LGrid_global(3))) ! server node for a given plane
  call trace_allocations(0,4.*float(2*Nserver+Lgrid_global(3)),.false.,'plane stuff')

! determine number of planes on each server
! NB: the plane number runs from 0 to Lgrid_global(3)-1
  nplanes(:)     = min_planes
  plane_start(:) = 0
  nrem   = LGrid_global(3) - SUM(nplanes) ! number of missing planes 
  if(nrem > 0) then
     nplanes(1:nrem) = nplanes(1:nrem) + 1
  endif
  do iplane = 2, NServer
     plane_start(iplane) = plane_start(iplane-1) + nplanes(iplane-1)
  enddo

  if(my_type == server)then
     Lgrid(1:2) = LGrid_global(1:2)
     LGrid(3)   = nplanes(GroupPE+1)
  else
     Lgrid(:)   = LGrid_global(:)   ! masters handle the global grid
  endif

! record which server a given plane lives on
  do iserver=1,NServer
     myserver(1+plane_start(iserver):1+plane_start(iserver)+nplanes(iserver)-1) = iserver-1
  enddo

#ifdef VERBOSE
  if(MyPE==0)then
     do iplane=1,lgrid_global(3)
        write (*,*) ' iplane = ',iplane,' server = ',myserver(iplane)
     enddo
  endif
#endif

! set and allocate Grid Variables for server nodes
  call allocate_server_data ()

! .. for master modes
  call allocate_master_data ()

  return
end subroutine divide_domain


subroutine allocate_server_data()

! Allocates grids for server data

  use fragment_mpi
  use grid
  implicit none
  integer(kind=singleI) :: ierr, n_data

  if(my_type == server)then
     local_z_start = plane_start(GroupPE+1)

! note the two extra planes in z-direction for isolated boundary conditions
     allocate(server_GroupID(LGrid(1),LGrid(2),0:LGrid(3)+1),stat=ierr)
     call trace_allocations(ierr,4.*float(Lgrid(1)*Lgrid(2)*(Lgrid(3)+2)),.false.,'server_GroupID')
     server_GroupID(:,:,:) = 0 ! initialize

#ifdef STORE_VEL
     n_data = Lgrid(1)*Lgrid(2)
     allocate(stored_vel(3,n_data,0:LGrid(3)+1),stat=ierr)
     call trace_allocations(ierr,12.*float(Lgrid(1)*Lgrid(2)*(Lgrid(3)+2)),.false.,'stored_vel')
#endif

#ifdef STORE_FMAX
     n_data = Lgrid(1)*Lgrid(2)
     allocate(stored_fmax(n_data,0:LGrid(3)+1),stat=ierr)
     call trace_allocations(ierr,4.*float(Lgrid(1)*Lgrid(2)*(Lgrid(3)+2)),.false.,'stored_fmax')
     allocate(stored_index(n_data,0:LGrid(3)+1),stat=ierr)
     call trace_allocations(ierr,4.*float(Lgrid(1)*Lgrid(2)*(Lgrid(3)+2)),.false.,'stored_index')
#endif

  endif

  return
end subroutine allocate_server_data


subroutine allocate_master_data()

! Allocates matrices for master processors

  use fragment_mpi
  use grid
  use GroupData
  implicit none
  integer(kind=singleI) :: ierr, ig
  integer(kind=doubleI) :: pp

  if(my_type == master)then
!     allocate(num_this_proc(NServer))
!     allocate(plane_done(LGrid_global(3)))

! allocate group data for master nodes except 0
     if(MyPE >= first_inode )then         !  **2G**
        allocate(ServerGroups(MaxGroupsOnServer),stat=ierr)
        call trace_allocations(ierr,float(GroupDataTypeSize*MaxGroupsOnServer),.false.,'ServerGroups')
! initialize
        do ig=1, MaxGroupsOnServer
           pp=-1
           call init_group(ServerGroups(ig),0,pp,0,0.,0.,0.,0.,0.,0.,.false.,.false.)
        enddo
        ngroups_on_server=0
     endif
     NewGroupName = 0 ! current name of new group
  endif

  return
end subroutine allocate_master_data


subroutine init_run ()
  
! Initialization of all runtime parameters and cosmological functions

  use fragment_mpi
  use grid
  use runtime_parameters
  use pinocchio_common
  implicit none

! local variables
  integer(kind=singleI) :: mpi_string, ierr, count, nsend, i
  integer(kind=singleI) :: r8_size, r4_size, char_size, string_size, integer_size, character_size

  if(MyPE == 0) then

     call read_info_files()

     InputDir     = './Data/'
     OutputDir    = './'
     Fmax_File    = 'Fmax.'//run//'.d'
     Velx_File    = 'Velx.'//run//'.d'
     Vely_File    = 'Vely.'//run//'.d'
     Velz_File    = 'Velz.'//run//'.d'
     RunString    = run

     pbc_flag = .true.
  endif

! create data type for strings
  call mpi_type_contiguous(string_length, mpi_character, mpi_string, ierr)
  call MPI_TYPE_COMMIT(mpi_string,ierr)

! determine size of variables
  call MPI_Pack_size(1,MPI_DOUBLE_PRECISION,MPI_COMM_WORLD,r8_size,ierr)
  call MPI_Pack_size(1,MPI_REAL,MPI_COMM_WORLD,r4_size,ierr)
  call MPI_Pack_size(1,mpi_string,MPI_COMM_WORLD,string_size,ierr)
  call MPI_Pack_size(1,mpi_character,MPI_COMM_WORLD,character_size,ierr)
  call MPI_Pack_size(1,mpi_integer,mpi_comm_world,integer_size,ierr)


! communicate these to all other processes
  count = 0
  nsend = 3*integer_size + 7 * string_size
  if(MyPE ==0) then
     call mpi_pack(LGrid_global,3,mpi_integer,comm_buffer,nsend,count,mpi_comm_world,ierr)
     call mpi_pack(InputDir, 1,mpi_string,comm_buffer,nsend,count,mpi_comm_world,ierr)
     call mpi_pack(OutputDir,1,mpi_string,comm_buffer,nsend,count,mpi_comm_world,ierr)
     call mpi_pack(Fmax_file,1,mpi_string,comm_buffer,nsend,count,mpi_comm_world,ierr)
     call mpi_pack(Velx_file,1,mpi_string,comm_buffer,nsend,count,mpi_comm_world,ierr)
     call mpi_pack(Vely_file,1,mpi_string,comm_buffer,nsend,count,mpi_comm_world,ierr)
     call mpi_pack(Velz_file,1,mpi_string,comm_buffer,nsend,count,mpi_comm_world,ierr)
     call mpi_pack(RunString,1,mpi_string,comm_buffer,nsend,count,mpi_comm_world,ierr)
  endif

  call mpi_bcast(comm_buffer, nsend, MPI_PACKED, 0, MPI_COMM_WORLD,ierr)

! unpack
  if(MyPE /= 0)then
     count = 0

     call mpi_unpack(comm_buffer, nsend, count, LGrid_global, 3, mpi_integer, mpi_comm_world, ierr)
     call mpi_unpack(comm_buffer, nsend, count, InputDir , string_length, mpi_character, mpi_comm_world, ierr)
     call mpi_unpack(comm_buffer, nsend, count, OutputDir, string_length, mpi_character, mpi_comm_world, ierr)
     call mpi_unpack(comm_buffer, nsend, count, Fmax_File, string_length, mpi_character, mpi_comm_world, ierr)
     call mpi_unpack(comm_buffer, nsend, count, Velx_File, string_length, mpi_character, mpi_comm_world, ierr)
     call mpi_unpack(comm_buffer, nsend, count, Vely_File, string_length, mpi_character, mpi_comm_world, ierr)
     call mpi_unpack(comm_buffer, nsend, count, Velz_File, string_length, mpi_character, mpi_comm_world, ierr)
     call mpi_unpack(comm_buffer, nsend, count, RunString, string_length, mpi_character, mpi_comm_world, ierr)

  endif

! initialize evolution loop
  if(MyPE == 0)then
     current_fmax = 1.e12
     final_fmax   = fmax_out(nout+1)
     write(*,*) 'Outputs:'
     do i=1,nout
        write(*,*)i,fmax_out(i),z_out(i)
     enddo
  endif
  call mpi_bcast(current_fmax, 1, mpi_real, 0, mpi_comm_world, ierr)
  call mpi_bcast(final_fmax, 1, mpi_real, 0, mpi_comm_world, ierr)

  return
end subroutine init_run


subroutine read_info_files()

! read all info files, initialize cosmology and parameters

  use numbers
  use pinocchio_common
  use runtime_parameters
  use grid
  use virial_params
  implicit none
  integer(kind=singleI)::Lgrid1,i,ierr
  real(kind=singleR)::var_grid,r,v,vt,Mtot,Mstar
  real(kind=singleR),external::grow,aM_star,an_h,aLambda,power
  logical::cubic=.false.,success=.false.,done,exist

! reads the run name from fragment.info
  open(21,file='fragment.info',status='old',err=57)
  read(21,'(a6)',err=55,end=55)run
  close(21)
  write(*,*)'run: ',run

! Reads the $run.info file
  do while (.not.success)
     open(unit=10,file=run//'.info',status='old',err=56)
     read(10,*)BoxSize(1)
     read(10,*)Omega0
     read(10,*)OmegaLambda
     read(10,*)PowerSpectrum
     read(10,*)Sigma8
     read(10,*)Hubble100
     if (.not.cubic) then
! first: tries to read the three grid lengths (non-cubic case)
        read(10,*,err=50) Lgrid_global
        success  = .true.
     else
! second: reads only one grid length and assumes a cubic box
        read(10,*,err=50) Lgrid1
        success      = .true.
        Lgrid_global = Lgrid1
     endif
50   close(10)
     if(success) then
        InterPartDist=BoxSize(1)/float(Lgrid_global(1))
        BoxSize = InterPartDist*Lgrid_global
     endif
     cubic = .true.
  enddo

  write(*,'("Hubble100     = ",f7.2)')Hubble100
  write(*,'("Omega0        = ",f7.2)')Omega0
  write(*,'("OmegaLambda   = ",f7.2)')OmegaLambda
  write(*,'("Sigma8        = ",f7.2)')Sigma8
  write(*,'("PowerSpectrum = ",f7.2)')PowerSpectrum
  write(*,'("BoxSize       = ",3(1x,f7.2))')BoxSize

! setting cosmology
  filter=2
  call normalization()
  delta_c=1.686d0
  flag_ps=4

!!!  select case (flag_ps)
!!!  case (0)
!!!     write(*,*)'Using Press & Schechter (1974) for the analytic mass function'
!!!  case (1)
!!!     write(*,*)'Using Sheth & Tormen (2001) for the analytic mass function'
!!!  case (2)
!!!     write(*,*)'Using Jenkins et al. (2001) for the analytic mass function'
!!!  case (3)
!!!     write(*,*)'Using Warren et al. (2006) for the analytic mass function'
!!!  case (4)
!!!     write(*,*)'Using Reed et al. (2007) for the analytic mass function'
!!!  end select

  call grow_start()

  write(*,'("Check of normalization: ",2(1x,f7.4))')sqrt(aLambda(8./Hubble100)),Sigma8

! Number of particles and maximum number of groups
  Npart = product(Lgrid_global)

! reads the grid-scale variance from the run.var file
  open(10,file=run//'.var',status='old')
  r=1.e10
  do while (r>0.)
     read(10,*,end=53)r,v,vt
  enddo
  var_grid=vt

! Opens fragment.info again and reads the kernel file name
  open(21,file='fragment.info',status='old')
  read(21,*)
  read(21,'(a6)',end=22)kernel
  write(*,*)'kernel for output files: '//kernel

! final redshift and inverse growing mode
  read(21,*,end=22) zfin
  Flim=1./grow(zfin)

! intermediate outputs
  read(21,*,end=22)nout
  allocate (z_out(nout+1)) ; allocate (fmax_out(nout+1))
  call trace_allocations(0,8.*float(nout+1),.false.,'z_out, fmax_out')
  read(21,*,end=22)z_out(1:nout)
  z_out(nout+1)=zfin

! sorts the output redshifts in decreasing order
  call sort(nout+1,z_out)
  do i=1,nout+1
     fmax_out(i)=1./grow(z_out(i))
  enddo

! computes the particle mass
  Mtot   = 2.778e11*Hubble100**2*Omega0*product(BoxSize)
  Mpart  = Mtot/float(Npart)
  Mstar  = aM_star(zfin)

!     print *,'Global box:'
!     print *,'Grid size:        ',Lgrid_global
!     print *,'No. of particles: ',product(Lgrid_global)
!     print *,'Box size:         ',BoxSize_global
!     print *
!     print *,'Local box:'
!     print *,'Grid size:        ',Lgrid
!     print *,'No. of particles: ',Npart
!     print *,'First particle:   ',Lstart
!     print *,'PBC flags:        ',pbc_flag
!     print *,'Box size:         ',BoxSize
!     print *,'Expected ngrpmax: ',ngrpmax
!     print *,'Mpart:            ',Mpart
!     print *,'M* at zfin:       ',Mstar

! sets parameters
!     0: default parameters
!     1: specify f_a
!     2: specify all parameters
  read(21,*,end=22)param_flag
  if (param_flag.eq.0) then
     f_m  = 0.35
     f_rm = 0.70
     f_s  = 0.06
     f_a  = 0.22 + (log10(var_grid/InterPartDist) - 0.36) * 0.11
     f_a  = min (f_a,0.26)
     f_a  = max (f_a,0.22)
     f_ra = 0.40 - (f_a - 0.22) * 3.5
  else if (param_flag.eq.1) then
     read(21,*,end=22)f_a
     f_m  = 0.35
     f_rm = 0.70
     f_s  = 0.06
     f_ra = 0.40 - (f_a - 0.22) * 3.5
  else if (param_flag.eq.2) then
     read(21,*,end=22)f_m,f_rm,f_a,f_ra,f_s
  else
     call abort_code ('Incorrect param_flag')
     goto 22
  endif
  espo=1.7

!other flags
  read(21,*,end=22)out_flag
  read(21,*,end=22)stop_flag

! endo of reading of fragment.info
  close(21)

  return
! error messages
22 call abort_code('Error in fragment.info')

53 call abort_code(run//'.var file not existent or corrupted')

55 call abort_code('Cannot read run in fragment.info')

56 call abort_code('File '//run//'.info does not exist')

57 call abort_code('File fragment.info does not exist')

end subroutine read_info_files


subroutine number_of_neighbours(GroupID,cleaned_list,n_groups,n_filaments)

! Finds the neighbours and filaments of a vertex

  use numbers
  use GridData, only : GroupsPerVertex
  use FilamentData, only : is_filament
  implicit none
  integer(kind=singleI),intent(in)  :: GroupID(GroupsPerVertex)
  integer(kind=singleI),intent(out) :: cleaned_list(GroupsPerVertex), n_groups, n_filaments

  integer(kind=singleI)::i

  n_filaments = 0
  do i=1,GroupsPerVertex
     if (GroupID(i) == is_filament) n_filaments=n_filaments+1
  enddo

  cleaned_list = GroupID
  call clean_list(GroupsPerVertex,cleaned_list)

  n_groups = 0
  do while (cleaned_list(n_groups+1)>0)
     n_groups=n_groups+1
     if (n_groups == GroupsPerVertex) exit
  enddo

  return
end subroutine number_of_neighbours


subroutine condition_for_accretion(i,j,k,vx,vy,vz,Fmax,grp,dd,rr)

! Condition for accretion

  use grid, only : Lgrid_global
  use GroupData

  implicit none
  integer(kind=singleI),intent(in)::i,j,k,grp
  real(kind=singleR),intent(in)::Fmax,vx,vy,vz
  real(kind=singleR),intent(out)::dd,rr

! local variables
  real(kind=singleR)::rvir,dx,dy,dz,d2
  real(kind=singleR),external::distance,virial

  rvir = virial(MasterGroups(grp)%Mass,Fmax,1)

#ifdef VERBOSE
  write(2,'(''Condition:'',3(1x,i6),4(1x,f12.6))')i,j,k,vx,vy,vz,Fmax
  write(2,'(8x,i6,6(1x,f12.6))')MasterGroups(grp)%Mass,MasterGroups(grp)%CoMPos,MasterGroups(grp)%CoMVel
  write(2,*)'rvir = ',rvir
#endif

  rr   = rvir*rvir
  dd   = 100.*rr
  dx = distance((i-.5)+vx/Fmax, MasterGroups(grp)%CoMPos(1)+MasterGroups(grp)%CoMVel(1)/Fmax &
       ,float(Lgrid_global(1)))
  if (dx < rvir) then
     dy = distance((j-.5)+vy/Fmax, MasterGroups(grp)%CoMPos(2)+MasterGroups(grp)%CoMVel(2)/Fmax &
          ,float(Lgrid_global(2)))
     d2 = dx*dx+dy*dy
     if (d2 < rr) then
        dz = distance((k-.5)+vz/Fmax, MasterGroups(grp)%CoMPos(3)+MasterGroups(grp)%CoMVel(3)/Fmax &
             ,float(Lgrid_global(3)))
        d2 = d2+dz*dz
        if (d2.le.rr) dd=d2
     endif
  endif

  return
end subroutine condition_for_accretion


subroutine condition_for_merging(Fmax,grp1,grp2,merge_flag)

! Checks whether two groups must be merged

  use grid, only : Lgrid_global
  use GroupData
  implicit none
  real(kind=singleR),intent(in)::Fmax
  integer(kind=singleI),intent(in)::grp1,grp2
  logical,intent(out)::merge_flag

  real(kind=singleR)::rvir,rvir1,rvir2,dd,rr,dx,dy,dz
  real(kind=singleR),external::distance,virial
      
  merge_flag=.false.
  rvir1=virial(MasterGroups(grp1)%Mass,Fmax,0)
  rvir2=virial(MasterGroups(grp2)%Mass,Fmax,0)
  rvir=max(rvir1,rvir2)
  rr=rvir*rvir
  dx=distance( &
        MasterGroups(grp1)%CoMPos(1)+MasterGroups(grp1)%CoMVel(1)/Fmax &
       ,MasterGroups(grp2)%CoMPos(1)+MasterGroups(grp2)%CoMVel(1)/Fmax &
       ,float(Lgrid_global(1)))
  if (dx.lt.rvir) then
     dy=distance( &
           MasterGroups(grp1)%CoMPos(2)+MasterGroups(grp1)%CoMVel(2)/Fmax &
          ,MasterGroups(grp2)%CoMPos(2)+MasterGroups(grp2)%CoMVel(2)/Fmax &
          ,float(Lgrid_global(2)))
     dd=dx*dx+dy*dy
     if (dd.lt.rr) then
        dz=distance( &
              MasterGroups(grp1)%CoMPos(3)+MasterGroups(grp1)%CoMVel(3)/Fmax &
             ,MasterGroups(grp2)%CoMPos(3)+MasterGroups(grp2)%CoMVel(3)/Fmax &
             ,float(Lgrid_global(3)))
        dd=dd+dz*dz
        if (dd.le.rr) merge_flag=.true.
     endif
  endif

! Bye!
  return
end subroutine condition_for_merging


function virial(grp_mass,F,flag)

! Gives the virial radius

  use fragment_mpi
  use virial_params
  use pinocchio_common, only: InterPartDist
  implicit none

  real(kind=singleR)::virial
  integer(kind=singleI),intent(in)::grp_mass,flag
  real(kind=singleR),intent(in)::F

! local_variables
  real(kind=singleR)::rlag,zero,corr,variance
  real(kind=singleR),external::aLambda
  real,parameter::exponent=1./3.

  rlag=float(grp_mass)**exponent
  if (flag.eq.0) then
! merging
     virial=rlag*f_m + f_rm
  else
! accretion
     variance=aLambda(rlag*InterPartDist)
     corr=(variance)**(0.5*espo)*F**(-espo)
     corr=min(corr,5.)
     zero=f_ra+f_s*rlag*corr
     virial=f_a*rlag+zero
  endif

  return
end function virial


subroutine accretion(group,i,j,k,Fmax,velx,vely,velz)

! Accretes the particle (i,j,k) on the group

  use GroupData
  use grid, only : Lgrid_global
  use runtime_parameters, only : MassMin
  implicit none
  integer(kind=singleI),intent(in)::group,i,j,k
  real(kind=singleR),intent(in)::Fmax,velx,vely,velz

  MasterGroups(group)%modified=.true.

  call update(MasterGroups(group)%CoMPos(1),MasterGroups(group)%Mass,MasterGroups(group)%CoMVel(1) &
       ,float(i)-0.5, 1, velx, Lgrid_global(1))
  call update(MasterGroups(group)%CoMPos(2),MasterGroups(group)%Mass,MasterGroups(group)%CoMVel(2) &
       ,float(j)-0.5, 1, vely, Lgrid_global(2))
  call update(MasterGroups(group)%CoMPos(3),MasterGroups(group)%Mass,MasterGroups(group)%CoMVel(3) &
       ,float(k)-0.5, 1, velz, Lgrid_global(3))

  MasterGroups(group)%Mass = MasterGroups(group)%Mass + 1

  if (MasterGroups(group)%Mass > MassMin .and. MasterGroups(group)%z_appear == -1.) then
     MasterGroups(group)%z_appear = Fmax
     call write_appearing_group(group,MasterGroups(group)%Mass,MasterGroups(group)%Peak,Fmax)
  endif

  MasterGroups(group)%modified = .true.

#ifdef VERBOSE
  write(2,'(''  > Accreting particle '',3(1x,i4),'' on group '',1x,i6,1x,i10)')i,j,k,group,MasterGroups(group)%Peak
  write(2,'(''  > Final group: '',1x,i6,7(1x,f12.6))')MasterGroups(group)%Mass,MasterGroups(group)%CoMPos &
       ,MasterGroups(group)%CoMVel,MasterGroups(group)%z_appear
#endif

  return
end subroutine accretion


subroutine write_appearing_group(group,mass,peak,fmax)

  use numbers
  use runtime_parameters, only : merger_unit,varnumber
  use GroupData, only : MasterGroups
  implicit none
  integer(kind=singleI),intent(in)::group,mass
  integer(kind=doubleI),intent(in)::peak
  real(kind=singleR),intent(in)::fmax

  integer(kind=doubleI)::peak2=0
!!!#ifdef HDF5_OUT
!!!  character(len=16)::hdf5label
!!!  integer(kind=singleI)::sI_vector(4)
!!!  integer(kind=doubleI)::dI_vector(2)
!!!  real   (kind=singleR)::sR_vector(3)
!!!#endif

  write(merger_unit)fmax,MasterGroups(group)%MasterID,peak,mass,fmax,0,peak2,0,0.0

!!!#ifdef HDF5_OUT
!!!  varnumber=varnumber+1
!!!  write(hdf5label,'(i12)')varnumber
!!!  hdf5label='sRea'//trim(adjustl(hdf5label))
!!!  sR_vector=(/fmax,fmax,0/)
!!!  call hdf5_write_data(merger_unit,trim(hdf5label),sR_vector)
!!!  hdf5label(1:4)='sInt'
!!!  sI_vector=(/MasterGroups(group)%MasterID,mass,0,0/)
!!!  call hdf5_write_data(merger_unit,trim(hdf5label),sI_vector)
!!!  hdf5label(1:4)='dInt'
!!!  dI_vector=(/peak,peak2/)
!!!  call hdf5_write_data(merger_unit,trim(hdf5label),dI_vector)
!!!#endif

  return
end subroutine write_appearing_group


subroutine merge_groups(grp1,grp2,fmax)

! Merges grp2 into grp1

  use runtime_parameters, only : merger_unit, MassMin, varnumber
  use grid, only : Lgrid_global
  use GroupData
  implicit none
  integer(kind=singleI),intent(in)::grp1,grp2
  real(kind=singleR),intent(in)::fmax

! local variables
  integer(kind=singleI)::i1,ic,mass_of_merger
!!!#ifdef HDF5_OUT
!!!  character(len=16)::hdf5label
!!!  integer(kind=singleI)::sI_vector(4)
!!!  integer(kind=doubleI)::dI_vector(2)
!!!  real   (kind=singleR)::sR_vector(3)
!!!#endif

  mass_of_merger = MasterGroups(grp1)%Mass + MasterGroups(grp2)%Mass

! appearance time
  if (mass_of_merger > MassMin .and. MasterGroups(grp1)%z_appear == -1.) &
       MasterGroups(grp1)%z_appear = Fmax

  if (MasterGroups(grp1)%Mass > MassMin .and. MasterGroups(grp2)%Mass > MassMin) then
! writes information on the merger to disc

     write(merger_unit)fmax &
          ,MasterGroups(grp1)%masterID,MasterGroups(grp1)%Peak,MasterGroups(grp1)%Mass,MasterGroups(grp1)%z_appear &
          ,MasterGroups(grp2)%masterID,MasterGroups(grp2)%Peak,MasterGroups(grp2)%Mass,MasterGroups(grp2)%z_appear

!!!#ifdef HDF5_OUT
!!!     varnumber=varnumber+1
!!!     write(hdf5label,'(i12)')varnumber
!!!     hdf5label='sRea'//trim(adjustl(hdf5label))
!!!     sR_vector=(/fmax,MasterGroups(grp1)%z_appear,MasterGroups(grp2)%z_appear/)
!!!     call hdf5_write_data(merger_unit,trim(hdf5label),sR_vector)
!!!     hdf5label(1:4)='sInt'
!!!     sI_vector=(/MasterGroups(grp1)%masterID,MasterGroups(grp1)%Mass,&
!!!          MasterGroups(grp2)%masterID,MasterGroups(grp2)%Mass/)
!!!     call hdf5_write_data(merger_unit,trim(hdf5label),sI_vector)
!!!     hdf5label(1:4)='dInt'
!!!     dI_vector=(/MasterGroups(grp1)%Peak,MasterGroups(grp2)%Peak/)
!!!     call hdf5_write_data(merger_unit,trim(hdf5label),dI_vector)
!!!#endif

  else if (MasterGroups(grp1)%Mass <= MassMin .and. mass_of_merger > MassMin) then
     call write_appearing_group(grp1,mass_of_merger,MasterGroups(grp1)%Peak,Fmax)
  endif

  MasterGroups(grp1)%modified=.true.
  MasterGroups(grp2)%modified=.true.
  MasterGroups(grp2)%alive=.false.
  MasterGroups(grp2)%z_merge=fmax

! updates centres and velocities
  do ic=1,3
     call update( &
           MasterGroups(grp1)%CoMPos(ic),MasterGroups(grp1)%Mass,MasterGroups(grp1)%CoMVel(ic) &
          ,MasterGroups(grp2)%CoMPos(ic),MasterGroups(grp2)%Mass,MasterGroups(grp2)%CoMVel(ic) &
          ,Lgrid_global(ic))
  enddo

! updates mass
  MasterGroups(grp1)%Mass = mass_of_merger

!!!! updates list of mergers
!!!  call update_merger_list(MasterGroups(grp1)%MasterID,MasterGroups(grp2)%MasterID)

! Bye!
  return
end subroutine merge_groups


subroutine update(x1,m1,v1,x2,m2,v2,Lgrid)

! Updates the centre position of group 1 respecting PBCs

  use numbers
  implicit none
  real(kind=singleR),intent(inout)::x1,v1
  real(kind=singleR),intent(in)::x2,v2
  integer(kind=singleI),intent(in)::m1,m2,Lgrid

  real(kind=singleR)::d
  real(kind=singleR),external::pbc

  d=abs(x1-x2)
  if (d.le.Lgrid/2) then
! in this case theres no problem
     x1=(x1*m1+x2*m2)/float(m1+m2)
  else
! PBC must be considered
     if (x1.gt.Lgrid/2) then
! in this case xc is in the second half, and x2 -> x2+Lgrid
        x1=(x1*m1+(x2+Lgrid)*m2)/float(m1+m2)
     else
! in this case xc is in the first half, and x2 -> x2-Lgrid
        x1=(x1*m1+(x2-Lgrid)*m2)/float(m1+m2)
     endif
! checks PBC again
     x1=pbc(x1,float(Lgrid))
  endif

! velocity
  v1=(v1*m1+v2*m2)/float(m1+m2)

! bye!
  return
end subroutine update


SUBROUTINE clean_list(n,arr)

  use numbers
  implicit none
  integer(kind=singleI),intent(in)::n
  integer(kind=singleI),intent(inout),dimension(n)::arr

  INTEGER(KIND=SINGLEI)::i,j,a
  logical::done

  do j=2,n

     do i=j-1,1,-1
        if (arr(i).eq.arr(j)) arr(j)=0
     enddo

     a=arr(j)
     done=.false.
     do i=j-1,1,-1
        if (arr(i)>0)  then
           done=.true.
           exit
        endif
        arr(i+1)=arr(i)
     enddo
     if (.not.done) i=0
     arr(i+1)=a
  enddo
  return
END SUBROUTINE clean_list


function fdate ()
  
  use numbers
  implicit none
  character(len=24) :: fdate
  character(len=8)  :: date
  character(len=10) :: time
  character(len=5)  :: timezone
  integer                   :: values(8)
  
  call date_and_time (date, time, timezone, values)
  
  write (fdate, &
       &   "(I2.2,'-',I2.2,'-',I4.4,' ',I2.2,':',I2.2,'.',I2.2,'     ')") &
       &   values(2), values(3), values(1), values(5), values(6), &
       &   values(7)
  
  return
end function fdate


subroutine output(iout,nfmax)

  use numbers
  use fragment_mpi
  use GroupData, only : ServerGroups, n_master_groups, MaxGroupsOnServer, first_inode
  use runtime_parameters
  use pinocchio_common, only : Mpart, InterPartDist, run, BoxSize
#ifdef HDF5_OUT
  use hdf5_wrapper; use pinocchio_hdf5
#endif
  implicit none
  integer(kind=singleI),intent(inout) :: iout
  real(kind=singleR),   intent(in)    :: nfmax
  real(kind=singleR),external :: grow, fomega, Hubble, pbc

! local variables 
  integer(kind=singleI) :: i, ierr, counter
  integer,parameter::nvector=8
  real(kind=singleR) :: vector(nvector)
  character(len=2*string_length) :: filename
  character(len=4) :: redshift
  character(len=2) :: proclabel

  type OneGroupType
     sequence
     integer(kind=singleI) :: masterID
     real(kind=singleR)    :: logMass, qPos(3), xPos(3), Vel(3)
     integer(kind=singleI) :: Mass
     integer(kind=doubleI) :: Peak
  end type OneGroupType
  type(OneGroupType) :: OneGroup

#ifdef HDF5_OUT
  real   (kind=singleR)::sR_vector(10)
  integer(kind=singleI)::sI_vector(2)
  integer(kind=doubleI)::dI_scalar
  character(len=80) :: var_name
#endif

  if (my_type==Master) then
     if (MyPE==0) then
! processor 0 determines whether to output and computes the coefficients
        if (fmax_out(iout)==nfmax) then
           vector(1)=z_out(iout)
           vector(2)=InterPartDist
           vector(3)=grow(z_out(iout))
           vector(4)=vector(2)*vector(3)*fomega(z_out(iout))*Hubble(z_out(iout))/(1.+z_out(iout))
           vector(5)=Mpart
           vector(6)=BoxSize(1)
           vector(7)=BoxSize(2)
           vector(8)=BoxSize(3)
           iout = iout + 1
        else
           vector=-99.
        endif
     endif

     call mpi_bcast(vector, nvector, mpi_real, 0, master_comm, ierr)

     if (MyPE>=first_inode) then
! other master processors write the group catalogues
        if (vector(1)/=-99.) then
           if (vector(1) < 10) then
              write(redshift,'(f4.2)')vector(1)
           else
              write(redshift,'(f4.1)')vector(1)
           endif
           if (MyPE < 10) then
              write(proclabel,'(''0'',i1)')MyPE
           else
              write(proclabel,'(i2)')MyPE
           endif

           filename = 'frag.'//redshift//'.'//trim(RunString)//'.grp.out.'//proclabel

#ifdef HDF5_OUT
           call hdf5_create_file(file_handle, filename)
           call rw_header('write')
#endif
#ifdef ASCII_OUT
           open(20,file=trim(filename),status='unknown')
#endif
#ifdef UNFORM_OUT
           open(20,file=trim(filename),status='unknown',form='unformatted')
#endif

           counter=0
           do i=1,MaxGroupsOnServer
              if (ServerGroups(i)%alive) then

                 OneGroup%MasterID = ServerGroups(i)%MasterID
                 OneGroup%LogMass  = log10(ServerGroups(i)%Mass*vector(5))
                 OneGroup%qPos(1)  = pbc(ServerGroups(i)%CoMPos(1)*vector(2),vector(6))
                 OneGroup%qPos(2)  = pbc(ServerGroups(i)%CoMPos(2)*vector(2),vector(7))
                 OneGroup%qPos(3)  = pbc(ServerGroups(i)%CoMPos(3)*vector(2),vector(8))
                 OneGroup%xPos(1)  = pbc((ServerGroups(i)%CoMPos(1)+ServerGroups(i)%CoMVel(1)*vector(3))*vector(2),vector(6))
                 OneGroup%xPos(2)  = pbc((ServerGroups(i)%CoMPos(2)+ServerGroups(i)%CoMVel(2)*vector(3))*vector(2),vector(7))
                 OneGroup%xPos(3)  = pbc((ServerGroups(i)%CoMPos(3)+ServerGroups(i)%CoMVel(3)*vector(3))*vector(2),vector(8))
                 OneGroup%Vel(:)   = ServerGroups(i)%CoMVel(:)*vector(4)
                 OneGroup%Mass     = ServerGroups(i)%Mass
                 OneGroup%Peak     = ServerGroups(i)%Peak
                 counter=counter+1

#ifdef ASCII_OUT
                 write(20,'(1x,i9,1x,f7.4,6(1x,f7.2),3(1x,f9.2),1x,i8,1x,i12)') OneGroup
#endif
#ifdef UNFORM_OUT
                 write(20) OneGroup
#endif
#ifdef HDF5_OUT
                 write(var_name,'(i12)')counter
                 var_name='/Group_sR'//trim(var_name)
                 sR_vector=(/OneGroup%LogMass,OneGroup%qPos(:),OneGroup%xPos(:),OneGroup%Vel(:)/)
                 call hdf5_write_data(file_handle,trim(var_name),sR_vector)
                 sI_vector=(/OneGroup%MasterID,OneGroup%Mass/)
                 write(var_name,'(i12)')counter
                 var_name='/Group_sI'//trim(var_name)
                 call hdf5_write_data(file_handle,trim(var_name),sI_vector)
                 dI_scalar=OneGroup%Peak
                 write(var_name,'(i12)')counter
                 var_name='/Group_dI'//trim(var_name)
                 call hdf5_write_data(file_handle,trim(var_name),dI_scalar)
#endif
              endif
           enddo
#ifdef ASCII_OUT
           if (counter==0) write(20,*)
           close(20)
#endif
#ifdef UNFORM_OUT
           if (counter==0) write(20)
           close(20)
#endif
#ifdef HDF5_OUT
           call hdf5_close_file(file_handle)
#endif

           write(*,'("Processor ",i2," has written ",i9," groups in ",a)')MyPE,counter,trim(filename)
        endif
     endif
  endif

  return
end subroutine output


subroutine write_mergers_info()

! writes the mergers.info file

  use fragment_mpi
  use fundamentals, only : first_inode, NMaster
  use GroupData, only : NewGroupName
  use runtime_parameters, only : merger_unit
  implicit none

  if (MyPE==0) then
     open(merger_unit,file='mergers.info',status='unknown')
     write(merger_unit,*)first_inode
     write(merger_unit,*)NMaster-1
     write(merger_unit,*)NewGroupName
     close(merger_unit)
  endif

  return
end subroutine write_mergers_info


subroutine trace_allocations(ierr,size,flag,name)

  use fragment_mpi
  implicit none
  integer(kind=singleI),intent(in)::ierr
  real(kind=singleR),intent(in)::size
  logical,intent(in)::flag
  character(len=*),intent(in)::name
#ifdef ALLOC_VERBOSE
  logical,parameter::verbose=.true.
#else
  logical,parameter::verbose=.false.
#endif

  real(kind=singleR),save :: total_allocated=0.,max_allocated=0.
  real(kind=singleR)      :: new_allocated
  logical,save::firsttime=.true.
  character(len=80)::filename
  integer,save::unit
  logical::open_flag

  if (firsttime) then
     if (MyPE<10) then
        write(filename,'("trace_allocations.",i1)')MyPE
     else if (MyPE<100) then
        write(filename,'("trace_allocations.",i2)')MyPE
     else
        write(filename,'("trace_allocations.",i3)')MyPE
     endif
     unit=1
     open(unit,file=trim(filename),status='unknown')
     firsttime=.false.
  endif

  if(ierr /= 0)then
     write (*,*) ' Processor ',MyPE
     write (*,*) ' tried to allocated memory for ',trim(name)
     write (*,*) ' total already allocated: ',total_allocated
     new_allocated = size/1024.**2
     write (*,*) ' required memory: ',new_allocated
     call abort_code('overflow in memory allocation')
  endif

  new_allocated = size/1024.**2
  total_allocated = total_allocated + new_allocated
  max_allocated = max( total_allocated, max_allocated )

!  inquire(file=trim(filename),opened=open_flag)
!  if (.not.open_flag) then
!     write(*,*)'CHIUSO!',ierr,size,flag,trim(name)
!     call abort_code('CHIUSO')
!  endif

  if (flag) then
     write(unit,'("Proc. ",i3,", status of allocations: ",2(1x,f9.3," Mb"))',err=2) &
          MyPE,total_allocated,max_allocated
  else if (verbose) then
     if (size>0.) then
        write(unit,'("Proc. ",i3,", allocation of ",a,3(1x,f9.3," Mb"))',err=2) &
             MyPE,trim(name),new_allocated,total_allocated,max_allocated
     else if (size<0.) then
        write(unit,'("Proc. ",i3,", deallocation of ",a,3(1x,f9.3," Mb"))',err=2) &
             MyPE,trim(name),new_allocated,total_allocated,max_allocated
     else
        write(unit,'("Proc. ",i3,", no allocation of ",a,3(1x,f9.3," Mb"))',err=2) &
             MyPE,trim(name),new_allocated,total_allocated,max_allocated
     endif
  endif

  return
!1 write(*,*)'EOF',ierr,size,flag,trim(name)
!  call abort_code('EOF')
2 write(*,*)'ERR',ierr,size,flag,trim(name)
  call abort_code('ERR')

end subroutine trace_allocations


#ifdef RECOVER
subroutine write_recover_files(next_fmax,counter)

  use fragment_mpi
  use GridData
  use GroupData
  use grid, only : Lgrid, server_GroupID
  implicit none
  real(kind=singleR),intent(in)    :: next_fmax
  integer(kind=singleI),intent(in) :: counter
  integer(kind=singleI),external :: system

! local variables
  character(len=80)     :: recover_file
  integer(kind=singleI) :: ierr, i, readPE, number, ii
  integer(kind=singleI),save :: nskip=0

  nskip=nskip+1
  if (nskip/=StepsToSkip) then
     return
  else
     nskip=0
  endif

  if (MyPE==0) then
     open (recover_unit,file='recoverfile',status='unknown',position='append')
     write(recover_unit,'(a11,1x,i6)')'saving step',counter
     close(recover_unit)

! queries all processors
     do readPE = first_inode, NumPEs-1
!        call mpi_recv(number, 1, mpi_integer, readPE,  readPE, mpi_comm_world, mpi_status, ierr)
!        write(*,'("Proc. ",i3," is writing recovery file")')readPE
        call mpi_recv(number, 1, mpi_integer, readPE,  readPE, mpi_comm_world, mpi_status, ierr)
        write(*,'("Recover file written by proc. ",i3)')readPE
     enddo
  else if (my_type == master) then
     if (GroupPE>=first_inode) then
! write the group catalogue to the disc
        write(recover_file,*)MyPE
        recover_file='recover.'//trim(adjustl(recover_file))
! saves a copy of the old recover file
        ii=system('/bin/mv -f '//trim(recover_file)//' '//trim(recover_file)//'.old')
!        number=1
!        call mpi_send(number, 1, mpi_integer, 0, MyPE, mpi_comm_world, ierr)
        open(unit=recover_unit,file=trim(recover_file),status='new',form='unformatted')
        write(recover_unit)counter
        write(recover_unit)ngroups_on_server
        do i=1,ngroups_on_server
           write(recover_unit)ServerGroups(i)
        enddo
        close(recover_unit)
        call mpi_send(number, 1, mpi_integer, 0, MyPE, mpi_comm_world, ierr)
     endif
  else if (my_type == server) then
! write the GroupID array to the disc
     write(recover_file,*)MyPE
     recover_file='recover.'//trim(adjustl(recover_file))
! saves a copy of the old recover file
     ii=system('/bin/mv -f '//trim(recover_file)//' '//trim(recover_file)//'.old')
!     number=1
!     call mpi_send(number, 1, mpi_integer, 0, MyPE, mpi_comm_world, ierr)
     open(unit=recover_unit,file=trim(recover_file),status='new',form='unformatted')
     write(recover_unit)counter
     write(recover_unit)Lgrid
     write(recover_unit)server_GroupID
     close(recover_unit)
     call mpi_send(number, 1, mpi_integer, 0, MyPE, mpi_comm_world, ierr)
  endif

  call mpi_barrier(mpi_comm_world,ierr)

  if (MyPE==0) then
     open(recover_unit,file='recoverfile',status='unknown',position='append')
     write(recover_unit,'(a11,1x,i6)')    'saved step ',counter
     write(recover_unit,'(a11,1x,f20.14)')'last fmax: ',next_fmax
     write(recover_unit,'(a11,1x,i15)')   'last group:',NewGroupName
     close(recover_unit)
  else
! when all recover files are saved, remove the old ones
     ii=system('/bin/rm -f '//trim(recover_file)//'.old')
  endif

  return
end subroutine write_recover_files


subroutine recover_run(current_fmax,counter)

! Checks the existence of a recoverfile and, if it exists, loads the groups catalogues
! and server_GroupID arrays from the files.
  
  use fragment_mpi
  use GroupData
  use GridData
  use grid, only : Lgrid, server_GroupID
  use runtime_parameters, only : merger_unit
  implicit none
  integer(kind=singleI),intent(inout) :: counter
  real   (kind=singleR),intent(inout) :: current_fmax

! local variables
  logical               :: flag,saving=.false., merger_exist
  integer(kind=singleI) :: last_counter,icounter,Lgrid_recover(3),ierr,i, nskip
  real(kind=singleR)    :: last_fmax, fmax_read
  character(len=11)     :: string
  character(len=80)     :: recover_file

  if (MyPE==0) then
     inquire(file='recoverfile',exist=flag)
     if (flag) then
        write(*,*) 'RECOVER FILE FOUND'
        open (recover_unit,file='recoverfile',status='old')
        do while (.true.)
           read (recover_unit,'(a11,1x,i6)',end=1)string,icounter
           saving = .true.
           read (recover_unit,'(a11,1x,i6)',end=1)string,last_counter
           saving = .false.
           read (recover_unit,'(a11,1x,f20.14)',end=1)string,last_fmax
           read (recover_unit,'(a11,1x,i15)',end=1)string,NewGroupName
        enddo
1       close(recover_unit)
        if (icounter /= last_counter) then
           write(*,*) 'Inconsistency in recoverfile'
           call abort_code ('Aborting...')
        endif
        if (saving) then
           write(*,*) 'Run crashed while saving, cannot be recovered'
           call abort_code ('Aborting...')
        endif
        write(*,*) 'RUN WILL START FROM FMAX = ',last_fmax,last_counter
        write(*,*) 'ALREADY CREATED ',NewGroupName,' GROUPS'
        current_fmax=last_fmax
        counter=last_counter+1
     endif
  endif

  call mpi_bcast(flag, 1, mpi_logical, 0, mpi_comm_world, ierr)
  call mpi_bcast(counter, 1, mpi_logical, 0, mpi_comm_world, ierr)
  call mpi_bcast(current_fmax, 1, mpi_real, 0, mpi_comm_world, ierr)

  if (flag) then
     if (my_type == master) then
        if (MyPE==0) then

! open list_of_mergers.dat and points to the correct 
           inquire(file='list_of_mergers.dat',exist=merger_exist)
           if (.not.merger_exist) then
              write(*,*)'SEVERE: list_of_mergers.dat does not exist'
              call abort_code('aborting...')
           endif

           open(merger_unit,file='list_of_mergers.dat',status='old',form='unformatted')

           fmax_read=huge(1.)
           nskip=0
           do while(fmax_read>current_fmax)
              nskip=nskip+1
              read(merger_unit,end=2)fmax_read
           enddo
2          nskip=nskip-1
           close(merger_unit)
           open(merger_unit,file='list_of_mergers.dat',status='old',form='unformatted')
           do i=1,nskip
              read(merger_unit)
           enddo
           write(*,'("ALIGNED list_of_mergers.dat SKIPPING ",i9," LINES")')nskip

        else if (GroupPE>=first_inode) then
! write the group catalogue to the disc
           write(recover_file,*)MyPE
           recover_file='recover.'//trim(adjustl(recover_file))
           open (unit=recover_unit,file=trim(recover_file),status='old',form='unformatted')
           read (recover_unit)icounter
           read (recover_unit)ngroups_on_server
           if (ngroups_on_server > MaxGroupsOnServer) then
              write(*,*) 'Inconsistent group number in file '//trim(recover_file)
              write(*,*) ngroups_on_server, MaxGroupsOnServer
              call abort_code('Aborting...')
           endif
           do i=1,ngroups_on_server
              read (recover_unit)ServerGroups(i)
           enddo
           close(recover_unit)
           write(*,'("Master ",i3," read ",i9," groups from file ",a)')MyPE,ngroups_on_server,trim(recover_file)
        endif

     else if (my_type == server) then

 ! write the GroupID array to the disc
        write(recover_file,*)MyPE
        recover_file='recover.'//trim(adjustl(recover_file))
        open (unit=recover_unit,file=trim(recover_file),status='old',form='unformatted')
        read (recover_unit)icounter
        if (icounter/=counter-1) then
           write(*,*) 'Inconsistent counter in file '//trim(recover_file)
           write(*,*) counter-1,icounter
           call abort_code('Aborting...')
        endif
        read (recover_unit)Lgrid_recover              
        if (Lgrid_recover(1) /= Lgrid(1) .or. Lgrid_recover(2) /= Lgrid(2) &
             .or. Lgrid_recover(3) /= Lgrid(3)) then
           write(*,*) 'Inconsistent grid in file '//trim(recover_file)
           write(*,*) Lgrid
           write(*,*) Lgrid_recover
           call abort_code('Aborting...')
        endif
        read (recover_unit)server_GroupID
        close(recover_unit)
        write(*,'("Server ",i3," read GroupIDs from file ",a)')MyPE,trim(recover_file)
     endif
  else
     if (MyPE==0) then
! opens the list_of_mergers.dat file
           open(merger_unit,file='list_of_mergers.dat',status='new',form='unformatted')
     endif
  endif

  return
end subroutine recover_run
#endif


!!!subroutine update_merger_list(g1,g2)
!!!
!!!  use merger_list_module
!!!  use fundamentals , only:MaxVertexOnMaster
!!!  implicit none
!!!  integer(kind=singleI),intent(in)::g1,g2
!!!  integer(kind=singleI),allocatable::tmp(:,:)
!!!  integer(kind=singleI)::i, list_size_old, ierr
!!!
!!!  if (.not.allocated(merger_list)) then
!!!     n_listed_groups=0
!!!     merger_list_size=MaxVertexOnMaster/30
!!!     allocate(merger_list(2,merger_list_size),stat=ierr)
!!!     call trace_allocations(ierr,4.*float(2*merger_list_size),.false.,'merger_list')
!!!  endif
!!!
!!!  do i=1,n_listed_groups
!!!     if (merger_list(1,i)==g2) then
!!!        merger_list(1,i)=g1
!!!     endif
!!!  enddo
!!!
!!!  n_listed_groups=n_listed_groups+1
!!!
!!!  if (n_listed_groups>merger_list_size) then
!!!
!!!!     write(*,*)'Reallocating merger_list matrix'
!!!
!!!     allocate(tmp(2,merger_list_size),stat=ierr)
!!!     call trace_allocations(ierr,4.*float(2*merger_list_size),.false.,'tmp')
!!!     tmp=merger_list
!!!     deallocate(merger_list)
!!!     call trace_allocations(0,-4.*float(2*merger_list_size),.false.,'merger_list')
!!!     list_size_old=merger_list_size
!!!     merger_list_size=merger_list_size+MaxVertexOnMaster/30
!!!     allocate (merger_list(2,merger_list_size),stat=ierr)
!!!     call trace_allocations(ierr,4.*float(2*merger_list_size),.false.,'merger_list')
!!!     merger_list(:,1:list_size_old)=tmp
!!!     merger_list(:,list_size_old+1:merger_list_size)=0
!!!     deallocate(tmp)
!!!     call trace_allocations(0,-4.*float(2*list_size_old),.false.,'tmp')
!!!  endif
!!!
!!!  merger_list(1,n_listed_groups)=g1
!!!  merger_list(2,n_listed_groups)=g2
!!!
!!!  return
!!!end subroutine update_merger_list

!   int MPI_Recv( void *buf, int count, MPI_Datatype datatype, int source, int tag, MPI_Comm comm, MPI_Status *status )
!   int MPI_Send( void *buf, int count, MPI_Datatype datatype, int dest, int tag, MPI_Comm comm )

! int MPI_Sendrecv( void *sendbuf, int sendcount, MPI_Datatype sendtype, 
!                 int dest, int sendtag, 
!                  void *recvbuf, int recvcount, MPI_Datatype recvtype, 
!                 int source, int recvtag, MPI_Comm comm, MPI_Status *status )

