!*****************************************************************
!*                       PINOCCHI0  V2.2                         *
!*  (PINpointing Orbit-Crossing Collapsed HIerarchical Objects)  *
!*****************************************************************
!
!This code was written by
!Pierluigi Monaco, Tom Theuns, Giuliano Taffoni
!Copyright (C) 2008
!
!web page: http//adlibitum.oats.inaf.it/monaco/Homepage/Pinocchio/
!
!This program is free software; you can redistribute it and/or modify
!it under the terms of the GNU General Public License as published by
!the Free Software Foundation; either version 2 of the License, or
!(at your option) any later version.
!
!This program is distributed in the hope that it will be useful,
!but WITHOUT ANY WARRANTY; without even the implied warranty of
!MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!GNU General Public License for more details.
!
!You should have received a copy of the GNU General Public License
!along with this program; if not, write to the Free Software
!Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

module scratch_io
  interface write_scratch
     module procedure write_scratch_r4
     module procedure write_scratch_i4
     module procedure write_scratch_plane_r4
  end interface

  interface read_scratch
     module procedure read_scratch_r4
     module procedure read_scratch_i4
  end interface

contains
  subroutine write_scratch_r4(unit, size, plane)
    use numbers
    use runtime_parameters
    use fragment_mpi
    implicit none
    integer, intent(in)             :: unit, size
    real(kind=singleR), intent(in)  :: plane(size)
#ifdef NOSCRATCH
    character(len=80)::label
    character(len=4) ::proclabel
    write(proclabel,'(i4)')MyPE
    write(label,*)unit
    label = 'tmp_local/scratch'//trim(adjustl(proclabel))//'.'//trim(adjustl(label))
    open (unit,file=trim(label),status='new',form='unformatted')
    write(unit) plane
    close(unit)
#else
    open(unit=unit,status='scratch',form='unformatted')
    write (unit) plane
#endif
  end subroutine write_scratch_r4

  subroutine write_scratch_plane_r4(unit, size, plane)
    use numbers
    use runtime_parameters
    use fragment_mpi
    implicit none
    integer, intent(in)             :: unit, size(2)
    real(kind=singleR), intent(in)  :: plane(size(1),size(2))
#ifdef NOSCRATCH
    character(len=80)::label
    character(len=4) ::proclabel
    write(proclabel,'(i4)')MyPE
    write(label,*)unit
    label = 'tmp_local/scratch'//trim(adjustl(proclabel))//'.'//trim(adjustl(label))
    open (unit,file=trim(label),status='new',form='unformatted')
    write(unit) plane
    close(unit)
#else
    open(unit=unit,status='scratch',form='unformatted')
    write (unit) plane
#endif
  end subroutine write_scratch_plane_r4

  subroutine read_scratch_r4(unit, size, plane)
    use numbers
    use runtime_parameters
    use fragment_mpi
    implicit none
    integer, intent(in)              :: unit, size
    real(kind=singleR), intent(out)  :: plane(size)
#ifdef NOSCRATCH
    character(len=80)::label
    character(len=4) ::proclabel
    write(proclabel,'(i4)')MyPE
    write(label,*)unit
    label = 'tmp_local/scratch'//trim(adjustl(proclabel))//'.'//trim(adjustl(label))
    open (unit,file=trim(label),status='old',form='unformatted')
    read (unit) plane
    close(unit)
#else
    rewind (unit)
    read   (unit) plane
#endif
  end subroutine read_scratch_r4

  subroutine write_scratch_i4(unit, size, plane)
    use numbers
    use runtime_parameters
    use fragment_mpi
    implicit none
    integer, intent(in)             :: unit, size
    integer(kind=singleI), intent(in)  :: plane(size)
#ifdef NOSCRATCH
    character(len=80)::label
    character(len=4) ::proclabel
    write(proclabel,'(i4)')MyPE
    write(label,*)unit
    label = 'tmp_local/scratch'//trim(adjustl(proclabel))//'.'//trim(adjustl(label))
    open (unit,file=trim(label),status='new',form='unformatted')
    write(unit) plane
    close(unit)
#else
    open(unit=unit,status='scratch',form='unformatted')
    write (unit) plane
#endif
  end subroutine write_scratch_i4

  subroutine read_scratch_i4(unit, size, plane)
    use numbers
    use runtime_parameters
    use fragment_mpi
    implicit none
    integer, intent(in)              :: unit, size
    integer(kind=singleI), intent(out)  :: plane(size)
#ifdef NOSCRATCH
    character(len=80)::label
    character(len=4) ::proclabel
    write(proclabel,'(i4)')MyPE
    write(label,*)unit
    label = 'tmp_local/scratch'//trim(adjustl(proclabel))//'.'//trim(adjustl(label))
    open (unit,file=trim(label),status='old',form='unformatted')
    read (unit) plane
    close(unit)
#else
    rewind (unit)
    read   (unit) plane
#endif
  end subroutine read_scratch_i4

end module scratch_io
