!*****************************************************************
!*                       PINOCCHI0  V2.2                         *
!*  (PINpointing Orbit-Crossing Collapsed HIerarchical Objects)  *
!*****************************************************************
!
!This code was written by
!Pierluigi Monaco, Tom Theuns, Giuliano Taffoni
!Copyright (C) 2008
!
!web page: http//adlibitum.oats.inaf.it/monaco/Homepage/Pinocchio/
!
!This program is free software; you can redistribute it and/or modify
!it under the terms of the GNU General Public License as published by
!the Free Software Foundation; either version 2 of the License, or
!(at your option) any later version.
!
!This program is distributed in the hope that it will be useful,
!but WITHOUT ANY WARRANTY; without even the implied warranty of
!MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!GNU General Public License for more details.
!
!You should have received a copy of the GNU General Public License
!along with this program; if not, write to the Free Software
!Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

subroutine load_data ()

!

  use numbers
  use fragment_mpi
  use grid
  use scratch_io
  use runtime_parameters
  use fast_fmax_module

  implicit none
  integer(kind=singleI) :: iplane_local, iplane_global, n_read, read_proc
  integer(kind=singleI) :: ierr, next_proc, readPE, n_data, unit, left, right, &
        iplane_send, iplane_recv , imod, ioff, isend, i, j, iserver, ii, jj, kk
  logical, allocatable :: plane_to_read(:)
  character(len=2*string_length) :: filename, file, label
  character(len=4) :: extension
  integer(kind=doubleI) :: ntotal
  logical :: recover_flag


! master PE decides who is allowed to read
  if(MyPE == 0)then

! loop over server nodes, allowing them to read, until all planes are read
     allocate(plane_to_read(LGrid_global(3)))
     call trace_allocations(0,4.*float(Lgrid_global(3)),.false.,'plane_to_read')
     allocate(fast_fmax(Nserver,nbin))
     call trace_allocations(0,8*Nserver*nbin,.false.,'fast_fmax')
     fast_fmax = 0

     plane_to_read(:) = .true.
     read_proc        = 0
! test if any more planes need reading
     do while (any(plane_to_read))
! test whether read_proc has any more planes to read
        loop: do iplane_global=1, LGrid_global(3)
           if (plane_to_read(iplane_global) .and. myserver(iplane_global) == read_proc) then

! send message to this server to start reading, and wait for it to finish
              readPE = read_proc + NMaster
              call mpi_send(iplane_global,1,mpi_integer,readPE,readPE,mpi_comm_world,ierr)
              call mpi_recv(fast_fmax_plane,nbin,mpi_integer,readPE,readPE,mpi_comm_world,mpi_status,ierr)
              fast_fmax(read_proc+1,:) = fast_fmax(read_proc+1,:) + fast_fmax_plane
              call mpi_recv(next_largest,1,mpi_real,readPE,readPE,mpi_comm_world,mpi_status,ierr)
              largest_fmax=max(largest_fmax,next_largest)
!              if (read_proc==0) write (*,'(" Server 0 processing plane : ",i5," largest fmax found: ",f9.4)')iplane_global, next_largest
              if (read_proc==0) write (*,'(i5," of ",i5)')iplane_global,nplanes(1)

!!!              call mpi_recv(iplane_global,1,mpi_integer,readPE,readPE,mpi_comm_world,mpi_status,ierr)
!!!!              if (read_proc==0) write (*,'(" Server 0 processing plane : ",i5)')iplane_global, largest_fmax
!!!              if (read_proc==0) write (*,'(i5," of ",i5)')iplane_global,nplanes(1)

              plane_to_read(iplane_global) = .false.
              exit loop
           endif
        enddo loop
        read_proc = mod(read_proc+1,NServer)
     enddo
     deallocate(plane_to_read)
     call trace_allocations(0,-4.*float(Lgrid_global(3)),.false.,'plane_to_read')

! cumulates the fmax histogram
     open(10,file='cumulative_fmax.out',status='unknown')
     ntotal = 0
     do ibin=2,nbin
        fbin = exp(Flargest-ibin*binsize)
        do read_proc=1,Nserver
           fast_fmax(read_proc,ibin) = fast_fmax(read_proc,ibin)+fast_fmax(read_proc,ibin-1)
        enddo
        write(10,'(1x,i4,1x,f8.4,99(1x,i19))')ibin,fbin,fast_fmax(:,ibin)
     enddo
     close(10)
     write(*,*)'Largest fmax: ',largest_fmax
     ntotal = sum(fast_fmax(:,nbin))
     write(*,*)'Number of collapsed particles: ',ntotal

  endif

! only server nodes read
  if(my_type == server)then

#ifdef RECOVER
     unit = plane_start(GroupPE+1) + velx_offset + 1
     write(file,'(i4)')MyPE
     write(label,'(i9)')unit
     file='tmp_local/scratch'//trim(adjustl(file))//'.'//trim(adjustl(label))
!     write(*,*)'inquirying file '//trim(file)
     inquire(file=trim(file),exist=recover_flag)
!     write(*,*)'result: '//trim(file),recover_flag
     if (recover_flag.and.GroupPE==0) write(*,*)'scratch files found, will not be overwritten' 
#else
     recover_flag=.false.
#endif
     n_data = LGrid(1)*LGrid(2)
     allocate(plane_to_read(LGrid(3)))
     call trace_allocations(0,4.*float(Lgrid(3)),.false.,'plane_to_read')
     allocate(plane1D(Lgrid(1)*LGrid(2)),stat=ierr)
     call trace_allocations(ierr,4.*float(Lgrid(1)*Lgrid(2)),.false.,'plane1D')
     allocate(recvplane1D(Lgrid(1)*LGrid(2)),stat=ierr)
     call trace_allocations(ierr,4.*float(Lgrid(1)*Lgrid(2)),.false.,'recvplane1D')
     allocate(plane(Lgrid(1),Lgrid(2)),stat=ierr)
     call trace_allocations(ierr,4.*float(Lgrid(1)*Lgrid(2)),.false.,'plane')
     allocate(indexx(n_data),stat=ierr)
     call trace_allocations(ierr,4.*float(Lgrid(1)*Lgrid(2)),.false.,'indexx')
     plane_to_read(:) = .true.

     do while (any(plane_to_read))
 ! wait to get start sign from master
        readPE = GroupPE + NMaster
        call mpi_recv(iplane_global,1,mpi_integer,0,readPE,mpi_comm_world,mpi_status,ierr)
        iplane_local = iplane_global - plane_start(GroupPE+1)
        plane_to_read(iplane_local) = .false.

#ifdef ONE_PLANE_PER_FILE
        if (iplane_local<10) then
           write(extension,'("000",i1)')iplane_local
        else if (iplane_local<100) then
           write(extension,'("00",i2)')iplane_local
        else if (iplane_local<1000) then
           write(extension,'("0",i3)')iplane_local
        else
           write(extension,'(i4)')iplane_local
        endif
#else
        extension=''
#endif

! read velx and write local copy
        if (.not.recover_flag) then
           filename = trim(InputDir)//trim(Velx_File)//trim(extension)
           call read_plane (filename, iplane_global-1, Lgrid(1:2), plane)
#ifdef STORE_VEL
           do jj=1,Lgrid(2)
              do ii=1,Lgrid(1)
                 kk=ii+(jj-1)*Lgrid(1)
                 stored_vel(1,kk,iplane_local) = plane(ii,jj)
              enddo
           enddo
#else
           unit = iplane_global + velx_offset
           call write_scratch (unit, Lgrid(1:2), plane)
#endif
        endif

! read vely and write local copy
        if (.not.recover_flag) then
           filename = trim(InputDir)//trim(Vely_File)//trim(extension)
           call read_plane (filename, iplane_global-1, Lgrid(1:2), plane)
#ifdef STORE_VEL
           do jj=1,Lgrid(2)
              do ii=1,Lgrid(1)
                 kk=ii+(jj-1)*Lgrid(1)
                 stored_vel(2,kk,iplane_local) = plane(ii,jj)
              enddo
           enddo
#else
           unit = iplane_global + vely_offset
           call write_scratch (unit, Lgrid(1:2), plane)
#endif
        endif

! read velz and write local copy
        if (.not.recover_flag) then
           filename = trim(InputDir)//trim(Velz_File)//trim(extension)
           call read_plane (filename, iplane_global-1, Lgrid(1:2), plane)
#ifdef STORE_VEL
           do jj=1,Lgrid(2)
              do ii=1,Lgrid(1)
                 kk=ii+(jj-1)*Lgrid(1)
                 stored_vel(3,kk,iplane_local) = plane(ii,jj)
              enddo
           enddo
#else
           unit = iplane_global + velz_offset
           call write_scratch (unit, Lgrid(1:2), plane)
#endif
        endif

! read Fmax and write local copy
        if (.not.recover_flag) then
           filename = trim(InputDir)//trim(Fmax_File)//trim(extension)
           call read_plane (filename, iplane_global-1, Lgrid(1:2), plane)
#ifdef STORE_FMAX
           do jj=1,Lgrid(2)
              do ii=1,Lgrid(1)
                 kk=ii+(jj-1)*Lgrid(1)
                 stored_fmax(kk,iplane_local) = plane(ii,jj)
              enddo
           enddo
#else
           unit = iplane_global + fmax_offset
           call write_scratch (unit, Lgrid(1:2), plane)
#endif

! sort the plane and write copy of index
           fast_fmax_plane=0

           do j=1,Lgrid(2)
              do i=1,Lgrid(1)
                 plane1D(i+(j-1)*Lgrid(1))=-plane(i,j)

! constructs a histogram of Fmax values
                 if (plane(i,j)>=1.) then
                    largest_fmax = max(largest_fmax,plane(i,j))
                    ibin = (Flargest-log(plane(i,j)))/binsize+1
                    ibin = max(ibin,1)
                    ibin = min(ibin,nbin-1)
                    if (ibin>nbin) then
                       print *,MyPE
                       print *,'ERROR 1 in building Fmax histogram!',plane(i,j),ibin,Flargest,binsize
                       call abort('aborting...')
                    endif
                    fast_fmax_plane(ibin) = fast_fmax_plane(ibin) + 1
                 endif

              enddo
           enddo
        else
#ifdef STORE_FMAX
           plane1D = stored_fmax(:,iplane_local)
#else
           unit = iplane_global + fmax_offset
           call read_scratch (unit, n_data, plane1D)
#endif

! constructs a histogram of Fmax values
           fast_fmax_plane=0
           do i=1,n_data
              if (plane1D(i)>=1.) then
                 largest_fmax = max(largest_fmax,plane1D(i))
                 ibin = (Flargest-log(plane1D(i)))/binsize+1
                 ibin = max(ibin,1)
                 ibin = min(ibin,nbin-1)
                 if (ibin>nbin) then
                    print *,MyPE
                    print *,'ERROR 2 in building Fmax histogram!',plane1D(i),ibin,Flargest,binsize
                    call abort('aborting...')
                 endif
                 fast_fmax_plane(ibin) = fast_fmax_plane(ibin) + 1
              endif
           enddo

        endif

! send message back to master to indicate planes have been read
        call mpi_send(fast_fmax_plane,nbin,mpi_integer,0,readPE,mpi_comm_world,ierr)
        call mpi_send(largest_fmax,1,mpi_real,0,readPE,mpi_comm_world,ierr)

!!!        call mpi_send(iplane_global,1,mpi_integer,0,readPE,mpi_comm_world,ierr)


        
        if (.not.recover_flag) then
           call index_sort_r4(n_data, plane1D, indexx)
#ifdef STORE_FMAX
           stored_index(:,iplane_local) = indexx
#else
           unit = iplane_global + index_offset
           call write_scratch (unit, n_data, indexx)
#endif
        endif
     enddo

! each processor also makes a copy of the boundary planes either side
! left-hand boundary

!     write (*,*) ' this server: ',GroupPE, plane_start(GroupPE+1), nplanes(GroupPE+1)

     left  = mod( GroupPE - 1 + NServer, NServer)
     right = mod( GroupPE + 1 + NServer, NServer)

     do isend=1,3
        if(isend == 1)then
           ioff = velx_offset
        else if (isend == 2) then
           ioff = vely_offset
        else
           ioff = velz_offset
        endif

! send to left, receive from right
        iplane_send = plane_start(GroupPE+1) + 1
        iplane_recv = plane_start(right+1) + 1
#ifdef STORE_VEL
        plane1D = stored_vel(isend,:,1)
#else
        call read_scratch(iplane_send + ioff, n_data, plane1D)
#endif
        call mpi_sendrecv(plane1D, n_data, mpi_real, left, GroupPE &
             , recvplane1D, n_data, mpi_real, right, right, server_comm, mpi_status, ierr)
#ifdef STORE_VEL
        if (.not.recover_flag) stored_vel(isend,:,LGrid(3)+1) = recvplane1D
#else
        if (.not.recover_flag) call write_scratch(iplane_recv + ioff, n_data, recvplane1D)
#endif

! send to right, receive from left
        iplane_send = plane_start(GroupPE+1) + nplanes(GroupPE+1)
        iplane_recv = plane_start(left+1) + nplanes(left+1)
#ifdef STORE_VEL
        plane1D = stored_vel(isend,:,LGrid(3))
#else
        call read_scratch(iplane_send + ioff, n_data, plane1D)
#endif
        call mpi_sendrecv(plane1D, n_data, mpi_real, right, GroupPE &
             , recvplane1D, n_data, mpi_real, left, left, server_comm, mpi_status, ierr)
#ifdef STORE_VEL
        if (.not.recover_flag) stored_vel(isend,:,0) = recvplane1D
#else
        if (.not.recover_flag) call write_scratch(iplane_recv + ioff, n_data, recvplane1D)
#endif
     enddo

     deallocate(plane1D)
     call trace_allocations(0,-4.*float(Lgrid(1)*Lgrid(2)),.false.,'plane1D')
     deallocate(recvplane1D)
     call trace_allocations(0,-4.*float(Lgrid(1)*Lgrid(2)),.false.,'recvplane1D')
     deallocate(plane)
     call trace_allocations(0,-4.*float(Lgrid(1)*Lgrid(2)),.false.,'plane')
     deallocate(plane_to_read)
     call trace_allocations(0,-4.*float(Lgrid(3)),.false.,'plane_to_read')
     deallocate(indexx)
     call trace_allocations(0,-4.*float(Lgrid(1)*Lgrid(2)),.false.,'indexx')

  endif

  return
end subroutine load_data


character(len=4) function write_string(i)

  use numbers
  implicit none
  integer(kind=singleI),intent(in)::i

  write(write_string,'(i4)')i
  if (i<=9) then 
     write_string(1:3)='000'
  else if (i<=99) then
     write_string(1:2)='00'
  else if (i<=999) then
     write_string(1:1)='0'
  endif

  return
end function write_string


subroutine read_plane(filename, iplane, size, plane)

! reads one plane from the fmax output files

  use numbers
#ifdef HDF5
  use hdf5_wrapper
#endif
  use runtime_parameters, only : string_length
  use fragment_mpi, only : MyPE
  implicit none
  integer, intent(in) :: iplane, size(2)                   
  real(kind=singleR), intent(out) :: plane(size(1),size(2))
  character(len=2*string_length), intent(in) :: filename
  character(len=5)::label

#ifdef HDF5
  character(len=11)         :: field
  character(len=80)         :: varname
  integer(kind=singleI)     :: file_handle
  character(len=4),external :: write_string
  integer(kind=singleI)     :: rank
  integer(kind=singleI),dimension(2)::dims
#else
  character(len=80)::info
  integer(kind=singleI),parameter::unit=29
  integer(kind=singleI)::last_plane,Lx,Ly,Lz,i,ii
  logical::flag
#endif

#ifdef HDF5
  call hdf5_open_file(file_handle, filename)
  field    ='/Plane'//write_string(iplane)//'/'
  varname  = trim(field)//'Data'
  call hdf5_get_dimensions(file_handle, varname, rank, dims)
  call hdf5_read_data(file_handle,varname,plane)
  call hdf5_close_file(file_handle)
#else

#ifdef ONE_FILE_PER_PLANE
  if (iplane<10) then
     write(label,'(".000",i1)')iplane
  else if (iplane<100) then
     write(label,'(".00",i2)')iplane
  else if (iplane<1000) then
     write(label,'(".0",i3)')iplane
  else 
     write(label,'(".",i4)')iplane
  endif
  open(unit=unit,file=trim(filename)//label,status='old',form='unformatted')
  read (unit)
  read (unit)plane
  close(unit)
#else

  open(unit,file=trim(filename),status='unknown',form='unformatted')
  read(unit)Lx, Ly, Lz
  if (Lx/=size(1) .or. Ly/=size(2) ) then
     call message('WARNING: SEVERE PROBLEM')
     call message('Inconsistent grid data in fortran file')
     write (info,'(''Parameters'',5(1x,i9))')Lx,Ly,Lz
     call message(trim(info))
     call abort()
  endif

  do i=1,Lx
     read(unit)ii
     if (ii==iplane) then
        read (unit)plane
        exit
     else
        read (unit)
     endif
  enddo
#endif
#endif

  return
end subroutine read_plane


!#ifdef HDF5
!subroutine rw_header(action)
!  use hdf5_wrapper; use pinocchio_hdf5
!  implicit none
!  character(*), intent(in) :: action
!
!  character,parameter :: read='read',write='write',test='test'
!  character(len=80)::info
!
!  if(action .eq. write) then
!
!     call hdf5_create_group(file_handle,'/Header/')
!     call hdf5_write_data(file_handle,'Header/Omega0',Omega0)
!     call hdf5_write_data(file_handle,'Header/OmegaLambda',OmegaLambda)
!     call hdf5_write_data(file_handle,'Header/Hubble100',Hubble100)
!     call hdf5_write_data(file_handle,'Header/Sigma8',Sigma8)
!     call hdf5_write_data(file_handle,'Header/BoxSize',BoxSize)
!     call hdf5_write_data(file_handle,'Header/PowerSpectrum',PowerSpectrum)
!     call hdf5_write_data(file_handle,'Header/Mpart',Mpart)
!     call hdf5_write_data(file_handle,'Header/InterPartDist',InterPartDist)
!     call hdf5_write_data(file_handle,'Header/Lgrid_global',Lgrid_global)
!
!  elseif(action .eq. read) then
!
!     call hdf5_read_data(file_handle,'Header/Omega0',Omega0)
!     call hdf5_read_data(file_handle,'Header/OmegaLambda',OmegaLambda)
!     call hdf5_read_data(file_handle,'Header/Hubble100',Hubble100)
!     call hdf5_read_data(file_handle,'Header/Sigma8',Sigma8)
!     call hdf5_read_data(file_handle,'Header/BoxSize',BoxSize)
!     call hdf5_read_data(file_handle,'Header/PowerSpectrum',PowerSpectrum)
!     call hdf5_read_data(file_handle,'Header/Mpart',Mpart)
!     call hdf5_read_data(file_handle,'Header/InterPartDist',InterPartDist)
!     call hdf5_read_data(file_handle,'Header/Lgrid_global',Lgrid_global)
!
!  else
!     call abort_code(' invalid action in rw_grid')
!  endif
!end subroutine rw_header
!
!#endif


