!*****************************************************************
!*                       PINOCCHI0  V2.2                         *
!*  (PINpointing Orbit-Crossing Collapsed HIerarchical Objects)  *
!*****************************************************************
!
!This code was written by
!Pierluigi Monaco, Tom Theuns, Giuliano Taffoni
!Copyright (C) 2008
!
!web page: http//adlibitum.oats.inaf.it/monaco/Homepage/Pinocchio/
!
!This program is free software; you can redistribute it and/or modify
!it under the terms of the GNU General Public License as published by
!the Free Software Foundation; either version 2 of the License, or
!(at your option) any later version.
!
!This program is distributed in the hope that it will be useful,
!but WITHOUT ANY WARRANTY; without even the implied warranty of
!MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!GNU General Public License for more details.
!
!You should have received a copy of the GNU General Public License
!along with this program; if not, write to the Free Software
!Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Program Mass_Function

! Produces the mass function of DM halos after a pinocchio run

  use numbers
  use pinocchio_common
  Implicit none

  character(len=4)::redshift
  character(len=80)::filein,fileout
  real(kind=singleR)::final_z
  real(kind=singleR),allocatable::outputs(:)
  integer(kind=singleI)::Lgrid(3),nz,nout,iout,Lgrid1

  real(kind=singleR),external::an_h,aLambda
  logical::cubic=.false.,success=.false.,done,exist

! reads the run name from fragment.info
  open(21,file='fragment.info',status='old',err=57)
  read(21,'(a6)',err=57,end=57)run
  read(21,*)
  read(21,*)final_z
  read(21,*)nz

  nout=nz+1
  allocate (outputs(nout))
  read(21,*)outputs(1:nz)
  outputs(nout)=final_z

  close(21)
  write(*,*)'run: ',run
  write(*,*)'number of outputs: ',nout
  write(*,*)'redshifts: ',outputs

! Reads the $run.info file
  do while (.not.success)
     open(unit=10,file=run//'.info',status='old',err=56)
     read(10,*)BoxSize(1)
     read(10,*)Omega0
     read(10,*)OmegaLambda
     read(10,*)PowerSpectrum
     read(10,*)Sigma8
     read(10,*)Hubble100
     if (.not.cubic) then
! first: tries to read the three grid lengths (non-cubic case)
        read(10,*,err=50) Lgrid1
        success  = .true.
     else
! second: reads only one grid length and assumes a cubic box
        read(10,*,err=50) Lgrid1
        success      = .true.
        Lgrid = Lgrid1
     endif
     read(10,*)Mpart
50   close(10)
     cubic = .true.
  enddo
  print *,'Particle mass: ',Mpart
  print *,'Box size: ',BoxSize(1)


! setting cosmology
  call normalization()
  filter=2
  delta_c=1.686d0
  flag_ps=4

  select case (flag_ps)
  case (0)
     write(*,*)'Using Press & Schechter (1974) for the analytic mass function'
  case (1)
     write(*,*)'Using Sheth & Tormen (2001) for the analytic mass function'
  case (2)
     write(*,*)'Using Jenkins et al. (2001) for the analytic mass function'
  case (3)
     write(*,*)'Using Warren et al. (2006) for the analytic mass function'
  case (4)
     write(*,*)'Using Reed et al. (2007) for the analytic mass function'
  end select

  call grow_start()
  print *,'Check of sigma8:',sqrt(aLambda(8./Hubble100)),Sigma8

! Loop on the outputs
  do iout = 1,nout
     
     if (outputs(iout)<10) then
        write(redshift,'(f4.2)')outputs(iout)
     else
        write(redshift,'(f4.1)')outputs(iout)
     endif
     filein='frag.'//redshift//'.'//run//'.grp.out'
     fileout='frag.'//redshift//'.'//run//'.mf.par.out'

     Call MF(filein,fileout,BoxSize(1),Mpart,outputs(iout))

     print *,'Finished redshift '//redshift
  enddo

  stop

56 print *,'Error: file '//run//'.info does not exist'
  stop

57 print *,'Problem with file fragment.info'
  stop

  Contains

  Subroutine MF(in,out,box,Mpart,z)

    use numbers
    implicit none
    Character(len=80),Intent(in)::in,out
    real(kind=singleR),intent(in)::box,Mpart,z
    Integer(kind=singleI), Parameter::bins=40
    Integer(kind=singleI)::k,nr,p
    Integer(kind=singleI)::is,js,i
    Real(kind=singleR)::as,bs,cs,ds,es,fs,gs,hs,ls,ms
    Real(kind=singleR)::Delta,nM,lgnM,Vol,MMin,MMax,lgnMup,lgnMdn,dm,theo,Mass,LgM
    character(len=3)::label
    integer(kind=singleI)::halos(0:9)=0,total,ibin
    integer(kind=singleI),dimension(bins)::massf
    logical::flag,to_read(0:9)
    real(kind=singleR),external::an_h

    print *
    print *,'Computing mass function at z =',z

    Vol=box*box*box
    MMax=16.5
    MMin=Mpart+1.
    Delta=(MMax-MMin)/real(bins)

    do i=0,9
       write(label,'(".0",i1)')i
       inquire(file=trim(in)//label,exist=flag)
       to_read(i)=flag
       if (flag) then
          print *,'Opening file '//trim(in)//label
#ifdef ASCII_OUT
          Open(10,file=trim(in)//label,status='old')
#else
          Open(10,file=trim(in)//label,status='old',form='unformatted')
#endif
          nr=0
          Do While (.True.)
#ifdef ASCII_OUT
             Read(10,*,end=30,err=30) is !,as,bs,cs,ds,es,fs,gs,hs,ls,ms,js
#else
             Read(10,end=30,err=30) is !,as,bs,cs,ds,es,fs,gs,hs,ls,ms,js
#endif
             nr=nr+1
          End Do
30        Close(10)
          print *,'found ',nr,' halos'
          if (nr==0) to_read(i)=.false.
          halos(i)=nr
          print *,'Closed file '//trim(in)//label
       endif
    enddo

    nr=sum(halos)
    print *,'Found ',nr,' groups in all files'
    
    massf=0
    total=0
    do i=0,9
       if (to_read(i)) then
          write(label,'(".0",i1)')i
          print *,'Opening again '//trim(in)//label
#ifdef ASCII_OUT
          Open(10,file=trim(in)//label,status='old')
#else
          Open(10,file=trim(in)//label,status='old',form='unformatted')
#endif
          Do k=total+1,total+halos(i)
#ifdef ASCII_OUT
             Read(10,*,end=40) is,LgM  !is,as,bs,cs,ds,es,fs,gs,hs,ls,ms,js
#else
             Read(10,end=40) is,LgM    !is,as,bs,cs,ds,es,fs,gs,hs,ls,ms,js
#endif
             ibin=(LgM-MMin)/Delta+1
             if (ibin>=1 .and. ibin<=bins) massf(ibin)=massf(ibin)+1
          End Do
40        Close(10)
          print *,'Closing '//trim(in)//label
          total=total+halos(i)
       endif
    enddo

    print *,'Opening '//trim(out)
    Open(50,file=trim(out),status='unknown')

    Do p=1,bins
       dm=10.**(MMin+p*Delta)-10.**(MMin+(p-1)*Delta)
       lgnM=float(massf(p))/Vol/dm
       lgnMup=log10(max(lgnM+sqrt(float(massf(p)))/Vol/dm,tiny(1.)))
       lgnMdn=log10(max(lgnM-sqrt(float(massf(p)))/Vol/dm,tiny(1.)))
       lgnM=log10(max(lgnM,tiny(1.)))
       Mass=MMin+(p-0.5)*Delta
       theo=log10(an_h(10.**Mass,z))

       write(50,'(4(1x,f15.8),3x,i10,3x,f12.6)',err=1)Mass,lgnM,lgnMup,lgnMdn,massf(p),theo

    End Do
    Close(50)

    return
1   print *,'Something went wrong...'
    stop

  End Subroutine MF


End Program Mass_Function


subroutine abort_code(messg)
  implicit none
  character(*) :: messg
  stop
end subroutine abort_code
