!*****************************************************************
!*                       PINOCCHI0  V2.2                         *
!*  (PINpointing Orbit-Crossing Collapsed HIerarchical Objects)  *
!*****************************************************************
!
!This code was written by
!Pierluigi Monaco, Tom Theuns, Giuliano Taffoni
!Copyright (C) 2008
!
!web page: http//adlibitum.oats.inaf.it/monaco/Homepage/Pinocchio/
!
!This program is free software; you can redistribute it and/or modify
!it under the terms of the GNU General Public License as published by
!the Free Software Foundation; either version 2 of the License, or
!(at your option) any later version.
!
!This program is distributed in the hope that it will be useful,
!but WITHOUT ANY WARRANTY; without even the implied warranty of
!MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!GNU General Public License for more details.
!
!You should have received a copy of the GNU General Public License
!along with this program; if not, write to the Free Software
!Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


! This file contains the main program of the fragment-2.2 code

program fragment

! fragment-2.2 version
! - serial version 
! - bug fixes with respect to 1.0 version
! - reads the fmax-1.0 and fmax-2.2 output (ascii, binary or HDF5)
! - handles non-cubic boxes with non-power-of-two grid points
! - coming soon: parallel version of fragment
!
!     ========================================================================
!     Summary: Compute merger history for dark matter haloes in a CDM
!              Cosmology, for a given realisation of a linear density field,
!              generated or specified on a rectangular grid
!     
!              fmax-2.2:    compute inverse collapse time for each grid point
!              fragment-2.2:  use output from fmax-2.2 to determine merger history
!
!     Implementation:
!     This is a fortran-90 translation of the original fragment code
!
!     Detail:
!     See the DOCUMENTATION file for all details.
!
!     Input is read from file 'fragment.info'

  use run_flags
  implicit none
  logical::first

  first=.true.
  do while (.true.)

! Initializes everything
     call init_fragment(first)

! Generates the group catalogue at Flim
     call groups()

! Writes out the results
     call out_write(zfin)

! Mass function
     call mf(zfin)

! Decides whether to stop or run again
     call next_step(first)

  enddo
end program fragment



subroutine next_step(first)

  use pinocchio_common ; use run_flags
  implicit none
  logical,intent(inout)::first

! decides whether to stop or run again

  close(31)
  if (stop_flag .eq. 1) stop

! In case of re-run, reads new param values from fragment.info
  print *,'Save fragment.info file and press return'
  read (*,'(a)')

! on re-run it will not sort the Fmax vector again
  first=.false.

  return
end subroutine next_step



subroutine init_fragment(first)

! Initializes all quantities for the run

  use pinocchio_common ; use run_flags !; use run_module
  use grid
  use matrices ; use link ; use merger ; use group_vectors 
  use virial_params
#ifdef PAIRS
  use pairs
#endif

  implicit none

  logical,intent(in)::first

  integer(kind=singleI)::Lgrid1,i,ierr
  real(kind=singleR)    :: rngrpmax
  real(kind=singleR)::var_grid,r,v,vt,Mtot,Mstar
  real(kind=singleR),external::grow,aM_star,an_h,aLambda            , Variance
  logical::cubic=.false.,success=.false.,done,partitioning,exist
  character(len=6)::partition_label
  character(len=30)::partition_file



! all this is done only on first run
  if (first) then

! outputs messages to explicitate pre-compiler options
#ifdef SPIN
     print *,'This version is compiled WITH the SPIN option'
     seed=-33343
#else
     print *,'This version is compiled WITHOUT the SPIN option'
#endif
#ifdef PAIRS
     print *,'This version is compiled WITH the PAIRS option'
#else
     print *,'This version is compiled WITHOUT the PAIRS option'
#endif

!     print *,'Label for partitioning (return for none)?'
!     read '(a)',partition_label
     partition_label=''
     if (trim(partition_label)=='') then
        print *,'No partitioning, the whole box will be considered'
        partitioning=.false.
     else
        print *,'The box will be partitioned'
        partitioning=.true.
     endif

! reads the run name from fragment.info
     open(21,file='fragment.info',status='old',err=57)
     read(21,'(a6)',err=55,end=55)run
     close(21)
     print *,'run: ',run

! Reads the $run.info file
     do while (.not.success)
        open(unit=10,file=run//'.info',status='old',err=56)
        read(10,*)BoxSize_global(1)
        read(10,*)Omega0
        read(10,*)OmegaLambda
        read(10,*)PowerSpectrum
        read(10,*)Sigma8
        read(10,*)Hubble100
        if (.not.cubic) then
! first: tries to read the three grid lengths (non-cubic case)
           read(10,*,err=50) Lgrid_global
           success  = .true.
        else
! second: reads only one grid length and assumes a cubic box
           read(10,*,err=50) Lgrid1
           success      = .true.
           Lgrid_global = Lgrid1
        endif
50      close(10)
        if(success) then
           InterPartDist=BoxSize_global(1)/float(Lgrid_global(1))
           BoxSize_global = InterPartDist*Lgrid_global
        endif
        cubic = .true.
     enddo

! Reads the partition of the box to read
     if (partitioning) then
        open(21,file='partitions.'//partition_label,status='old',err=58)
        read(21,*,end=58,err=58)Lgrid
        read(21,*,end=58,err=58)Lstart
!        read(21,'(a6)',end=58,err=58)partition_label
        close(21)
        print *,'partitions.'//partition_label//' file read'
     else
        Lgrid=0 ; Lstart=0
     endif

     do i=1,3
        if (Lgrid(i) == 0 .or. Lgrid(i)==Lgrid_global(i)) then
           Lgrid(i)    = Lgrid_global(i)
           if (Lstart(i)==0) then
! In this case the whole box with pbc is to be used on direction i
              Lstart(i)   = 1
              pbc_flag(i) = .true.
           else if (Lstart(i)>0 .and. Lstart(i)<=Lgrid_global(i)) then
! In this case PBC are not used on direction i
              pbc_flag(i) = .false.
           else
! In this case there's something wrong with Lstart
              print *,'Default Lgrid but wrong Lstart on dim.',i
              stop
           endif
        else if (Lgrid(i)>0 .and. Lgrid(i)<Lgrid_global(i)) then
           if (Lstart(i)>0 .and. Lstart(i)<=Lgrid_global(i)) then
! In this case only a partition of the box is analyzed, 
! and PBCs on direction i are switched off
              pbc_flag(i) = .false.
           else
! In this case there's something wrong with Lstart
              print *,'Wrong Lstart  on dim. ',i
              stop
           endif
        else
! In this case there's something wrong with Lgrid
           print *,'Wrong Lgrid  on dim. ',i
           stop
        endif
     enddo

! setting cosmology
     filter=2
     delta_c=1.686
     flag_ps=4

     select case (flag_ps)
     case (0)
        write(*,*)'Using Press & Schechter (1974) for the analytic mass function'
     case (1)
        write(*,*)'Using Sheth & Tormen (2001) for the analytic mass function'
     case (2)
        write(*,*)'Using Jenkins et al. (2001) for the analytic mass function'
     case (3)
        write(*,*)'Using Warren et al. (2006) for the analytic mass function'
     case (4)
        write(*,*)'Using Reed et al. (2007) for the analytic mass function'
     end select

     call normalization()
     write(*,'("Check of normalization: ",2(1x,f7.2))')sqrt(aLambda(8./Hubble100)),Sigma8

     BoxSize = InterPartDist*Lgrid

! Number of particles and maximum number of groups
     Npart = product(Lgrid)
     
!     ngrpmax = max(Npart/100,100)

! computes the particle mass
     Mtot   = 2.778e11*Hubble100**2*Omega0*product(BoxSize)
     Mpart  = Mtot/float(Npart)
     Mstar  = aM_star(zfin)
     rngrpmax = Mpart*an_h(Mpart,0.)*product(BoxSize)
     ngrpmax = int(rngrpmax)*10

     print *,'Global box:'
     print *,'Grid size:        ',Lgrid_global
     print *,'No. of particles: ',product(Lgrid_global)
     print *,'Box size:         ',BoxSize_global
     print *
     print *,'Local box:'
     print *,'Grid size:        ',Lgrid
     print *,'No. of particles: ',Npart
     print *,'First particle:   ',Lstart
     print *,'PBC flags:        ',pbc_flag
     print *,'Box size:         ',BoxSize
     print *,'Expected ngrpmax: ',ngrpmax
     print *,'Mpart:            ',Mpart
     print *,'M* at zfin:       ',Mstar

! reads the grid-scale variance from the run.var file
     open(10,file=run//'.var',status='old')
     r=1.e10
     do while (r>0.)
        read(10,*,end=53)r,v,vt
     enddo
     var_grid=vt
! allocates matrices
     allocate (Fmax     (Npart),stat=ierr)
     if (ierr>0) then ; stop 'could not allocate Fmax     ' ; else ; print *,'allocated Fmax' ; endif
     allocate (vx       (Npart),stat=ierr)
     if (ierr>0) then ; stop 'could not allocate vx       ' ; else ; print *,'allocated vx' ; endif
     allocate (vy       (Npart),stat=ierr)
     if (ierr>0) then ; stop 'could not allocate vy       ' ; else ; print *,'allocated vy' ; endif
     allocate (vz       (Npart),stat=ierr)
     if (ierr>0) then ; stop 'could not allocate vz       ' ; else ; print *,'allocated vz' ; endif
     allocate (indices  (Npart),stat=ierr)
     if (ierr>0) then ; stop 'could not allocate indices  ' ; else ; print *,'allocated indices' ; endif
     allocate (group_ID (Npart),stat=ierr)
     if (ierr>0) then ; stop 'could not allocate group_ID ' ; else ; print *,'allocated group_ID' ; endif
     allocate (link_list(Npart),stat=ierr)
     if (ierr>0) then ; stop 'could not allocate link_list' ; else ; print *,'allocated link_list' ; endif

     allocate (mass   (  ngrpmax))
     allocate (peak   (3,ngrpmax))
     allocate (centres(3,ngrpmax))
     allocate (velgrp (3,ngrpmax))
#ifdef SPIN
     allocate (spin   (3,ngrpmax))
#endif
     allocate (gr_ll         (ngrpmax))
     allocate (Mcentr        (ngrpmax))
     allocate (halo_app      (ngrpmax))
     allocate (mass_at_merger(ngrpmax))
     allocate (merged_with   (ngrpmax))
     allocate (majorz        (ngrpmax))
     allocate (t_acc         (ngrpmax))
     allocate (t_appear      (ngrpmax))
     allocate (point         (ngrpmax))
     allocate (bottom        (ngrpmax))

     print *,'Allocations done'

! Reads the Fmax and vx-y-z matrices
     call read_Fmax()

! Sorts the Fmax matrix
     call index_sort(Npart,Fmax,indices)
     print *,'Sorting done'

  endif

  done=.false.
  do while (.not.done)

! Opens fragment.info again and reads the kernel file name
     open(21,file='fragment.info',status='old')
     read(21,*)
     read(21,'(a6)',end=22)kernel
     if (partitioning) kernel=partition_label
     print *,'kernel for output files: '//kernel

! Opens the frag.$kernel.out file
     open(31,file='frag.'//kernel//'.out',status='unknown')
     print *,'Parameters and infos will be written in the file frag.'//kernel//'.out'
     write(31,'(a,3(f10.4))')'BoxSize_global = ',BoxSize_global
     write(31,'(a,f10.4)')   'Omega0         = ',Omega0
     write(31,'(a,f10.4)')   'OmegaLambda    = ',OmegaLambda
     write(31,'(a,f10.4)')   'PowerSpectrum  = ',PowerSpectrum
     write(31,'(a,f10.4)')   'Sigma8         = ',Sigma8
     write(31,'(a,f10.4)')   'Hubble100      = ',Hubble100

     write(31,'(''Original field has dimension '',3(1x,i6))')Lgrid_global
     write(31,'(''Fragment will work on a grid '',3(1x,i6))')Lgrid

! final redshift and inverse growing mode
     read(21,*,end=22)zfin
     write(31,'(a7,10(3x,f6.3))')'z    = ',zfin
     Flim=1./grow(zfin)
     write(31,'(a7,10(3x,f6.3))')'Flim = ',Flim

! intermediate outputs
     read(21,*,end=22)nout
     write(31,'(a7,3x,i2)')'nout = ',nout
     if (allocated(zout)) then
        deallocate(zout)
        deallocate(fout)
     endif
     allocate (zout(nout+1)) ; allocate (fout(nout+1))
     read(21,*,end=22)zout(1:nout)
     zout(nout+1)=zfin

! sorts the output redshifts in decreasing order
     call sort(nout+1,zout)

     do i=1,nout+1
        fout(i)=1./grow(zout(i))
     enddo
     write(31,'(''Sorted output redshifts:'')')
     write(31,'(a7,10(3x,f6.3))')'zout = ',(zout(i),i=1,nout)
     write(31,'(a7,10(3x,f6.3))')'fout = ',(fout(i),i=1,nout)

! sets parameters
!     0: default parameters
!     1: specify f_a
!     2: specify all parameters
     read(21,*,end=22)param_flag
     if (param_flag.eq.0) then
        write(31,*)'Parameters are default'
        f_m  = 0.35
        f_rm = 0.70
        f_s  = 0.06
        f_a  = 0.22 + (log10(var_grid/InterPartDist) - 0.36) * 0.11
        f_a  = min (f_a,0.26)
        f_a  = max (f_a,0.22)
        f_ra = 0.40 - (f_a - 0.22) * 3.5
!        f_ra = 0.37 - (f_a - 0.22) * 3.5
!        limit_var = 0.2 !0.011*var_grid +0.1
     else if (param_flag.eq.1) then
        write(31,*)'f_a is specified, other parameters are default'
        read(21,*,end=22)f_a
        f_m  = 0.35
        f_rm = 0.70
        f_s  = 0.06
        f_ra = 0.40 - (f_a - 0.22) * 3.5
     else if (param_flag.eq.2) then
        write(31,*)'All parameters are specified'
        read(21,*,end=22)f_m,f_rm,f_a,f_ra,f_s
     else
        print *,'Incorrect param_flag'
        goto 22
     endif
     espo=1.7
     write(31,'(a7,10(3x,f6.3))')'f_m  = ',f_m
     write(31,'(a7,10(3x,f6.3))')'f_a  = ',f_a
     write(31,'(a7,10(3x,f6.3))')'f_rm = ',f_rm
     write(31,'(a7,10(3x,f6.3))')'f_ra = ',f_ra
     write(31,'(a7,10(3x,f6.3))')'f_s  = ',f_s
     write(31,'(a7,10(3x,f6.3))')'sig2 = ',var_grid/InterPartDist

!other flags
     read(21,*,end=22)out_flag
     write(31,'(a7,3x,i2)')'outfl= ',out_flag
     read(21,*,end=22)stop_flag
     write(31,'(a7,3x,i2)')'stop = ',stop_flag

! endo of reading of fragment.info
     close(21)
     done=.true.
     exit
! in case of error
22   print *,'Error in fragment.info'
     print *,'Please correct fragment.info and press return'
     read (*,'(a)')
     close(21)
     close(31)
     cycle

! end of reading - fragment.info cycle
  enddo

  write(31,*)'Mpart    = ',Mpart
  write(31,*)'Mtot     = ',Mtot
  write(31,*)'M*       = ',Mstar
  write(31,*)'Mtot/M*  = ',Mtot/Mstar
  write(31,*)'M*/Mpart = ',Mstar/Mpart

#ifdef PAIRS
! Reads info on pair checking from file
  open (22,file='fragment.pair',status='unknown',err=54)
  read (22,*,end=54)mmin_pair
  read (22,*,end=54)Fstart
  read (22,*,end=54)deltaF
  read (22,*,end=54)njmax
  close(22)
  write(31,*)'Read check-pair parameters from file fragment.pair'
  write(31,*)'F_start = ',Fstart
  write(31,*)'delta_F = ',deltaF
  write(31,*)'minimum mass = ',mmin_pair
  write(31,*)'check done after each ',njmax,' steps'
  if (mmin_pair.lt.MASSMIN) then
     write(31,*)'WARNING: mmin_pair < MASSMIN'
     write(31,*)'Forcing mmin_pair=MASSMIN'
     mmin_pair=MASSMIN
  endif
#endif

! bye!
  return

! other error messages
53 print *,run//'.var file not existent or corrupted'
  stop

54 print *,'Please provide a correct fragment.pair file'
  print *,'  or compile fragment WITHOUT the -DPAIRS option'
  stop

55 print *,'Cannot read run in fragment.info'
  stop

56 print *,'File '//run//'.info does not exist'
  stop

57 print *,'File fragment.info does not exist'
  stop

58 print *,'Error in reading partitioning file'
  stop

end subroutine init_fragment



subroutine read_Fmax()

! Reads the Fmax and vmax matrices

  use pinocchio_common ; use grid ; use matrices ; use run_flags !; use run_module

  implicit none

! local variables
  character(len=30)::fF,fx,fy,fz
  integer(kind=singleI) i
  real(kind=singleR)::c
  logical::ex1,ex2

  fF=dir//'Fmax.ell.'//run//'.out'
  inquire(file=fF,exist=ex1)
  fF=dir//'Fmax.'//run//'.d'
  inquire(file=fF,exist=ex2)

  if (ex1.or.ex2) then
     if (ex2) then
        print *,'Found fmax-2.* output files'
        ex1=.false.
     else
        print *,'Found fmax-1.0 output files'

! No sectioning for pinocchio-1.0
        if (Lgrid(1) /= Lgrid_global(1) .or. &
            Lgrid(2) /= Lgrid_global(2) .or. &
            Lgrid(3) /= Lgrid_global(3)) then

           print *,'Warning: no sectioning in this case'
           print *,'Please trasform 1.0 to 2.* files'
           stop

        endif

     endif
  else
     print *,'Error in '//dir
     print *,'File Fmax is not present'
     stop
  endif

  if (ex1) then
     fF=dir//'Fmax.ell.'//run//'.out'
     fx=dir//'vmxx.ell.'//run//'.out'
     fy=dir//'vmxy.ell.'//run//'.out'
     fz=dir//'vmxz.ell.'//run//'.out'
  else
     fF=dir//'Fmax.'//run//'.d'
     fx=dir//'Velx.'//run//'.d'
     fy=dir//'Vely.'//run//'.d'
     fz=dir//'Velz.'//run//'.d'
  endif

  call read_field(fF,ex2,Fmax)
  call read_field(fx,ex2,vx  )
  call read_field(fy,ex2,vy  )
  call read_field(fz,ex2,vz  )

! Displacements are rescaled into grid units
  if (ex1) then
     c=1./InterPartDist
     do i=1,Npart
        vx(i)=vx(i)*c
        vy(i)=vy(i)*c
        vz(i)=vz(i)*c
     enddo
  endif

  return
end subroutine read_Fmax


subroutine groups()

! Fragment the collapsed medium into groups

  use pinocchio_common ; use grid ; use run_flags
  use merger ; use link ; use group_vectors
  use matrices 
#ifdef PAIRS
  use pairs
#endif

  implicit none

  integer(kind=singleI),parameter::nv=6
  integer(kind=singleI),dimension(nv,nv)::merge
  integer(kind=singleI),dimension(nv)::neigh
  integer(kind=singleI),dimension(3,nv)::fil_list
  integer(kind=singleI),dimension(6)::nm
  integer(kind=singleI)::NMERG,iout,Lgrid12,nn,i,j,k,mmax,ifil,indx,neigrp
  integer(kind=singleI)::igrp,ia,iz,in,kk,i1,i2,j1,j2,k1,k2,accrg
  integer(kind=singleI)::merge_counter,nmeff,ig3,small,large
  integer(kind=singleI)::nf,to_group,accgrp,ig1,ig2
  integer(kind=singleI)::ngr_ex,accrflag,nstep,nmerge,nstep_p,perc
  integer(kind=singleI)::nwei1,nwei2
  real(kind=singleR)::ratio,best_ratio,d2,r2
  logical::mask(-1:1,-1:1,-1:1)
  logical::merge_flag
  integer(kind=singleI),external::ipbc
!  real(kind=singleR),external::inverse_grow_s
  real(kind=singleR),external::inverse_grow

! Initializations

  NMERG=0                   ! NUMBER OF MAJOR MERGER EVENTS
  ngroups=1                 ! number of groups + filaments
  accrg=0                   ! number of accreting events on groups
  nwei1=0                   ! number of points not accreted at merger
                            !   at first attempt
  nwei2=0                   ! number of points never accreted at merger
  Lgrid12=Lgrid(1)*Lgrid(2)
  iout=1

#ifdef PAIRS
  Fold=Fmax(indices(Npart-1000))  ! first cleaning of pair list
  nptm=0                    ! number of pairs of groups still to merge
  njump=0
#endif

  mass=0     ; point=0     ; bottom=0
  GR_LL=-1   ; HALO_APP=-1 ; MCENTR=-1
  MASS_AT_MERGER=0  ;  MERGED_WITH=-1
  MAJORZ=-1. ; T_ACC=-1.   ; T_APPEAR=-1.
  centres=0. ; velgrp=0.   
  group_ID=0 ; link_list=0
  nm=0

#ifdef SPIN
  spin=0.
#endif

  point(1)=Npart+1        ! filaments are not grouped!
  bottom(1)=Npart+1

! This defines the number of neighbours considered

  if (nv.eq.26) then
! In this case all the 26 neighbouring points are considered
! (the central one is obviously excluded)
     mask=.true.
     mask(0,0,0)=.false.

  else if (nv.eq.6) then
! In this case considers only 6 neighbouring points
     mask=.false.
     mask(-1, 0, 0) =.true.
     mask( 1, 0, 0) =.true.
     mask( 0,-1, 0) =.true.
     mask( 0, 1, 0) =.true.
     mask( 0, 0,-1) =.true.
     mask( 0, 0, 1) =.true.
  else
     stop 'wrong nv'
  endif


! ********************************************
  print *,'Starting the fragmentation process'
! ********************************************

! Calculates the number of steps required

  iz=Npart
  
  do while (Fmax(indices(iz)) >= Flim)
     iz=iz-1
  enddo
  nstep=Npart-iz
  print *,'Number of steps required: ',nstep
  write(31,*)'Number of steps required: ',nstep
  nstep_p=nstep/20

! ************************************************************************
! START OF THE CYCLE ON POINTS
! ************************************************************************
  do iz=Npart,Npart-nstep+1,-1

! Outputs if relevant
     if (Fmax(indices(iz)).lt.fout(iout)) then
        call out_write(zout(iout))
        call mf(zout(iout))
        iout=iout+1
     endif

! Fraction of steps done
     k=Npart-iz+1
     perc=k/(nstep_p)*5
     if (mod(k,nstep_p).eq.0) then
        print '(''*** '',i3,''% Done, F ='',f6.2,'',  z = '',f6.2)' &
             ,perc,Fmax(indices(iz)),inverse_grow(1./Fmax(indices(iz)))
     endif

! More initializations

     neigrp=0               ! number of neighbouring groups
     nf=0                   ! number of neighbouring filament points
     accrflag=0             ! if =1 all the neighbouring filaments
                            !   are accreted 
     neigh=0                ! neighbours

! grid coordinates from the indices

     k=1+(indices(iz)-1)/Lgrid12
     kk=indices(iz)-1-(k-1)*Lgrid12
     j=1+kk/Lgrid(1)
     i=1+(kk-(j-1)*Lgrid(1))

! skips if the point is at the border and PBCs are swithced off

     if (.not.pbc_flag(1) .and. (i==1 .or. i==Lgrid(1))) cycle
     if (.not.pbc_flag(2) .and. (j==1 .or. j==Lgrid(2))) cycle
     if (.not.pbc_flag(3) .and. (k==1 .or. k==Lgrid(3))) cycle

! List of neighbouring groups
     nn=0
     do k1=-1,+1
        do j1=-1,+1
           do i1=-1,+1
              if (mask(i1,j1,k1)) then
                 i2=ipbc(i+i1,Lgrid(1))
                 j2=ipbc(j+j1,Lgrid(2))
                 k2=ipbc(k+k1,Lgrid(3))
                 nn=nn+1
                 neigh(nn)=group_ID(i2+(j2-1)*Lgrid(1)+(k2-1)*Lgrid12)
! list of touching points that are in filaments (not counted as groups)
! NB: IT ASSUMES THAT GROUP 1 CONTAINS FILAMENTS
                 if (neigh(nn)<=1 .and. neigh(nn)/=0) then
                    nf=nf+1
                    neigh(nn)=0
                    fil_list(1,nf)=i2
                    fil_list(2,nf)=j2
                    fil_list(3,nf)=k2
                 endif
              endif
           enddo
        enddo
     enddo

! Cleans the list of neighbouring groups
     call clean_list(nv,neigh)

! Number of neighbouring groups
     neigrp=0
     do while (neigh(neigrp+1)>1) 
        neigrp=neigrp+1
        if (neigrp==nv) exit
     enddo


! Is the point isolated?

     if (neigrp==0 .and. nf==0) then
! **********************************************************************
!                       FIRST CASE: PEAK
! **********************************************************************

! New group
        ngroups=ngroups+1
        if (ngroups > ngrpmax) then
!           print *,'Increase ngrpmax!',ngroups,ngrpmax
!           stop
           
! Increases ngrpmax by 10 per cent
           call increase_ngrpmax()

        endif
        centres(1,ngroups)=i-.5
        centres(2,ngroups)=j-.5
        centres(3,ngroups)=k-.5
        velgrp(1,ngroups)=vx(indices(iz))
        velgrp(2,ngroups)=vy(indices(iz))
        velgrp(3,ngroups)=vz(indices(iz))
        mass(ngroups)=1
! NB qui va controllato se ci sono le PBC
        peak(1,ngroups)=ipbc(i+Lstart(1)-1,Lgrid_global(1))
        peak(2,ngroups)=ipbc(j+Lstart(2)-1,Lgrid_global(2))
        peak(3,ngroups)=ipbc(k+Lstart(3)-1,Lgrid_global(3))
        group_ID(indices(iz))=ngroups
        point(ngroups)=indices(iz)
        bottom(ngroups)=indices(iz)
        link_list(indices(iz))=indices(iz)
        GR_LL(NGROUPS)=group_ID(indices(iz))
        HALO_APP(NGROUPS)=NGROUPS

     else if (neigrp.eq.1) then
! **********************************************************************
!                       SECOND CASE: 1 GROUP
! **********************************************************************

! if the points touches only one group, check whether to accrete the 
! point on it

        call accr_cond(i,j,k &
             ,vx(indices(iz)),vy(indices(iz)),vz(indices(iz)) &
             ,Fmax(indices(iz)),neigh(1),d2,r2)
        if (d2.lt.r2) then

! *********************
! accretion on a group!
! *********************

           accrg=accrg+1
           accrflag=1
           to_group=neigh(1)
           call accretion(to_group,i,j,k,indices(iz),Fmax(indices(iz)))
        else

! *********
! filament!
! *********

           mass(1)=mass(1)+1
           group_ID(indices(iz))=1
           link_list(indices(iz))=indices(iz)
        endif

     else if (neigrp.gt.1) then
! **********************************************************************
!                       THIRD CASE: >1 GROUP
! **********************************************************************

! In this case the point touches more than one group

! *********************
! accretion on a group?
! *********************

        best_ratio=(10.*Lgrid(1))**2
        accgrp=0
        do ig1=1,neigrp
           call accr_cond(i,j,k &
                ,vx(indices(iz)),vy(indices(iz)),vz(indices(iz)) &
                ,Fmax(indices(iz)),neigh(ig1),d2,r2)
           ratio=d2/r2
           if (ratio.lt.1..and.ratio.lt.best_ratio) then
              best_ratio=ratio
              accgrp=ig1
           endif
        enddo
        if (accgrp.gt.0) then
           accrg=accrg+1
           accrflag=1
           to_group=neigh(accgrp)
           call accretion(neigh(accgrp),i,j,k,indices(iz),Fmax(indices(iz)))
        endif

! Then checks whether the groups must be merged together

        nmerge=0
        do ig1=1,neigrp
           do ig2=1,ig1-1
              merge(ig1,ig2)=0
              call merge_cond(Fmax(indices(iz)),neigh(ig1),neigh(ig2), &
                   merge_flag)
              if (merge_flag) then
                 merge(ig1,ig2)=1
                 nmerge=nmerge+1
#ifdef PAIRS
              else if (mass(neigh(ig1)).gt.mmin_pair.and &
                   .mass(neigh(ig2)).gt.mmin_pair) then

! If the merging condition is not satisfied, then records the pair
                 nptm=nptm+1
                 if (nptm.gt.npair) stop 'Increase npair'
                 pairs(1,nptm)=min(neigh(ig1),neigh(ig2))
                 pairs(2,nptm)=max(neigh(ig1),neigh(ig2))
#endif
              endif
           enddo
        enddo

        if (nmerge.gt.1.and.accgrp.eq.0) nwei1=nwei1+1

! ******************
! merging of groups!
! ******************

! The group number of the largest group is preserved

        if (nmerge.gt.0) then
           nmeff=0
           do ig1=1,neigrp
              do ig2=1,ig1-1
                 if (merge(ig1,ig2).eq.1.and.neigh(ig1).ne.neigh(ig2)) then
                    if (mass(neigh(ig1)).gt.mass(neigh(ig2))) then
                       call merge_groups(neigh(ig1),neigh(ig2),Fmax(indices(iz)))
                       large=neigh(ig1)
                       small=neigh(ig2)
                       if (to_group.eq.small) to_group=large
                       do ig3=1,neigrp
                          if (neigh(ig3).eq.small) neigh(ig3)=large
                       enddo
                       nmeff=nmeff+1
                    else
                       call merge_groups(neigh(ig2),neigh(ig1),Fmax(indices(iz)))
                       small=neigh(ig1)
                       large=neigh(ig2)
                       if (to_group.eq.small) to_group=large
                       do ig3=1,neigrp
                          if (neigh(ig3).eq.small) neigh(ig3)=large
                       enddo
                       nmeff=nmeff+1
                    endif
                 endif
              enddo
           enddo
           if (nmeff>0) nm(nmeff)=nm(nmeff)+1
        endif

! If relevant, it tries again to accrete the point

        if (accgrp.eq.0) then
           call clean_list(nv,neigh)

           neigrp =0
           do while(neigh(neigrp+1).gt.1)
              neigrp = neigrp + 1
              if (neigrp==nv) exit
           enddo

           accgrp=0
           best_ratio=(10.*Lgrid(1))**2
           do ig1=1,neigrp
              call accr_cond(i,j,k,vx(indices(iz)),vy(indices(iz)) &
                   ,vz(indices(iz)),Fmax(indices(iz)),neigh(ig1),d2,r2)
              ratio=d2/r2
              if (ratio.lt.best_ratio) then
                 best_ratio=ratio
                 accgrp=ig1
              endif
           enddo

           if (best_ratio.lt.1) then
              accrg=accrg+1
              accrflag=1
              to_group=neigh(accgrp)
              call accretion(neigh(accgrp),i,j,k,indices(iz),Fmax(indices(iz)))
           else

! If the point has not been accreted at all:

! *********
! filament!
! *********

              mass(1)=mass(1)+1
              group_ID(indices(iz))=1
              link_list(indices(iz))=indices(iz)

              if (nmerge.gt.1) nwei2=nwei2+1
           endif
! end of accretion if
        endif

     else
! **********************************************************************
!                       FOURTH CASE: FILAMENTS
! **********************************************************************
        mass(1)=mass(1)+1
        group_ID(indices(iz))=1
        link_list(indices(iz))=indices(iz)

! end of cases

     endif

! Checks whether to accrete all the neighbouring filaments

     if (accrflag.eq.1) then
        do ifil=1,nf
           indx=fil_list(1,ifil)+(fil_list(2,ifil)-1)*Lgrid(1)+ &
                (fil_list(3,ifil)-1)*Lgrid12
           call accretion(to_group,fil_list(1,ifil),fil_list(2,ifil) &
                ,fil_list(3,ifil),indx,Fmax(indices(iz)))
           mass(1)=mass(1)-1
        enddo
     endif


#ifdef PAIRS

! **********************************************************************
! Check for merging pairs
! **********************************************************************

     if (nptm.gt.0.and. &
          (Fmax(indices(iz  )) < Fold-deltaF.or &
          .Fmax(indices(iz-1)) < Flim       .or &
          .Fmax(indices(iz-1)) < fout(iout))) then
        njump=njump+1
        Fold=Fold-deltaF

! First stage of cleaning: if the group has merged we consider the group it
! has merged into.  If the pair has already merged it is removed

        do i=1,nptm
           pairs(1,i)=HALO_APP(pairs(1,i))
           pairs(2,i)=HALO_APP(pairs(2,i))
           if (pairs(1,i).eq.pairs(2,i)) then
! In this case the groups have already merged
              pairs(1,i)=0
              pairs(2,i)=0
           endif
        enddo

! Second stage of cleaning: the list is ordered and duplicates are removed
! NB: nptm is updated

        call clean_list_2(pairs,npair,nptm)

        i=0
        do while (i<nptm .and. pairs(1,i+1)/=0)
           i=i+1
        enddo
        nptm=i

! Checks merging condition of groups and forces it if necessary

        if ((njump == njmax                 .or &
             .Fmax(indices(iz-1)) < Flim    .or &
             .Fmax(indices(iz-1)) < fout(iout)) &
             .and.Fmax(indices(iz)) < Fstart   ) then

           write(31,*)'Checking merging condition for ',nptm,' groups'
           njump=0
           npmerge=0
           do i=1,nptm
              pairs(1,i)=HALO_APP(pairs(1,i))
              pairs(2,i)=HALO_APP(pairs(2,i))
              if (pairs(1,i).eq.pairs(2,i)) then
                 pairs(1,i)=0
                 pairs(2,i)=0
                 cycle
              endif

              call merge_cond(Fmax(indices(iz)),pairs(1,i),pairs(2,i),merge_flag)

              if (merge_flag) then
                 npmerge=npmerge+1
                 if (mass(pairs(1,i)).gt.mass(pairs(2,i))) then
                    call merge_groups(pairs(1,i),pairs(2,i),Fmax(indices(iz)))
                 else
                    call merge_groups(pairs(2,i),pairs(1,i),Fmax(indices(iz)))
                 endif
                 pairs(1,i)=0
                 pairs(2,i)=0
              endif
           enddo

           write(31,*)'Number of groups that have been merged: ',npmerge &
                ,float(npmerge)/float(nptm)

        else
           if (njump.eq.njmax) njump=0
        endif
     endif
         
#endif
! ************************************************************************
! END OF DO-CYCLE ON COLLAPSED POINTS
! ************************************************************************

  enddo

! Counts many things

  ngr_ex=0
  mmax=0
  do i=2,ngroups
     mmax=max(mass(i),mmax)
     if (point(i).gt.0.and.point(i).ne.Npart+1) then
        ngr_ex=ngr_ex+1
     endif
  enddo

  merge_counter=sum(nm)


  write(31,*)'Total number of points considered: ',Npart-iz
  write(31,*)'Total number of groups: ',ngroups
  write(31,*)'Total number of mergers between groups: ',merge_counter
  write(31,*)'Total number of accretion on groups: ',accrg
  write(31,*)'Number of existing groups: ',ngr_ex
  write(31,*)'Largest group mass: ',mmax
  write(31,*)'Statistics of mergers: ',nm
  write(31,*)'Number of points not accreted immediately at merger: ',nwei1
  write(31,*)'Number of points never accreted at merger: ',nwei2

!
! Bye!
!
  return
end subroutine groups


subroutine mf(z)

! Mass function of the groups

  use pinocchio_common ; use group_vectors ; use link

  implicit none
  real(kind=singleR),intent(in)::z

  integer(kind=singleI),parameter::nbin=40
  integer(kind=singleI),dimension(nbin)::ninbin(nbin)
  integer(kind=singleI)::i,ibin,nc
  character(len=30)::filename
  character(len=4)::label
  real(kind=singleR)::mmin,mmax,lmass,deltam,vol,amass
  real(kind=singleR)::x,dm,a,a1,a2,a3
!  real(kind=singleR),external::an_h_s
  real(kind=singleR),external::an_h

  write(31,'(''Calculating the mass function at the redshift '',f5.2)')z

!  if (ilim1.eq.0) then
     vol=BoxSize(1)*BoxSize(2)*BoxSize(3)
!  else
!     vol=(xlim2-xlim1)*(ylim2-ylim1)*(zlim2-zlim1)
!  endif

  mmin=log10(10.*Mpart) ; mmax=16.5
  deltam=(mmax-mmin)/float(nbin)
  lmass=0.
  nc=0 ; ninbin=0

  do i=2,ngroups
     if (point(i) > 0) then
        amass=log10(mass(i)*Mpart)
        if (amass.gt.mmin) nc=nc+1
        lmass=max(amass,lmass)
        ibin=int((amass-mmin)/deltam+1.)
        if (ibin < 1 .or. ibin > nbin) cycle
        ninbin(ibin)=ninbin(ibin)+1
     endif
  enddo
  write(31,*)'Number of large enough objects: ',nc
  write(31,*)'Largest mass: ',10.**lmass

! Writes result

  if (z.lt.10) then
     write(label,'(f4.2)')z
  else
     write(label,'(f4.1)')z
  endif
  filename='frag.'//label//'.'//kernel//'.mf.out'
  open(10,file=filename,status='unknown')
  do i=1,nbin
     x=mmin+(i-.5)*deltam
     dm=10.**(mmin+i*deltam)-10.**(mmin+(i-1)*deltam)
     a=ninbin(i)/vol/dm
     a1=(ninbin(i)+sqrt(float(ninbin(i))))/vol/dm
     a2=(ninbin(i)-sqrt(float(ninbin(i))))/vol/dm
     a3=an_h(10.**x,z)
     a =log10(max(a ,tiny(a )))
     a1=log10(max(a1,tiny(a1)))
     a2=log10(max(a2,tiny(a2)))
     a3=log10(max(a3,tiny(a3)))
     write(10,'(4(1x,f15.8),3x,i10,3x,f12.6)')x,a,a1,a2,ninbin(i),a3
  enddo
  close(10)
  print *,'Written file '//filename

! Bye!

  return
end subroutine mf


subroutine accr_cond(i,j,k,vx,vy,vz,Fmax,grp,dd,rr)

! Condition for accretion

  use grid ; use group_vectors 

  implicit none
  integer(kind=singleI),intent(in)::i,j,k,grp
  real(kind=singleR),intent(in)::Fmax,vx,vy,vz
  real(kind=singleR),intent(out)::dd,rr

  real(kind=singleR)::rvir,dx,dy,dz,d2
  real(kind=singleR),external::distance,virial

  rvir=virial(mass(grp),Fmax,1)
  rr=rvir*rvir
  dd=100.*rr
  dx=distance((i-.5)+vx/Fmax,centres(1,grp)+velgrp(1,grp)/Fmax &
       ,float(Lgrid(1)))
  if (dx.lt.rvir) then
     dy=distance((j-.5)+vy/Fmax,centres(2,grp)+velgrp(2,grp)/Fmax &
          ,float(Lgrid(2)))
     d2=dx*dx+dy*dy
     if (d2.lt.rr) then
        dz=distance((k-.5)+vz/Fmax,centres(3,grp)+velgrp(3,grp)/Fmax &
             ,float(Lgrid(3)))
        d2=d2+dz*dz
        if (d2.le.rr) dd=d2
     endif
  endif

  return
end subroutine accr_cond


subroutine merge_cond(Fmax,grp1,grp2,merge_flag)

! Checks whether two groups must be merged

  use grid ; use group_vectors

  implicit none
  real(kind=singleR),intent(in)::Fmax
  integer(kind=singleI),intent(in)::grp1,grp2
  logical,intent(out)::merge_flag

  real(kind=singleR)::rvir,rvir1,rvir2,dd,rr,dx,dy,dz
  real(kind=singleR),external::distance,virial
      
  merge_flag=.false.
  rvir1=virial(mass(grp1),Fmax,0)
  rvir2=virial(mass(grp2),Fmax,0)
  rvir=max(rvir1,rvir2)
  rr=rvir*rvir
  dx=distance( &
       centres(1,grp1)+velgrp(1,grp1)/Fmax &
       ,centres(1,grp2)+velgrp(1,grp2)/Fmax &
       ,float(Lgrid(1)))
  if (dx.lt.rvir) then
     dy=distance( &
          centres(2,grp1)+velgrp(2,grp1)/Fmax &
          ,centres(2,grp2)+velgrp(2,grp2)/Fmax &
          ,float(Lgrid(2)))
     dd=dx*dx+dy*dy
     if (dd.lt.rr) then
        dz=distance( &
             centres(3,grp1)+velgrp(3,grp1)/Fmax &
             ,centres(3,grp2)+velgrp(3,grp2)/Fmax &
             ,float(Lgrid(3)))
        dd=dd+dz*dz
        if (dd.le.rr) merge_flag=.true.
     endif
  endif

! Bye!

  return
end subroutine merge_cond



subroutine merge_groups(grp1,grp2,time)

! Merges grp2 into grp1

  use pinocchio_common ; use grid
  use link ; use merger ; use group_vectors
  implicit none

  integer(kind=singleI),intent(in)::grp1,grp2
  real(kind=singleR),intent(in)::time

  integer(kind=singleI)::grp(2),i1,ic

! MAJOR MERGING EVENTS

  IF (mass(GRP2).gt.mass(GRP1)/2..AND.MASS(GRP2).GT.MASSMIN) THEN
     MAJORZ(GRP1)=TIME
  ENDIF

! updates the linking list

  grp(1)=grp1
  grp(2)=grp2

  i1=point(grp2)
  do while (link_list(i1).ne.point(grp2))
     group_ID(i1)=grp1
     i1=link_list(i1)
  enddo
  group_ID(i1)=grp1

  link_list(bottom(grp1))=point(grp2)
  link_list(bottom(grp2))=point(grp1)
  bottom (grp1)=bottom(grp2)
  point(grp2)=-1
  bottom (grp2)=-1
#ifdef SPIN

! updates the spin of the group

  call spin_update(time &
       ,mass(grp(1)) &
       ,centres(1,grp(1)),centres(2,grp(1)),centres(3,grp(1)) &
       ,velgrp(1,grp(1)),velgrp(2,grp(1)),velgrp(3,grp(1)) &
       ,spin(1,grp(1)),spin(2,grp(1)),spin(3,grp(1)) &
       ,mass(grp(2)) &
       ,centres(1,grp(2)),centres(2,grp(2)),centres(3,grp(2)) &
       ,velgrp(1,grp(2)),velgrp(2,grp(2)),velgrp(3,grp(2)) &
       ,spin(1,grp(2)),spin(2,grp(2)),spin(3,grp(2)))
#endif

! updates the centre position and masses, and the existence flag

  do ic=1,3
     call update(centres(ic,grp(1)),mass(grp(1)),velgrp(ic,grp(1)), &
          centres(ic,grp(2)),mass(grp(2)),velgrp(ic,grp(2)),Lgrid(ic))
  enddo

! UPDATES THE SUBSTRUCTURE IN THE NEW GROUP

  if (mass(grp(1)).gt.MASSMIN.and.mass(grp(2)).gt.MASSMIN) then
     CALL UPDATESUBSTRUCT(GRP(1),GRP(2),TIME)
  endif

  mass(grp(1))=mass(grp(1))+mass(grp(2))
  
  if (mass(grp(1)).gt.MASSMIN.and.T_APPEAR(grp(1)).eq.-1) &
       T_APPEAR(grp(1))=time

! Bye!

  return
end subroutine merge_groups



SUBROUTINE UPDATESUBSTRUCT(G1,G2,TIME)

! UPGRADE THE SUBSTRUCTURE OF A GROUP: M(G1)>M(G2) IE G2 --> G1 

  use pinocchio_common ; use merger ; use group_vectors ; use grid

  IMPLICIT NONE
  integer(kind=singleI),intent(in)::G1,G2
  real(kind=singleR),intent(in)::time

  integer(kind=singleI)::old_i,first_g2


! IF GROUPS HAVE NO SUBSTRUCTURE
  IF(GR_LL(G1).EQ.G1.AND.GR_LL(G2).EQ.G2) THEN 
     MCENTR(G1)=MASS(G1)
     MCENTR(G2)=MASS(G2)
     GR_LL(G1)=G2
     GR_LL(G2)=G1

! IF G1 HAS SUBSTRUCTURE BUT G2 IS A SINGLEHALO
  ELSEIF(GR_LL(G1).NE.G1.AND.GR_LL(G2).EQ.G2) THEN
! old procedure
!     GR_LL(G2)=GR_LL(G1)
!     GR_LL(G1)=G2
! new procedure
     gr_ll(g2)=g1
     old_i=g1
     do while (gr_ll(old_i).ne.g1)
        old_i=gr_ll(old_i)
     enddo
     gr_ll(old_i)=g2
!     end new procedure
     MCENTR(G2)=MASS(G2)

! IF G1 IS A SINGLE HALO AND G2 HAS SUBSTRUCTURE (RARE!)
  ELSEIF(GR_LL(G1).EQ.G1.AND.GR_LL(G2).NE.G2) THEN
     OLD_I=G2
     DO WHILE (GR_LL(OLD_I).NE.G2)
        OLD_I=GR_LL(OLD_I)
        HALO_APP(OLD_I)=G1
     ENDDO
     HALO_APP(G2)=G1
     GR_LL(G1)=GR_LL(G2)
     GR_LL(G2)=G1
     MCENTR(G1)=MASS(G1)
! BOTH THE GROUPS HAVE SUBSTRUCTURE
  ELSEIF(GR_LL(G1).NE.G1.AND.GR_LL(G2).NE.G2) THEN
     OLD_I=G2
     DO WHILE (GR_LL(OLD_I).NE.G2)
        OLD_I=GR_LL(OLD_I)
        HALO_APP(OLD_I)=G1
     ENDDO
! old procedure
!         INDX=GR_LL(G1)
!         GR_LL(G1)=GR_LL(G2)
!         GR_LL(G2)=INDX
! new procedure
     first_g2=old_i
     old_i=g1
     do while (gr_ll(old_i).ne.g1)
        old_i=gr_ll(old_i)
     enddo
     gr_ll(old_i)=gr_ll(g2)
     gr_ll(g2)=g1
! end new procedure

  ENDIF

  HALO_APP(G2)=G1
  T_ACC(G2)=TIME
  MASS_AT_MERGER(G2)=MASS(G1)
  MERGED_WITH(G2)=G1

  RETURN
END SUBROUTINE UPDATESUBSTRUCT



subroutine accretion(group,i,j,k,indx,F)

! Accretes the point (i,j,k) on the group

  use pinocchio_common ; use merger ; use group_vectors
  use grid ; use matrices ; use link
  implicit none

  integer(kind=singleI),intent(in)::group,i,j,k,indx
  real(kind=singleR),intent(in)::F

  if(group.le.0 .or. group.gt.ngrpmax)then
     write (*,*) ' error in accretion: '
     write (*,*) group,i,j,k,indx
     stop 'accretion'
  endif

#ifdef SPIN
  call spin_update(F &
       ,mass(group),centres(1,group),centres(2,group),centres(3,group) &
       ,velgrp(1,group),velgrp(2,group),velgrp(3,group) &
       ,spin(1,group),spin(2,group),spin(3,group) &
       ,1,i-0.5,j-0.5,k-0.5,vx(indx),vy(indx),vz(indx),0.,0.,0.)
#endif

  call update(centres(1,group),mass(group) &
       ,velgrp(1,group),i-.5,1,vx(indx),Lgrid(1))
  call update(centres(2,group),mass(group) &
       ,velgrp(2,group),j-.5,1,vy(indx),Lgrid(2))
  call update(centres(3,group),mass(group) &
       ,velgrp(3,group),k-.5,1,vz(indx),Lgrid(3))

  mass(group)=mass(group)+1

  if (mass(group).gt.MASSMIN.and.T_APPEAR(group).eq.-1) T_APPEAR(group)=F
  group_ID(indx)=group

! Updates the linking list

  link_list(bottom(group))=indx
  bottom(group)=indx
  link_list(indx)=point(group)

! Bye!

  return
end subroutine accretion


subroutine update(x1,m1,v1,x2,m2,v2,Lgrid)

! Updates the centre position of group 1 respecting PBCs

  use numbers
  implicit none
  real(kind=singleR),intent(inout)::x1,v1
  real(kind=singleR),intent(in)::x2,v2
  integer(kind=singleI),intent(in)::m1,m2,Lgrid

  real(kind=singleR)::d
  real(kind=singleR),external::pbc

  d=abs(x1-x2)
  if (d.le.Lgrid/2) then
! in this case theres no problem
     x1=(x1*m1+x2*m2)/float(m1+m2)
  else
! PBC must be considered
     if (x1.gt.Lgrid/2) then
! in this case xc is in the second half, and x2 -> x2+Lgrid
        x1=(x1*m1+(x2+Lgrid)*m2)/float(m1+m2)
     else
! in this case xc is in the first half, and x2 -> x2-Lgrid
        x1=(x1*m1+(x2-Lgrid)*m2)/float(m1+m2)
     endif
! checks PBC again
     x1=pbc(x1,float(Lgrid))
  endif

! velocity
  v1=(v1*m1+v2*m2)/float(m1+m2)

! bye!
  return
end subroutine update


#ifdef SPIN
subroutine spin_update(F &
     ,m1,q1x,q1y,q1z,v1x,v1y,v1z,s1x,s1y,s1z &
     ,m2,q2x,q2y,q2z,v2x,v2y,v2z,s2x,s2y,s2z)

! Updates the spin of a halo in case of merging or accretion

  use pinocchio_common ; use grid
  implicit none

  integer(kind=singleI),intent(in)::m1,m2
  real(kind=singleR),intent(in)::F,q1x,q1y,q1z,v1x,v1y,v1z,q2x,q2y,q2z,v2x,v2y,v2z,s2x,s2y,s2z
  real(kind=singleR),intent(inout)::s1x,s1y,s1z

  integer(kind=singleI)::ic,ix,iy,iz

  real(kind=singleR)::rng(3),z,runit,vunit,tmass
  real(kind=singleR)::z_collapse,t_collapse,t_me
  real(kind=singleR),dimension(3)::q1,v1,s1,q2,v2,s2,pq2,cm,vcm,rq1,rq2,rv1,rv2
  real(kind=singleR),external::pbc,distance !,inverse_grow_s,Hubble_s,fomega_s,scale_s,time_s
  real(kind=singleR),external::inverse_grow,Hubble,fomega,pbc,distance,scale,time

  q1(1)=q1x ; q1(2)=q1y ; q1(3)=q1z
  v1(1)=v1x ; v1(2)=v1y ; v1(3)=v1z
  s1(1)=s1x ; s1(2)=s1y ; s1(3)=s1z
  q2(1)=q2x ; q2(2)=q2y ; q2(3)=q2z
  v2(1)=v2x ; v2(2)=v2y ; v2(3)=v2z
  s2(1)=s2x ; s2(2)=s2y ; s2(3)=s2z
  rng=float(Lgrid)

! phisical units
! NB TIME IS INVERSE GROWING MODE!!!

  z_collapse= inverse_grow(1./F)
  t_collapse= time(z_collapse)
  t_me      = 0.5*t_collapse
  z         = 1./scale(t_me)-1.
  runit     = InterPartDist/(1.+z)
  vunit     = runit*F*fomega(z)*Hubble(z)
! compute position and velocity of center of mass
  tmass = float(m1+m2)
  cm  = 0.0
  vcm = 0.0

  do ic=1,3
     pq2(ic) = q2(ic)
     if (abs(q1(ic)-q2(ic)).gt..5*rng(ic))then
        if(q1(ic) .lt. q2(ic))then
           pq2(ic)=pq2(ic)-rng(ic)
        else
           pq2(ic)=pq2(ic)+rng(ic)
        endif
     endif
     cm(ic)  = pbc((m1*q1(ic)+m2*pq2(ic))/tmass,rng(ic))
     vcm(ic) = (m1*v1(ic)+m2*v2(ic))/tmass
     rq1(ic) = distance(cm(ic),q1(ic),rng(ic))
     rq2(ic) = distance(cm(ic),pq2(ic),rng(ic))
     rv1(ic) = v1(ic)-vcm(ic)
     rv2(ic) = v2(ic)-vcm(ic)
  enddo
  do ix=1,3
     iy=mod(ix,3)+1
     iz=mod(iy,3)+1
     s1(ix) = s1(ix) + s2(ix)  &
          + runit*vunit*(       &
          m1*(rq1(iy)*rv1(iz)-rq1(iz)*rv1(iy)) &
          + m2*(rq2(iy)*rv2(iz)-rq2(iz)*rv2(iy)) &
          )
  enddo
 
! Bye!

  s1x = s1(1)
  s1y = s1(2)
  s1z = s1(3)
  return
end subroutine spin_update
#endif


subroutine increase_ngrpmax()

! Increases ngrpmax by 10 per cent

  use grid ; use group_vectors ; use merger ; use link
  implicit none
  
  integer(kind=singleI)::ngrpmax_old
  integer(kind=singleI),allocatable,dimension(:)::l_mass
  integer(kind=singleI),allocatable,dimension(:,:)::l_peak
  real(kind=singleR),allocatable,dimension(:,:)::l_centres,l_velgrp
#ifdef SPIN
  real(kind=singleR),allocatable,dimension(:,:)::l_spin
#endif
  integer(kind=singleI),allocatable,dimension(:)::l_GR_LL,l_MCENTR,l_HALO_APP &
       ,l_MASS_AT_MERGER,l_MERGED_WITH,l_point,l_bottom
  real(kind=singleR),allocatable,dimension(:)::l_MAJORZ,l_T_ACC,l_T_APPEAR


! allocates tmp vectors
  allocate (l_mass   (  ngrpmax))
  allocate (l_peak   (3,ngrpmax))
  allocate (l_centres(3,ngrpmax))
  allocate (l_velgrp (3,ngrpmax))
#ifdef SPIN
  allocate (l_spin   (3,ngrpmax))
#endif
  allocate (l_gr_ll         (ngrpmax))
  allocate (l_Mcentr        (ngrpmax))
  allocate (l_halo_app      (ngrpmax))
  allocate (l_mass_at_merger(ngrpmax))
  allocate (l_merged_with   (ngrpmax))
  allocate (l_majorz        (ngrpmax))
  allocate (l_t_acc         (ngrpmax))
  allocate (l_t_appear      (ngrpmax))
  allocate (l_point         (ngrpmax))
  allocate (l_bottom        (ngrpmax))

! saves current vectors
  l_mass           = mass          
  l_peak           = peak
  l_centres        = centres       
  l_velgrp         = velgrp        
#ifdef SPIN
  l_spin           = spin          
#endif
  l_gr_ll          = gr_ll         
  l_Mcentr         = Mcentr        
  l_halo_app       = halo_app      
  l_mass_at_merger = mass_at_merger
  l_merged_with    = merged_with   
  l_majorz         = majorz        
  l_t_acc          = t_acc         
  l_t_appear       = t_appear      
  l_point          = point
  l_bottom         = bottom

! deallocate current vectors
  deallocate (mass)
  deallocate (peak)
  deallocate (centres)
  deallocate (velgrp)
#ifdef SPIN   
  deallocate (spin)
#endif        
  deallocate (gr_ll)
  deallocate (Mcentr)
  deallocate (halo_app)
  deallocate (mass_at_merger)
  deallocate (merged_with)
  deallocate (majorz)
  deallocate (t_acc)
  deallocate (t_appear)
  deallocate (point)
  deallocate (bottom)

! updates ngrpmax
  ngrpmax_old=ngrpmax
  ngrpmax=1.1*ngrpmax

! allocates new vectors
  allocate (mass          (  ngrpmax))
  allocate (peak          (3,ngrpmax))
  allocate (centres       (3,ngrpmax))
  allocate (velgrp        (3,ngrpmax))
#ifdef SPIN 
  allocate (spin          (3,ngrpmax))
#endif      
  allocate (gr_ll         (  ngrpmax))
  allocate (Mcentr        (  ngrpmax))
  allocate (halo_app      (  ngrpmax))
  allocate (mass_at_merger(  ngrpmax))
  allocate (merged_with   (  ngrpmax))
  allocate (majorz        (  ngrpmax))
  allocate (t_acc         (  ngrpmax))
  allocate (t_appear      (  ngrpmax))
  allocate (point         (  ngrpmax))
  allocate (bottom        (  ngrpmax))

! restores data
  mass          (  1:ngrpmax_old) = l_mass          
  peak          (:,1:ngrpmax_old) = l_peak
  centres       (:,1:ngrpmax_old) = l_centres       
  velgrp        (:,1:ngrpmax_old) = l_velgrp        
#ifdef SPIN
  spin          (:,1:ngrpmax_old) = l_spin          
#endif
  gr_ll         (  1:ngrpmax_old) = l_gr_ll         
  Mcentr        (  1:ngrpmax_old) = l_Mcentr        
  halo_app      (  1:ngrpmax_old) = l_halo_app      
  mass_at_merger(  1:ngrpmax_old) = l_mass_at_merger
  merged_with   (  1:ngrpmax_old) = l_merged_with   
  majorz        (  1:ngrpmax_old) = l_majorz        
  t_acc         (  1:ngrpmax_old) = l_t_acc         
  t_appear      (  1:ngrpmax_old) = l_t_appear      
  point         (  1:ngrpmax_old) = l_point
  bottom        (  1:ngrpmax_old) = l_bottom

  mass          (  ngrpmax_old+1:ngrpmax) = 0
  peak          (:,ngrpmax_old+1:ngrpmax) = 0.
  centres       (:,ngrpmax_old+1:ngrpmax) = 0.
  velgrp        (:,ngrpmax_old+1:ngrpmax) = 0.
#ifdef SPIN
  spin          (:,ngrpmax_old+1:ngrpmax) = 0.
#endif
  gr_ll         (  ngrpmax_old+1:ngrpmax) = -1 
  Mcentr        (  ngrpmax_old+1:ngrpmax) = -1
  halo_app      (  ngrpmax_old+1:ngrpmax) = -1
  mass_at_merger(  ngrpmax_old+1:ngrpmax) = 0
  merged_with   (  ngrpmax_old+1:ngrpmax) = -1
  majorz        (  ngrpmax_old+1:ngrpmax) = -1.
  t_acc         (  ngrpmax_old+1:ngrpmax) = -1.
  t_appear      (  ngrpmax_old+1:ngrpmax) = -1.
  point         (  ngrpmax_old+1:ngrpmax) = 0
  bottom        (  ngrpmax_old+1:ngrpmax) = 0

! deallocates tmp vectors
  deallocate (l_mass)
  deallocate (l_peak)
  deallocate (l_centres)
  deallocate (l_velgrp)
#ifdef SPIN
  deallocate (l_spin)
#endif
  deallocate (l_gr_ll)
  deallocate (l_Mcentr)
  deallocate (l_halo_app)
  deallocate (l_mass_at_merger)
  deallocate (l_merged_with)
  deallocate (l_majorz)
  deallocate (l_t_acc)
  deallocate (l_t_appear)
  deallocate (l_point)
  deallocate (l_bottom)

  print *,'ngrpmax increased from ',ngrpmax_old,' to ',ngrpmax

! Bye!
  return
end subroutine increase_ngrpmax


subroutine out_write(z)

! Writes the group catalogues and linking list

  use pinocchio_common ; use run_flags
  use grid
  use link ; use merger ; use group_vectors
  implicit none

  real(kind=singleR),intent(in)::z

!  real(kind=singleR),parameter::nsafe=0.7
  integer(kind=singleI)::ic,i,j,k,kk,ipart,Lgrid12,safe
  real(kind=singleR)::c1,c2,c3,mstar
  character(len=30)::filename
  character(len=4)::label    
  real(kind=singleR),external::pbc !,grow_s,fomega_s,Hubble_s,inverse_grow_s,aM_star_s
  real(kind=singleR),external::grow,fomega,Hubble,inverse_grow,aM_star,an_h
  real(kind=singleR),dimension(3)::xpos,qpos,c4
  logical,dimension(:),allocatable::good_groups
  real(kind=singleR)::Mlarge,massfn,Vol
  logical::cond

  integer(kind=singleI)::ig,jg,kg,gg
  integer(kind=singleI),external::ipbc


#ifdef SPIN
  real(kind=singleR)::spin_per_unit_mass(3),fact2,logmass
  real(kind=singleR),parameter::fact=0.8,slope=0.15
!  real(kind=singleR),external::random_s
  real(kind=singleR),external::random
#endif

!  Vol    = product(BoxSize)
!  Mlarge = 1.e17
!  massfn = 0.
!  cond   = .true.
!
!  do while (cond)
!     Mlarge=Mlarge*0.9
!     massfn = massfn+an_h(Mlarge,z)*0.1*Mlarge*Vol
!     cond = (massfn .lt. 1.)
!  enddo
!
!  safe=nint(nsafe*(Mlarge/Mpart)**0.3333)
!  print *,'Safe value: ',safe,Mlarge

  safe=3

! Checks which groups do not touch the borders
  allocate (good_groups(ngroups))
  good_groups=.true.
  good_groups(1)=.false.
  Lgrid12=Lgrid(1)*Lgrid(2)

!  do i=2,ngroups
!     if ((.not.pbc_flag(1) .and. (centres(1,i)<=1+safe .or. centres(1,i)>=Lgrid(1)-safe))  .or. &
!         (.not.pbc_flag(2) .and. (centres(2,i)<=1+safe .or. centres(2,i)>=Lgrid(2)-safe))  .or. &
!         (.not.pbc_flag(3) .and. (centres(3,i)<=1+safe .or. centres(3,i)>=Lgrid(3)-safe))) &
!         good_groups(i)=.false.
!  enddo


  do ipart=1,Npart
     k=1+(ipart-1)/Lgrid12
     kk=ipart-1-(k-1)*Lgrid12
     j=1+kk/Lgrid(1)
     i=1+(kk-(j-1)*Lgrid(1))

     if (group_ID(ipart)>1 .and. &
        (.not.pbc_flag(1) .and. (i<=1+safe .or. i>=Lgrid(1)-safe)  .or. &
         .not.pbc_flag(2) .and. (j<=1+safe .or. j>=Lgrid(2)-safe)  .or. &
         .not.pbc_flag(3) .and. (k<=1+safe .or. k>=Lgrid(3)-safe))) &
         good_groups(group_ID(ipart))=.false.

  enddo

! Out_flag = 0: no writing out

  if (out_flag.eq.0) return

  if (z.lt.10) then
     write(label,'(f4.2)')z
  else
     write(label,'(f4.1)')z
  endif

! Outflag = 1, 3 or 4: saves catalogue

  if (out_flag.eq.1.or.out_flag.ge.3) then
     write(31,*)'Writing the group catalogue at redshift ',z
     filename='frag.'//label//'.'//kernel//'.grp.out'
     open(10,file=filename,status='unknown')

     c1=InterPartDist
     c2=grow(z)
     c3=c1*c2*fomega(z)*Hubble(z)/(1.+z)
     c4=InterPartDist*(Lstart-1)
     mstar=log10(aM_star(z))

     do i=2,ngroups
        if (point(i).gt.1 .and. mass(i).gt.0 .and. good_groups(i)) then 

           do j=1,3
              if (pbc_flag(j)) then
                 qpos(j)=pbc(centres(j,i)*c1+c4(j),BoxSize(j))
                 xpos(j)=pbc((centres(j,i)+velgrp(j,i)*c2)*c1+c4(j),BoxSize(j))
              else
                 qpos(j)=centres(j,i)*c1+c4(j)
                 xpos(j)=(centres(j,i)+velgrp(j,i)*c2)*c1+c4(j)
              endif
           enddo

#ifdef SPIN
           logmass=log10(mass(i)*Mpart)
           fact2=max(fact-slope*(logmass-mstar),0.)
           fact2=min(fact2,1.)
           do ic=1,3
              spin_per_unit_mass(ic)=spin(ic,i)/float(mass(i)) &
                   *((1.-fact2)+fact2*float(random(seed)))
           enddo

           write(10,'(1x,i7,1x,f7.4,6(1x,f7.2),6(1x,f9.2),4(1x,i8))') &
                i,logmass &
                ,qpos &
                ,xpos &
                ,velgrp(:,i)*c3 &
                ,spin_per_unit_mass
                ,mass(i) !&
!                ,peak(:,i)
#else
           write(10,'(1x,i7,1x,f7.4,6(1x,f7.2),3(1x,f9.2),4(1x,i8))') &
                i,log10(mass(i)*Mpart) &
                ,qpos &
                ,xpos &
                ,velgrp(:,i)*c3 &
                ,mass(i) !&
!                ,peak(:,i)

#endif
        endif
     enddo
     close(10)
     print *,'Written file '//filename
  endif

! Out_flag = 4: Saves linking list

  if (out_flag.eq.4) then
     write(31,*)'Writing the linking list at redshift ',z
     filename='frag.'//label//'.'//kernel//'.lnk.out'
     open(10,file=filename,status='unknown',form='unformatted')
     write(10)Npart
     write(10)group_ID

!!!! ***********************************************
!!!     open(10,file=filename,status='unknown')
!!!     write(10,*)Npart
!!!
!!!     do ipart=1,Npart
!!!        k=1+(ipart-1)/Lgrid12
!!!        kk=ipart-1-(k-1)*Lgrid12
!!!        j=1+kk/Lgrid(1)
!!!        i=1+(kk-(j-1)*Lgrid(1))
!!!        
!!!        ig=ipbc(i+Lstart(1)-1,Lgrid_global(1))
!!!        jg=ipbc(j+Lstart(2)-1,Lgrid_global(2))
!!!        kg=ipbc(k+Lstart(3)-1,Lgrid_global(3))
!!!
!!!        if (group_ID(ipart)>1) then
!!!           gg =    peak(1,group_ID(ipart)) &
!!!                + (peak(2,group_ID(ipart))-1)*Lgrid_global(1) &
!!!                + (peak(3,group_ID(ipart))-1)*Lgrid_global(1)*Lgrid_global(2)
!!!           write(10,'(99(1x,i7))') ig,jg,kg,gg,peak(:,group_ID(ipart))
!!!        else
!!!           write(10,'(99(1x,i7))') ig,jg,kg,0
!!!        endif
!!!
!!!
!!!     enddo
!!!! ***********************************************


     close(10)
     print *,'Written file '//filename
  endif

! Out_flag = 2, 3, or 4: Saves histories

  if (out_flag.ge.2) then
     write(31,*)'Writing group histories at redshift ',z
     filename='frag.'//label//'.'//kernel//'.halo.out'
     OPEN(10,FILE=filename,STATUS='unknown')
     DO i=2,ngroups
        IF (MASS(I).gt.MASSMIN .and. good_groups(i)) THEN
           if (T_ACC(I).gt.0.) then
              c1=inverse_grow(1./T_ACC(I))
           else
              c1=-1.
           endif
           if (MAJORZ(I).gt.0.) then
              c2=inverse_grow(1./MAJORZ(I))
           else
              c2=-1.
           endif
           if (T_APPEAR(I).gt.0.) then
              c3=inverse_grow(1./T_APPEAR(I))
           else
              c3=-1.
           endif
           WRITE(10,'(7(1x,i7),3(1X,F9.4))') &
                I,HALO_APP(I),GR_LL(I),MERGED_WITH(I) &
                ,MASS(I),MASS_AT_MERGER(I),MCENTR(I),c1,c2,c3
        ENDIF
     ENDDO
     CLOSE(10)
     print *,'Written file '//filename

  endif

! Group-peak table
  if (z==zfin) then
     filename='frag.'//kernel//'.peaks.out'
     open(10,file=filename,status='unknown')
     do i=2,ngroups
        write(10,'(5(1x,i8))')i,peak(:,i),mass(i)
     enddo
     close(10)
     print *,'Written file '//filename
  endif

  deallocate(good_groups)

! Bye!

  return
end subroutine out_write



SUBROUTINE clean_list(n,arr)

  use numbers
  implicit none
  integer(kind=singleI),intent(in)::n
  integer(kind=singleI),intent(inout),dimension(n)::arr

  INTEGER(KIND=SINGLEI)::i,j,a
  logical::done

  do j=2,n

     do i=j-1,1,-1
        if (arr(i).eq.arr(j)) arr(j)=0
     enddo

     a=arr(j)
     done=.false.
     do i=j-1,1,-1
        if (arr(i)>0)  then
           done=.true.
           exit
        endif
        arr(i+1)=arr(i)
     enddo
     if (.not.done) i=0
     arr(i+1)=a
  enddo
  return
END SUBROUTINE clean_list



SUBROUTINE clean_list_2(arr,nmax,n)

  use numbers
  implicit none
  integer(kind=singleI),intent(in)::n,nmax
  integer(kind=singleI),intent(inout),dimension(2,nmax)::arr

  integer(kind=singleI)::i,j,a(2)

  do j=2,n

     do i=j-1,1,-1
        if (arr(1,i) == arr(1,j) .and. arr(2,i) == arr(2,j)) then
           arr(:,j)=0
        endif
     enddo

     a=arr(:,j)
     do i=j-1,1,-1
        if (arr(1,i)>0) exit
        arr(:,i+1)=arr(:,i)
     enddo
     if (i<=1) i=0
     arr(:,i+1)=a
  enddo

  return
END SUBROUTINE clean_list_2



real(kind=singleR) function virial(grp,F,flag)

! Gives the virial radius

  use virial_params ; use pinocchio_common
  implicit none

  integer(kind=singleI),intent(in)::grp,flag
  real(kind=singleR),intent(in)::F

  real(kind=singleR)::rlag,zero,corr,variance
  real(kind=singleR),external::aLambda

  rlag=float(grp)**.3333333333333333333333
  if (flag.eq.0) then
! merging
     virial=rlag*f_m + f_rm
  else
! accretion
     variance=aLambda(rlag*InterPartDist)/F**2
     corr=(variance)**(0.5*espo)    !*F**(-espo)
     corr=min(corr,5.)
     zero=f_ra+f_s*rlag*corr
     virial=f_a*rlag+zero   !+ exp(max(limit_var-variance,0.)) - 1.

!     variance=variance/F**2
!     if (variance<limit_var) virial = virial+(exp(0.5*(virial+limit_var-variance))-1.)

  endif

  return
end function virial


subroutine sort(n,vector)

! very simple sorting subroutine

  use numbers
  implicit none
  integer(kind=singleI),intent(in)::n
  real(kind=singleR),intent(inout)::vector(n)

! local variables
  integer(kind=singleI)::i,j
  real(kind=singleR)::a

  c2: do j=2,n
     a=vector(j)
     c1: do i=j-1,1,-1
        if (vector(i)>a) then
           vector(i+1)=a
           cycle c2
        else
           vector(i+1)=vector(i)
        endif
     enddo c1
     vector(i+1)=a
  enddo c2

  return
end subroutine sort
