!*****************************************************************
!*                       PINOCCHI0  V2.2                         *
!*  (PINpointing Orbit-Crossing Collapsed HIerarchical Objects)  *
!*****************************************************************
!
!This code was written by
!Pierluigi Monaco, Tom Theuns, Giuliano Taffoni
!Copyright (C) 2008
!
!web page: http//adlibitum.oats.inaf.it/monaco/Homepage/Pinocchio/
!
!This program is free software; you can redistribute it and/or modify
!it under the terms of the GNU General Public License as published by
!the Free Software Foundation; either version 2 of the License, or
!(at your option) any later version.
!
!This program is distributed in the hope that it will be useful,
!but WITHOUT ANY WARRANTY; without even the implied warranty of
!MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!GNU General Public License for more details.
!
!You should have received a copy of the GNU General Public License
!along with this program; if not, write to the Free Software
!Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA



module grid
  use numbers
  integer(kind=singleI)::Lgrid(3),Lgrid_global(3),Lstart(3),Npart
  integer(kind=singleI)::ngrpmax
  real(kind=singleR)   ::BoxSize_global(3)  !,var_grid
  logical,dimension(3) ::pbc_flag
end module grid

module link
  use numbers
  integer(kind=singleI),allocatable,dimension(:)::indices,group_ID,link_list,point,bottom
end module link

module merger
  use numbers
  integer(kind=singleI),parameter::MASSMIN=10
  integer(kind=singleI),allocatable,dimension(:)::GR_LL,MCENTR,HALO_APP &
       ,MASS_AT_MERGER,MERGED_WITH
  real(kind=singleR),allocatable,dimension(:)::MAJORZ,T_ACC,T_APPEAR
end module merger

module group_vectors
  use numbers
  integer(kind=singleI)::ngroups
  integer(kind=singleI),allocatable,dimension(:)::mass
  integer(kind=singleI),allocatable,dimension(:,:)::peak
  real(kind=singleR),allocatable,dimension(:,:)::centres,velgrp
#ifdef SPIN
  integer(kind=singleI)::seed
  real(kind=singleR),allocatable,dimension(:,:)::spin
#endif
end module group_vectors

module matrices
  use numbers
  real(kind=singleR),allocatable,dimension(:)::Fmax,vx,vy,vz
end module matrices

module virial_params
  use numbers
  real(kind=singleR)::f_m,f_rm,f_a,f_s,f_ra,espo,limit_var
end module virial_params

module run_flags
  use numbers
  integer(kind=singleI)::stop_flag,out_flag,param_flag
  real(kind=singleR)::Flim,zfin
  character(len=*),parameter::dir='Data/'
  integer(kind=singleI)::nout
  real(kind=singleR),allocatable,dimension(:)::fout,zout
end module run_flags

module pairs
#ifdef PAIRS
  use numbers
  integer(kind=singleI),parameter::npair=1e6
  integer(kind=singleI),dimension(2,npair)::pairs
  integer(kind=singleI)::njmax,nptm,npmerge,njump,mmin_pair
  real(kind=singleR)::Fold,deltaF,Fstart
#endif
end module pairs


module pinocchio_common
! This module contains the basic properties of the (highest-resolution) box, 
! the cosmological parameters, the run name and other parameters needed by
! cosmo-sp-2.2
  use numbers
! Cosmological parameters
  real(kind=singleR):: Omega0  
  real(kind=singleR):: OmegaLambda
  real(kind=singleR):: Hubble100
  real(kind=singleR):: Sigma8
  real(kind=singleR):: PowerSpectrum  ! flag for the transfer function
  real(kind=singleR):: csp            ! normalization constant for P(k)
  real(kind=singleR):: delta_c        ! delta_c Press & Schechter parameter
  integer(kind=singleI):: filter      ! flag for window filter
  integer(kind=singleI):: flag_ps     ! flag for the analytical mass function
! Basic properties of the highest resolution box
  real(kind=singleR):: InterPartDist  ! comoving inter-particle distance
  real(kind=singleR):: BoxSize(3)     ! 3D comoving size
  real(kind=singleR):: Mpart          ! particle mass
!
  character(len=6)  :: run, kernel    ! name of the run

  real(kind=singleR),parameter::Omegab=0.0456   ! baryonic Omega 
end module pinocchio_common

module fftw_variables
  use numbers
! grid size
  integer(kind=singleI)     :: nx_fftw, ny_fftw, nz_fftw
end module fftw_variables


module pinocchio_constants
! CONSTANTS
  use numbers
  real(kind=singleR), parameter :: pi   =3.1415926535897931d0
  real(kind=singleR), parameter :: twopi=2.d0*pi
end module pinocchio_constants


#ifdef HDF5
module pinocchio_hdf5
! numbers required to handle hdf5
  use numbers
  integer(kind=singleI) :: file_handle, group_handle, attribute_handle
end module pinocchio_hdf5
#endif
