!*****************************************************************
!*                       PINOCCHI0  V2.2                         *
!*  (PINpointing Orbit-Crossing Collapsed HIerarchical Objects)  *
!*****************************************************************
!
!This code was written by
!Pierluigi Monaco, Tom Theuns, Giuliano Taffoni
!Copyright (C) 2008
!
!web page: http//adlibitum.oats.inaf.it/monaco/Homepage/Pinocchio/
!
!This program is free software; you can redistribute it and/or modify
!it under the terms of the GNU General Public License as published by
!the Free Software Foundation; either version 2 of the License, or
!(at your option) any later version.
!
!This program is distributed in the hope that it will be useful,
!but WITHOUT ANY WARRANTY; without even the implied warranty of
!MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!GNU General Public License for more details.
!
!You should have received a copy of the GNU General Public License
!along with this program; if not, write to the Free Software
!Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


#ifdef ASCII_READ
#  define READ_STATEMENT read(1,*,err=20)
#else
#  define READ_STATEMENT read(1,err=20)
#endif
#ifdef ASCII_WRITE
#  define WRITE_STATEMENT write(1,*,err=20)
#else
#  define WRITE_STATEMENT write(1,err=20)
#endif


subroutine abort_code(messg)

  implicit none
  character(*) :: messg
  
  write (*,'(a)') messg
  stop
end subroutine abort_code


subroutine open_file_to_read(iproc,unit,file,read_headers)

! opens a file for reading in standard or hdf5 mode
! processor 0 reads the headers

  use numbers
#ifdef HDF5
  use hdf5_wrapper; use pinocchio_hdf5
#endif
  implicit none
  integer(kind=singleI),intent(in)::iproc
  integer(kind=singleI),intent(inout)::unit
  character(len=*),intent(in)::file
  logical,intent(in)::read_headers

! local variables
  logical::flag
#ifdef ASCII_WRITE
  character(len=*),parameter :: form='formatted'
#else
  character(len=*),parameter :: form='unformatted'
#endif
  character,parameter :: read='read',write='write', test='test'

  inquire(file=file,exist=flag)
  if (.not.flag) call abort_code ('File '//trim(file)//' does not exist')

  write(*,'(a)')'opening file '//trim(file)//' for reading'

#ifdef HDF5
  call hdf5_open_file(file_handle, file, readonly=.true.)
  if (iproc==0 .and. read_headers) then
     call rw_cosmology(test)
!     call rw_grid(test)
!     call rw_fmax(test)
  endif
  return
#endif
! open file and read header
  open(unit=unit,file=file,status='old',form=form)
!!  if (iproc==0) then
!     call read_pedantic_header()
     call read_cosm_header()
!!  endif

  return
end subroutine open_file_to_read


subroutine read_cosm_header()

! read grid parameters

  use pinocchio_constants
  use pinocchio_common
  use fftw_variables, ONLY: nx_fftw, ny_fftw, nz_fftw
  implicit none

! local variables
  integer(kind=singleI)::Lx,Ly,Lz
  character(len=80)::info

  READ_STATEMENT Lx, Ly, Lz
  if (Lx /= nx_fftw .or. Ly /= ny_fftw .or. Lz /= nz_fftw) then
     write(*,'(a)')'WARNING: SEVERE PROBLEM'
     write(*,'(a)')'Inconsistent grid data in fortran file'
     write(*,'("Parameters:",5(1x,i9))')Lx,Ly,Lz
!  else
!     write(*,'(a)')'grid lengths match'
  endif

  return
20 write(*,'(a)')' error reading cosmological header '
  call abort_code('read_cosm_header  ')
end subroutine read_cosm_header


#ifdef HDF5
subroutine rw_cosmology(action)

  use hdf5_wrapper; use pinocchio_hdf5
  use pinocchio_common
  implicit none
  character(*), intent(in) :: action
  character(len=200)        :: info
  character,parameter :: read='read',write='write',test='test'

  real(kind=singleR)::diff
  real(kind=doubleR)::par1,par2,par3,par4,par5(3),par6,par7,par8
  real(kind=singleR),parameter::accuracy=1.e-3

!!!  if (action == write) then
!!!
!!!     call hdf5_create_group(file_handle,'/Cosmology/')
!!!     call hdf5_write_data(file_handle,'Cosmology/Omega0',Omega0)
!!!     call hdf5_write_data(file_handle,'Cosmology/OmegaLambda',OmegaLambda)
!!!     call hdf5_write_data(file_handle,'Cosmology/Hubble100',Hubble100)
!!!     call hdf5_write_data(file_handle,'Cosmology/Sigma8',Sigma8)
!!!     call hdf5_write_data(file_handle,'Cosmology/BoxSize',BoxSize)
!!!     call hdf5_write_data(file_handle,'Cosmology/PowerSpectrum',PowerSpectrum)
!!!     call hdf5_write_data(file_handle,'Cosmology/Mpart',Mpart)
!!!     call hdf5_write_data(file_handle,'Cosmology/InterPartDist',InterPartDist)
!!!
!!!  else if (action == read) then
!!!
!!!     call hdf5_read_data(file_handle,'Cosmology/Omega0',Omega0)
!!!     call hdf5_read_data(file_handle,'Cosmology/OmegaLambda',OmegaLambda)
!!!     call hdf5_read_data(file_handle,'Cosmology/Hubble100',Hubble100)
!!!     call hdf5_read_data(file_handle,'Cosmology/Sigma8',Sigma8)
!!!     call hdf5_read_data(file_handle,'Cosmology/BoxSize',BoxSize)
!!!     call hdf5_read_data(file_handle,'Cosmology/PowerSpectrum',PowerSpectrum)
!!!     call hdf5_read_data(file_handle,'Cosmology/Mpart',Mpart)
!!!     call hdf5_read_data(file_handle,'Cosmology/InterPartDist',InterPartDist)
!!!
!!!  else 

  if (action == test) then

     call hdf5_read_data(file_handle,'Cosmology/Omega0',       par1)
     call hdf5_read_data(file_handle,'Cosmology/OmegaLambda',  par2)
     call hdf5_read_data(file_handle,'Cosmology/Hubble100',    par3)
     call hdf5_read_data(file_handle,'Cosmology/Sigma8',       par4)
     call hdf5_read_data(file_handle,'Cosmology/BoxSize',      par5)
     call hdf5_read_data(file_handle,'Cosmology/PowerSpectrum',par6)
     call hdf5_read_data(file_handle,'Cosmology/Mpart',        par7)
     call hdf5_read_data(file_handle,'Cosmology/InterPartDist',par8)
     
     diff = maxval( abs( &
          ((/par1,par2,par3,par4,par5(:),par6,par7,par8/)&
       -  (/Omega0,OmegaLambda,Hubble100,Sigma8,BoxSize(:),PowerSpectrum,Mpart,InterPartDist/))&
     / abs((/Omega0,OmegaLambda,Hubble100,Sigma8,BoxSize(:),PowerSpectrum,Mpart,InterPartDist/))))


     if (diff>accuracy) then
        write(*,'(a)')'WARNING: SEVERE PROBLEM'
        write(*,'(a)')'Inconsistent grid data in hdf5 file'
        write(*,'("Parameters ",10(1x,e12.6))') par1,par2,par3,par4,par5(:),par6,par7,par8

        write(*,'(a)')'run parameters should be : '
        write(*,'(''Parameters '',10(1x,e12.6))') Omega0,OmegaLambda,Hubble100,Sigma8,BoxSize(:)&
             ,PowerSpectrum,Mpart,InterPartDist
     endif

  else
 
    call abort_code(' invalid action in rw_cosmology ')

  endif
end subroutine rw_cosmology

#endif


subroutine read_plane_fragment(unit,iplane,Lx,Ly,plane)

! Writes a plane in standard or hdf5 format
! In this case the plane is given as an argument and not in a module
! This is used by fragment-2.2
  
  use numbers
#ifdef HDF5
  use hdf5_wrapper; use pinocchio_hdf5
#endif
  implicit none
  integer(kind=singleI),intent(in):: unit,iplane,Lx,Ly
  real(kind=singleR),dimension(Lx,Ly),intent(inout)  :: plane
  character(len=4),external       :: write_string
  character(len=11)               :: field
  integer(kind=singleI)           :: ii
  character(len=80)               :: info, var_name

#ifdef HDF5
  field='/Plane'//write_string(iplane)//'/'
  var_name = trim(field)//'Data'
  call hdf5_read_data(file_handle,var_name,plane)
  return
#endif
! plane ID
  READ_STATEMENT ii
  if (ii/=iplane .and. iplane .ge. 0) then
     write(info,'(''ERROR read_plane, index: '',2(1x,i9))')ii,iplane
     call abort_code (trim(info))
  endif
! data
#ifdef ASCII_READ
  read (1,*,err=20) plane
#else
  read (1,err=20) plane
#endif

  return
20 call abort_code ('ERROR in read_plane')
end subroutine read_plane_fragment


subroutine close_file(unit)

! Close a file

  use numbers
#ifdef HDF5
  use hdf5_wrapper; use pinocchio_hdf5
#endif
  implicit none
  integer(kind=singleI),intent(in):: unit

#ifdef HDF5
  call hdf5_close_file(file_handle)
#else
  close(1)
#endif

  return
end subroutine close_file

character(len=4) function write_string(i)

  use numbers
  implicit none
  integer(kind=singleI),intent(in)::i

  write(write_string,'(i4)')i
  if (i<=9) then 
     write_string(1:3)='000'
  else if (i<=99) then
     write_string(1:2)='00'
  else if (i<=999) then
     write_string(1:1)='0'
  endif
  return
end function write_string
