!*****************************************************************
!*                       PINOCCHI0  V2.2                         *
!*  (PINpointing Orbit-Crossing Collapsed HIerarchical Objects)  *
!*****************************************************************
!
!This code was written by
!Pierluigi Monaco, Tom Theuns, Giuliano Taffoni
!Copyright (C) 2008
!
!web page: http//adlibitum.oats.inaf.it/monaco/Homepage/Pinocchio/
!
!This program is free software; you can redistribute it and/or modify
!it under the terms of the GNU General Public License as published by
!the Free Software Foundation; either version 2 of the License, or
!(at your option) any later version.
!
!This program is distributed in the hope that it will be useful,
!but WITHOUT ANY WARRANTY; without even the implied warranty of
!MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!GNU General Public License for more details.
!
!You should have received a copy of the GNU General Public License
!along with this program; if not, write to the Free Software
!Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


! Reading and writing subroutines
! NB: Fragment version, no mpi


subroutine read_field(filename,flag,vector)

!  use run_module
  use pinocchio_common
  use grid
  use fftw_variables, ONLY: nx_fftw, ny_fftw, nz_fftw
  implicit none

! external variables
  character(*), intent(in) :: filename
  logical,      intent(in) :: flag
  real(kind=singleR), intent(out):: vector(Npart)
  logical, external        :: read_pedantic_header

! local variables
  real(kind=singleR),allocatable,dimension(:,:)::plane
  integer(kind=singleI)::iplane, in_plane, Last(3), kplane, ix, iy, index, unit


  nx_fftw = Lgrid_global(1)
  ny_fftw = Lgrid_global(2)
  nz_fftw = Lgrid_global(3)

  if (flag) then
!reads fmax-2.0 file

     Last=Lstart+Lgrid-Lgrid_global

     unit=1
     call open_file_to_read(0,unit,trim(filename),.true.)
     print *,'opened file '//trim(filename)

     allocate (plane(Lgrid_global(1),Lgrid_global(2)))

! read planes and saves the useful part
     do iplane=1,Lgrid_global(3)

        call read_plane_fragment(unit,iplane-1,Lgrid_global(1),Lgrid_global(2),plane)

        if ((iplane>=Lstart(3) .and. iplane<Lstart(3)+Lgrid(3)) .or. &
             iplane<Last(3) ) then
           if (iplane<Last(3)) then
              kplane=iplane-Lstart(3)+1+Lgrid_global(3)
           else
              kplane=iplane-Lstart(3)+1
           endif
           call shift_plane(Lgrid_global(1:2),Lstart(1:2),plane)

           do iy = 1,Lgrid(2)
              do ix = 1,Lgrid(1)

                 index = 1 + (ix-1) + ((iy-1) + (kplane-1)*Lgrid(2))*Lgrid(1)

                 vector(index) = plane (ix,iy)
              enddo
           enddo
        endif
     enddo

     call close_file(unit)
     deallocate(plane)

  else
! reads fmax-1.1 file
     open(10,file=filename,status='old',form='unformatted')
     read(10)vector
     close(10)
     print *,'Input file read: '//filename
  endif

! done
  print *,'Input file read: '//filename
  return
10 print *,' error opening input file '//filename
  stop
20 print *,' error reading input file '//filename
  stop

end subroutine read_field


subroutine shift_plane(L,S,plane)

! shifts a 2D array putting S as (1,1) [with PBCs]

  use numbers
  implicit none
  integer(kind=singleI),intent(in),dimension(2)::L,S
  real(kind=singleR),intent(inout)::plane(L(1),L(2))

! local variables
  integer(kind=singleI)::i
  real(kind=singleR),allocatable,save,dimension(:,:)::p


! x-direction
  allocate (p(L(1),L(2)))

  if (S(1)>1 .and. S(1)<=L(1)) then
     p=plane
     plane(1:L(1)-S(1)+1,:) = p(S(1):L(1),:)
     plane(L(1)-S(1)+2:L(1),:) = p(1:S(1)-1,:)
  endif

  if (S(2)>1 .and. S(2)<=L(2)) then
     p=plane
     plane(:,1:L(2)-S(2)+1)  = p(:,S(2):L(2))
     plane(:,L(2)-S(2)+2:L(2)) = p(:,1:S(2)-1)
  endif

  deallocate (p)

  return
end subroutine shift_plane
