!*****************************************************************
!*                       PINOCCHI0  V2.2                         *
!*  (PINpointing Orbit-Crossing Collapsed HIerarchical Objects)  *
!*****************************************************************
!
!This code was written by
!Pierluigi Monaco, Tom Theuns, Giuliano Taffoni
!Copyright (C) 2008
!
!web page: http//adlibitum.oats.inaf.it/monaco/Homepage/Pinocchio/
!
!This program is free software; you can redistribute it and/or modify
!it under the terms of the GNU General Public License as published by
!the Free Software Foundation; either version 2 of the License, or
!(at your option) any later version.
!
!This program is distributed in the hope that it will be useful,
!but WITHOUT ANY WARRANTY; without even the implied warranty of
!MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!GNU General Public License for more details.
!
!You should have received a copy of the GNU General Public License
!along with this program; if not, write to the Free Software
!Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

!
!
! XPOS: from grid to space (correcting for PBC)
! IPOS: from space to grid (correcting for PBC)
! DISTANCE: PBC-corrected distance between two space points
! IDISTANCE: PBC-corrected distance between two grid points
! PBC: PBC-corrected space position
! IPBC: PBC-corrected grid position
!
! Contains: xpos, ipos, distance, idistance, pbc, ipbc (functions)
!
real(kind=singleR) function xpos(i,ng,lbox)

! space position for grid number i (correcting for pbc)

  use numbers
  implicit none
  integer(kind=singleI),intent(in)::i,ng
  real(kind=singleR),intent(in)::lbox

! NB NB NB NB NB
  real(kind=singleR),parameter::center=0.5
! NB NB NB NB NB
  integer(kind=singleI)::i2

  i2=i
  if (i2.gt.ng) i2=i2-ng
  if (i2.lt.1) i2=i2+ng
  xpos=(i2-center)*lbox/float(ng)
  return
end function xpos


integer(kind=singleI) function ipos(x,ng,lbox)

! grid number for space position x (correcting for pbc)

  use numbers
  implicit none
  integer(kind=singleI),intent(in)::ng
  real(kind=singleR),intent(in)::x,lbox

  real(kind=singleR)::x2

  x2=x
  if (x2.ge.lbox) x2=x2-lbox
  if (x2.lt.0.) x2=x2+lbox
  ipos=int(x2*ng/lbox)+1
  return
end function ipos


real(kind=singleR) function distance(x1,x2,lbox)

! Gives x1 and x2 in a box of size lbox with PBC, it gives the 
! smallest distance x2-x1

  use numbers
  implicit none
  real(kind=singleR),intent(in)::x1,x2,lbox

  distance=x2-x1
  if (abs(distance).gt.lbox/2.) then
     if (x1.gt.lbox/2.) then
        distance=distance+lbox
     else
        distance=distance-lbox
     endif
  endif
  return
end function distance


integer(kind=singleI) function idistance(i1,i2,ng)

! Givex i1 and i2 in a box of size ng with PBC, it gives the 
! smallest distance i2-i1

  use numbers
  implicit none
  integer(kind=singleI),intent(in)::i1,i2,ng


  idistance=i2-i1
  if (abs(idistance).gt.ng/2) then
     if (i1.gt.ng/2) then
        idistance=idistance+ng
     else
        idistance=idistance-ng
     endif
  endif
  return
end function idistance


integer(kind=singleI) function ipbc(i,ng)

  use numbers
  implicit none
  integer(kind=singleI),intent(in)::i,ng
  ipbc=i
  if (i.gt.ng) ipbc=ipbc-ng
  if (i.lt.1) ipbc=ipbc+ng
  return
end function ipbc


real(kind=singleR) function pbc(x,lbox)

  use numbers
  implicit none
  real(kind=singleR),intent(in)::x,lbox
  pbc=x
  if (x.ge.lbox) pbc=pbc-lbox
  if (x.lt.0.) pbc=pbc+lbox
  return
end function pbc
